/*	Copyright (C) 1997-2008 Günter Woigk
				  kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	2000-05-06 kio	method 'TopOfTree()' renamed to 'Root()'
	2000-05-07 kio	member 'name' added
					root must not have neighbours; => root always isFirst and isLast
					revised behaviour of 'operator=()'
	2001-03-17 kio	Link()
	2002-09-01 kio	changed type of data member name to class String
					preserve this.name in operator=()
	2008-04-29 kio	reworked Check() macro
*/


#define LOG 	1
#define	SAFE	3


#include "HierObj.h"




/**********************************************************
					Hierarchical Objects
**********************************************************/


/* ==== Exzessive testing =================================
*/

/*static*/ 
void HierObj::Check ( cHierObj* p ) 
{
	ASSERT( p!=NULL );
	HierObj* c = p->child;
	if(c) do 
	{
		TRAP( c->parent!=p );
		TRAP( c->next->prev!=c );
		if( c->child ) Check(c);
		c = c->next;
	}
	while( c!=p->child );
}

/*static*/ 
inline void HierObj::Check( cHierObj& q ) { Check(&q); }



// ========================================================
// ====	Querries ==========================================
// ========================================================


/* ----	does this object contain the other object? ------------------	
		does not test this==other !!
	note:
		objects are either
		• identical (this==other)
		• intersect: 
			• a contains b or 
			• b contains a
		• disjunct 
			(a is not b and they do not intersect)
*/
bool HierObj::Contains ( cHierObj* q ) const
{
	do { q = q->parent; } while( q&&q!=this );
	return q==this;
}


/* ----	find object by name ------------------------------- kio 2001-08-10 / 2008-05-01 ----
		search for named object in own child tree
		does not compare name with own name!
		scans direct childs first and then childs of childs
		returns NULL if not found
		does not find unnamed objects!
		does not include the resume_position!
*/
HierObj* HierObj::FindChild ( cstr name, bool recursive, HierObj* p /*resume_position*/ ) const
{
	XXASSERT( p==NULL || this->Contains(p) );

	if(p) goto b;				// resume

	p = this->FirstChild();		// step in  (( start search ))
	if( p==NULL ) goto x;		// no childs => not found

a:	do							// scan direct childs
	{
		if( SameStr(name,p->GetName()) ) return p;
		p = p->next;
	}
	while( p->IsNotFirst() );
	
	if( !recursive ) goto x;	// not found

	do							// scan childs of childs
	{
		if( p->child ) { p = p->FirstChild(); goto a; }		// step in
b:		p = p->next;
	}
	while( p->IsNotFirst() );
	
	p = p->parent;				// step out
	if( p!=this ) goto b;		// resume with higher neighbours

x:	return NULL;				// not found
}




/* ----	find object by isa_id ------------------------------- kio 2008-05-01 ----
		search for next object of type id in own child tree
		does not compare id with own name!
		scans direct childs first and then childs of childs
		returns NULL if not found
		does not include the resume_position!
*/
HierObj* HierObj::FindChild ( int id, bool recursive, HierObj* p /*resume_position*/ ) const
{
	XXASSERT( p==NULL || this->Contains(p) );

	if(p) goto b;				// resume

	p = this->FirstChild();		// step in  (( start search ))
	if( p==NULL ) goto x;		// no childs => not found

a:	do							// scan direct childs
	{
		if( p->IsA(id) ) return p;
		p = p->next;
	}
	while( p->IsNotFirst() );
	
	if( !recursive ) goto x;	// not found

	do							// scan childs of childs
	{
		if( p->child ) { p = p->FirstChild(); goto a; }		// step in
b:		p = p->next;
	}
	while( p->IsNotFirst() );
	
	p = p->parent;				// step out
	if( p!=this ) goto b;		// resume with higher neighbours

x:	return NULL;				// not found
}


/* ----	count childs -------------------------------------
*/
long HierObj::CountChilds ( ) const
{
	XXXCheck(this);

	int      n = 0;
	HierObj* p = child;

	if(p) do { n++; p=p->next; } while( p!=child );

	return n;
}


/* ----	count all tree members ----------------------------
		incl. this object
*/
long HierObj::CountObjects ( ) const
{	
	XXXCheck(this);
	
	int      n = 1;
	HierObj* p = child;

	if(p) do 
	{
		n += p->child ? p->CountObjects() : 1;
		p = p->next;
	}	
	while( p!=child );

	return n;
}


/* ----	find top of tree ----------------------------------
*/
HierObj* HierObj::Root ( ) const
{
	XXXCheck(this);

	HierObj* p = (HierObj*)this;
	while( p->parent ) { p = p->parent; }

	return p;
}


/* ----	get own location in parent's child list -----------------
		indexes start at 0
*/
int HierObj::Index ( ) const
{
	int n = 0;

	if( parent )
	{
		XXXCheck(this);
		for ( HierObj* p = parent->child; p!=this; p=p->next ) { n++; }		
	}
	
	return n;
}


/* ----	compare locations in parent's child list -----------------
*/
bool HierObj::IsBefore ( HierObj const* other ) const
{
	if( parent && parent==other->parent )
	{
		XXXCheck(this);
	
		HierObj* p;
		for ( p = parent->child; p!=this&&p!=other; p=p->next ) { }
		return p!=other;
	}
	else
	{	// this->parent != other->parent
		// this and other are neighbours but !parent
		return no;
	}
}


/* ----	compare locations in parent's child list -----------------
*/
bool HierObj::IsBehind ( const HierObj* other ) const
{
	if( parent && parent==other->parent )
	{
		XXXCheck(this);	

		HierObj* p;
		for ( p = parent->child; p!=this&&p!=other; p=p->next ) { }
		return p!=this;
	}
	else
	{	// this->parent != other->parent
		// this and other are neighbours but !parent
		return no;
	}
}



// ============================================================
// ==== Managing the List =====================================
// ============================================================


/* ----	unlink object -------------------------------------
		unlink this object from parent and neighbours
		childs remain linked to this object
*/
void HierObj::Unlink ( )
{
	XXXCheck(this);

	if( parent )
	{
	// unlink from parent
		if( parent->child==this ) parent->child = this==next ? NULL : next;
		parent = NULL;

	// unlink from neighbours
		prev->next = next;
		next->prev = prev;
		prev = next = this;
	}
}


/* ----	insert this object into existing tree ------------------
		virtual!
		insert before/behind other object or
		in front/at end of other object's child tree
*/
void HierObj::Link ( int where, HierObj* other )
{
	XXASSERT( other!=this );

	XXXCheck(this);
	XXXCheck(other);

	if (parent) Unlink();
	assert(parent==0&&next==this);

	switch(where)
	{
	case before:
bf:	{
		XXASSERT( other->parent!=NULL );

		HierObj* p = other->prev;		// p -- this -- other
	
		p->next      = this;
		other->prev  = this;
		this->next   = other;
		this->prev   = p;
		this->parent = other->parent;
		if ( parent->child==other ) parent->child = this;
	}	break;
	
	case behind:
bh:	{
		XXASSERT( other->parent!=NULL );

		HierObj* n = other->next;		// other -- this -- n

		other->next	 = this;
		n->prev 	 = this;
		this->prev	 = other;
		this->next	 = n;
		this->parent = other->parent;
	}	break;

	case asfirstchild:
		if( other->child ) { other = other->child; goto bf; }
		goto ac;
	
	default:
	case aslastchild:
		if( other->child ) { other = other->child->prev; goto bh; }

ac:		parent = other;
		other->child = this;
		break;
	}
}



// ============================================================
// ==== Constructor / Destructor ==============================
// ============================================================


/* ----	setup this object ---------------------------------
		initialize cleared
		does not init name!
*/
inline void HierObj::_init ( )
{
	next  = prev = this; 
	parent= child= NULL; 
}


/* ----	destroy object on this location ------------------
		kill this object and all childs.
		this object becomes invalid.
		does not destroy name!
*/
void HierObj::_kill ( )
{
	if (parent) Unlink();		// also does check()
	while(child) delete child; 
}


/* ----	copy source to this location ----------------------
		copy an object and it's child tree from the source 
		position to the currently void target position. 
		source remains valid.
		does not copy name!
*/
void HierObj::_copy ( HierObj const& q )
{
	XXXCheck(q);

	_init();

	HierObj* c = q.child;
	if(c) do
	{
		c->Copy()->Link( aslastchild, this ); 
		c = c->next;
	}	while( c!=q.child );
}


/* ----	move source to this location ----------------------
		move an object from it's source memory position to
		the currently void target memory position. 
		source position becomes invalid.
		does not move name!
*/
void HierObj::_move ( HierObj const& q )
{
	XXXCheck(q);

	parent = q.parent;
	next   = q.next;
	prev   = q.prev;
	child  = q.child;

	next->prev = this;
	prev->next = this;

	if( parent&&parent->child==&q ) parent->child = this;

	HierObj* c = child;
	if(c) do { c->parent = this; c=c->next; } while( c!=child );
}


/* ----	destroy object ------------------------------------
*/
HierObj::~HierObj ( )
{
	_kill();
	if(*name) delete name;
}


/* ----	create object --------------------------------------
*/
HierObj::HierObj( )
:	name("")
{
	_init();
}

HierObj::HierObj ( cstr name )
:	name(name&&*name?NewCopy(name):"")
{
	_init();
}

HierObj::HierObj ( cHierObj& q )
{
	_copy(q); 
}



/* ----	copy entire tree ----------------------------------
		returns a copy of this object and all it's childs.
		this is a virtual function
		every derived object must provide it's version:
		
		DerivedObj* DerivedObj::Copy ( ) const
		{
			return new DerivedObj(*this);  
		}
*/
HierObj* HierObj::Copy ( ) const
{
	XXXCheck(this);
	return new HierObj(*this);
}


/* ----	assignment ------------------------------[13.mai.00]--------
		• store a copy of the source object and it's child tree 
		in this object replacing this object's contents and childs.
		• this object remains link at it's old position because in 
		case of self-assignment this object remains linked too.
		• the old target tree and source tree may intersect 
		because in this case the copy of the source is created prior 
		to deleting the old target tree.
		NOTE: Declare an operator=() for all derived classes, 
		whether you implement it or not. the flat copy C provides 
		by default is probably not what you expect!
		2002-09-01 kio:	this.name is preserved!
*/
HierObj& HierObj::operator= ( HierObj const& q )
{
	XXXCheck(this);
	XXXCheck(q);

	if( this!=&q )
	{
		bool x = false;
		for( HierObj*p=parent;   p&&!x; p=p->parent ) { x = p==&q;   }
		for( HierObj*p=q.parent; p&&!x; p=p->parent ) { x = p==this; }

		HierObj* pa = parent; if( pa && pa->child!=this ) pa=NULL;
		HierObj* pr = prev;   

		if(x)	// intersection between q and this
		{
			char c[sizeof(HierObj)];
			HierObj& z = (*(HierObj*)c);
			z._copy(q);				// create copy of source
			_kill();				// unlink and delete old target contents
			_move(z);				// move copy of source to target
		}	
		else	// no intersection between q and this
		{
			_kill();
			_copy(q);
		}

		if(pa) Link( asfirstchild, pa ); 
		else if(pr!=this) Link( behind, pr );
	}

	return *this;
}


















