/*	Copyright  (c)	Günter Woigk   1997-2008
  					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


Hierarchically organized objects
--------------------------------

	HierObjects are organized into a family tree.

	each object has one parent, two neighbours and no or more childs.
	each object has a name, which may be empty but is never NULL.

	neighbours are linked into a cyclic list. all neighbours have the same parent.

	the top object of a family tree (root) has no parent; that is: this->parent = NULL.
	the top object must not have neighbours => it is automatically first and last in it's neighbourhood list.

	if an object has childs, the child pointer points to the first child.
	other childs are accessible via their neighbourhood links.
	since childs are linked in a cyclic list, the last child is accessible via this->child->prev.

notes:

	the default creator for HierObjects creates an object which is the root object of a new tree 
	which only consists of this object.

	•important•: the destructor also •deletes• all childs of the object.
	therefore all objects in the tree must be dynamically allocated using •new•!
	if you put static/global/local objects into a tree, make shure that they are always unlinked first!
	however, it is ok to use a static/global/local object as top of the tree.

	every derived class must provide Copy() which must duplicate the entire object and it's child tree.
	the copy creator of the HierObj class which is called to copy the HierObj data member of your 
	derived class copies the child tree. therefore in most cases Copy() can be implemented like this:
	
		HierObj* DerivedObj::Copy ( ) const  {  return new DerivedObj(*this);  }
	
	duplicating a HierObj e.g. by operator=() or by the copy creator duplicates the entire subtree 
	starting at this object. therefore it is in most cases recommended to pass HierObj derived 
	parameters and results by reference, not by value.
*/


#ifndef	HierObj_h
#define	HierObj_h

#include	"cstrings/cstrings.h"

#undef		check			// may be defined in AssertMacros.h

typedef class			HierObj;
typedef HierObj const	cHierObj;


class HierObj 
{
	HierObj*	next;				// lower neighbour		last object: next   = top object
	HierObj*	prev;				// higher neighbour		top object:  prev	= bottom object
	HierObj*	child;				// bottom child			no childs:	 child  = NULL
	HierObj*	parent;				// parent object		top object:	 parent = NULL
	cstr		name;

inline void		_init				( );					// init this except this.name
	void		_copy				( cHierObj& q );		// copy this + tree except this.name
	void		_kill				( );					// destroy this except this.name
	void		_move				( cHierObj& q );		// move this except this.name, pull pointers
static void		Check				( cHierObj* q );		// extensive tree check
static void		Check				( cHierObj& q );

public:
				HierObj				( );
				HierObj				( cstr name );
				HierObj				( cHierObj& q );
virtual HierObj& operator=			( cHierObj& q );
virtual			~HierObj			( );
virtual HierObj* Copy				( )	const;				// create a copy of this with all childs

// ----	Querries ----
	cstr		GetName				( ) const				{ return name; }
virtual bool	IsA					( int id ) const		{ return no; }
	long		CountChilds			( )	const;				// number of childs
	long		CountObjects		( )	const;				// recursively count objects incl. this
	int			Index				( ) const;				// get own ordinal number in parent's child list (first child = 0)
	bool		IsFirst				( )	const				{ return !parent||parent->child==this; }
	bool		IsLast				( )	const				{ return !parent||parent->child==next; }
	bool		IsNotFirst			( )	const				{ return  parent&&parent->child!=this; }
	bool		IsNotLast			( )	const				{ return  parent&&parent->child!=next; }
	bool		IsBefore			( cHierObj* other ) const;
	bool		IsBehind			( cHierObj* other ) const;
	bool		Contains			( cHierObj* other ) const;									// does not test this==other !!
	bool		Intersect			( cHierObj* q ) const	{ return this->Contains(q) || q->Contains(this); }

	HierObj*	Prev				( )	const				{ return prev;   }					// may be this (cyclic ring)
	HierObj*	Next				( )	const				{ return next;   }					// may be this (cyclic ring)
	HierObj*	Parent				( )	const				{ return parent; }					// may be NULL (top of tree)
	HierObj*	Child				( )	const				{ return child;  }					// may be NULL
	HierObj*	FirstChild			( )	const				{ return child;  }					// may be NULL
	HierObj*	LastChild			( )	const				{ return child?child->prev:NULL;  }	// may be NULL
	HierObj*	Root				( )	const;				// return top of tree
	HierObj*	FindChild			( cstr name, bool recursive=no, HierObj* resume_position=NULL ) const;
	HierObj*	FindChild			( int id,    bool recursive=no, HierObj* resume_position=NULL ) const;

// ----	Actions ----
	void		Unlink				( );
	void		Link				( int where, HierObj* other );	enum { before,behind,asfirstchild,aslastchild };

	void		/*Insert*/LinkBeforeMe		( HierObj* newobj  )	{ newobj->Link ( before,       this ); } 
	void		/*Insert*/LinkBehindMe		( HierObj* newobj  )	{ newobj->Link ( behind,       this ); }
	void		/*Insert*/LinkAsFirstChild	( HierObj* newobj  )	{ newobj->Link ( asfirstchild, this ); } 
	void		/*Insert*/LinkAsLastChild	( HierObj* newobj  )	{ newobj->Link ( aslastchild,  this ); }

	void		MoveToFront			( )						{ if (IsNotFirst()) Link ( asfirstchild, parent ); }
	void		MoveToEnd			( )						{ if (IsNotLast())  Link ( aslastchild,  parent ); }
	void		MoveFore			( )						{ if (IsNotFirst()) Link ( before,       prev   ); }
	void		MoveBack			( )						{ if (IsNotLast())  Link ( behind,       next   ); }

	void		SetName				( cstr s )				{ if(*name)delete name; name=s&&*s?NewCopy(s):""; }
};





#endif






