/*	Copyright  (c)	Günter Woigk 2006 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		2

#import	<IOKit/IOCFPlugIn.h>
#import	<IOKit/hid/IOHIDLib.h>
#import	<IOKit/hid/IOHIDUsageTables.h>

#import	"Joystick.h"
#import	"KbdPanel.h"
#import	"Machine.h"
#import	"AlertPanel.h"


struct HIDJoystickCookies 
{
	IOHIDElementCookie		x_axis;
	IOHIDElementCookie		y_axis;
	IOHIDElementCookie		button_1;
	IOHIDElementCookie		button_2;
};

class HIDJoystick
{
	PLock						lock;
	IOHIDDeviceInterface122**	deviceInterface;
	HIDJoystickCookies			cookies;	

public:	
			HIDJoystick		( )			: lock(), deviceInterface(NULL) { }
	//		~HIDJoystick	( )			{ Unplug(); }

	void	Unplug			( );
	void	Plugin			( io_object_t );
	
	bool	IsPluggedIn		( )			{ return deviceInterface!=NULL; }
	uchar	GetState		( );
};


static HIDJoystick HIDJoysticks[2];

static void ShowDeviceProperties ( io_object_t myDevice );
static IOHIDDeviceInterface122** NewDeviceInterface ( io_object_t myDevice );
static bool GetCookies ( IOHIDDeviceInterface122**, HIDJoystickCookies& );



// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



/* ----	initialize physical device ----
		create device interface
		get cookies for x/y axis and buttons
		open device
*/
void HIDJoystick::Plugin ( io_object_t myDevice )
{	
	XLogIn("HIDJoystick:Plugin");
	if(XXXLOG) ShowDeviceProperties(myDevice);		// <-- very longish!
	XXXASSERT( deviceInterface==NULL );

	IOHIDDeviceInterface122** dev_if = NewDeviceInterface( myDevice );
	if( !dev_if || !*dev_if ) { ShowAlert( "HIDJoystick:Plugin:NewDeviceInterface = NULL" ); return; };

	bool is_js = GetCookies(dev_if,cookies);
	if( !is_js ) 
	{
		XLogLine( "this is no joystick!"); 
	}
	else
	{
		bool err = no;
		if( !cookies.x_axis )   { LogLine( "the device has no x axis."); err=yes; }
		if( !cookies.y_axis )   { LogLine( "the device has no y axis."); err=yes; }
		if( !cookies.button_1 ) { LogLine( "the device has no button."); err=yes; }

		if( !err ) 
		{
			IOReturn ioerror = (*dev_if)->open(dev_if, 0); 
			if( ioerror==kIOReturnSuccess ) 
			{
				PLocker z(lock); 
				deviceInterface = dev_if; 
				return; 	// OK: device is open
			}		
			ShowAlert( "HIDJoystick:Plugin:OpenDeviceInterface: error = $%X\n", (uint)ioerror ); 
		}
	}
	
// error:
	(*dev_if)->Release(dev_if);
}


/* ----	detach from physical device ----
*/
void HIDJoystick::Unplug ( )
{	
	XLogIn("HIDJoystick:Unplug");
	
	lock.Lock();
		IOHIDDeviceInterface122** dev_if = deviceInterface;
		deviceInterface = NULL;
	lock.Unlock();
	
	if( dev_if && *dev_if )
	{
		(*dev_if)->close(dev_if);
		(*dev_if)->Release(dev_if);
	}
}


/* ----	Get Joystick buttons state ------------
		returns result = %000FUDLR
*/
uchar HIDJoystick::GetState ( )
{
	XXXLogIn( "HIDJoystick:GetState" );

	PLocker z(lock);
	IOHIDDeviceInterface122** dev_if = deviceInterface;
	if( !dev_if || !*dev_if ) return 0;
	
	HRESULT/*SInt32*/	herror;
	IOHIDEventStruct	hidEvent;
	uchar x,y,b1; 
	
	int what=0;
	herror = (*dev_if)->getElementValue( dev_if, cookies.x_axis, &hidEvent );
	if (herror) goto xx;
	x = hidEvent.value;		// seen values: 	min=0  ..  dflt=127..130  ..  max=255

	what=1;
	herror = (*dev_if)->getElementValue( dev_if, cookies.y_axis, &hidEvent );
	if (herror) goto xx;
	y = hidEvent.value;

	what=2;
	herror = (*dev_if)->getElementValue( dev_if, cookies.button_1, &hidEvent );
	if (herror) goto xx;
	b1 = hidEvent.value; 

	XXXLogLine( "x-axis = %i; y-axis = %i; button1 = %i", (int)x, (int)y, (int)b1 );

	static uchar xval[] = { joystick_paddle_left_mask, 0, 0, joystick_paddle_right_mask };
	static uchar yval[] = { joystick_paddle_up_mask,   0, 0, joystick_paddle_down_mask  };

	return	( b1 ? joystick_button1_mask : 0 ) | xval[x>>6] | yval[y>>6];
//	return	( b1 ? joystick_button1_mask : 0 ) | 
//			(x==127?0:x<127?joystick_paddle_left_mask:joystick_paddle_right_mask) | 
//			(y==127?0:y<127?joystick_paddle_up_mask:joystick_paddle_down_mask);

// show error:
// note: ShowAlert() may show render errors because of thread != mainThread!
xx:	static cstr names[]={"x-axis","y-axis","button1"};
	ShowAlert( "HIDJoystick:GetState:getElementValue:%s: error = $%X\n", names[what], (uint)herror ); 
	
// unplug:
	//Unplug();						needs lock created with ATTR = PTHREAD_MUTEX_RECURSIVE
	(*dev_if)->close(dev_if);		
	(*dev_if)->Release(dev_if);
	deviceInterface = NULL;
	return 0;
}



// ================================================================
// ================================================================
// ================================================================



/* ----	Print Device Properties -------------
		((from the XCode Docs))
		print the property dictionary 
		for the supplied device.
		warning: prints really a lot! 
*/
static void ShowDeviceProperties ( io_registry_entry_t /* io_object_t */ myDevice )
{
	kern_return_t			result;
	CFMutableDictionaryRef	properties = NULL;
	char					path[512];

	LogIn("\nShowDeviceProperties");

	result = IORegistryEntryGetPath ( myDevice, kIOServicePlane, path );
	if( result == KERN_SUCCESS ) { LogLine( "path = \"%s\"\n", path ); }
	else						 { LogLine( "IORegistryEntryGetPath() error = $%X\n", (uint)result ); }

// Create a CF dictionary representation of the I/O Registry entry's properties
	result = IORegistryEntryCreateCFProperties( myDevice, &properties, kCFAllocatorDefault, kNilOptions );
	if ( (result != KERN_SUCCESS) || properties==NULL )
	{ LogLine( "IORegistryEntryCreateCFProperties() error = $%X\n", (uint)result ); return; }

// Some common properties of interest:
//	kIOHIDTransportKey, kIOHIDVendorIDKey, kIOHIDProductIDKey, kIOHIDVersionNumberKey,
//	kIOHIDManufacturerKey, kIOHIDProductKey, kIOHIDSerialNumberKey, kIOHIDLocationIDKey,
//	kIOHIDPrimaryUsageKey, kIOHIDPrimaryUsagePageKey, kIOHIDElementKey. 
	CFShow( properties );
	
	CFRelease( properties );
	LogNL();
}



/* ----	Get a device interface for a device -------------
		((from the XCode Docs))
		struct IOHIDDeviceInterface122 is the extended 
		IOHIDDeviceInterface since OSX 10.3
*/
static IOHIDDeviceInterface122** NewDeviceInterface ( io_object_t myDevice )
{
	XLogIn("NewDeviceInterface");
 
//	io_name_t                 className;
//	ioerror = IOObjectGetClass( myDevice, className );
//	if( ioerror != kIOReturnSuccess ) 
//	{ LogLine( "NewDeviceInterface:IOObjectGetClass: error = $%X", (uint)ioerror ); return NULL; }
//	XLogLine( "Device class = %s", className );	// e.g. "IOUSBHIDDriver"
 
    IOCFPlugInInterface**	  plugInInterface = NULL;
	IOHIDDeviceInterface122** deviceInterface = NULL;
    HRESULT                   herror  = S_OK;
    IOReturn                  ioerror = kIOReturnSuccess;
    SInt32                    score   = 0;

// Create a plugin interface
    ioerror = IOCreatePlugInInterfaceForService (
			myDevice, kIOHIDDeviceUserClientTypeID, kIOCFPlugInInterfaceID,
			&plugInInterface, &score );
 
    if( ioerror != kIOReturnSuccess ) 
	{ ShowAlert( "NewDeviceInterface:IOCreatePlugInInterfaceForService: error = $%X\n", (uint)ioerror ); return NULL; }	
	XXASSERT( plugInInterface && *plugInInterface );

// Create the device interface
	herror = (*plugInInterface)->QueryInterface (
			plugInInterface, 
			CFUUIDGetUUIDBytes( kIOHIDDeviceInterfaceID ),
			(void**)&deviceInterface );

// Purge the plugin interface
	(*plugInInterface)->Release(plugInInterface);

    if( herror != S_OK )
	{ ShowAlert( "NewDeviceInterface:QueryInterface: error = $%X\n", (uint)herror ); return NULL; }

	return deviceInterface;
}



/* ----	Get Cookies for x/y axis and button ----
		((from the XCode docs))
		retrieves cookies for HIDJoystick::cookies	((struct HIDJoystickCookies))
		via HIDJoystick::deviceInterface			((IOHIDDeviceInterface122**))
		and determines HIDJoystick::isJoystick.
*/
bool GetCookies ( IOHIDDeviceInterface122** deviceInterface, HIDJoystickCookies& cookies )
{       
 //	HIDJoystickCookies	cookies;					cookies to be returned here
 // bool isJoystick;								set if this is a joystick or equivalent
 // IOHIDDeviceInterface122** deviceInterface;		interface to get cookies from
	CFTypeRef			object;
	long				number;         
	IOHIDElementCookie	cookie;
	long				usage;
	long				usagePage;
	CFArrayRef			elements;
	CFDictionaryRef		element;
	IOReturn			ioerror;
	
	XLogIn( "GetCookies" );

	XXXASSERT( deviceInterface && *deviceInterface );
	memset( &cookies, 0, sizeof(cookies) );
	bool isJoystick = 0;

// O-Ton Apple: Copy all elements, since we're grabbing most of the elements for this device anyway, 
// and thus, it's faster to iterate them ourselves. When grabbing only one or two elements, 
// a matching dictionary should be passed in here instead of NULL.
	ioerror = (*deviceInterface)->copyMatchingElements(deviceInterface, NULL, &elements);
	if( ioerror != kIOReturnSuccess ) 
	{ ShowAlert( "GetCookies: copyMatchingElements: error = $%X", (uint)ioerror ); return no/*is_js*/; }
	XLogLine( "%i elements", (int)CFArrayGetCount(elements) );

// Loop over elements:
	for ( CFIndex i=0; i<CFArrayGetCount(elements); i++ )
	{
		element = (CFDictionaryRef)CFArrayGetValueAtIndex(elements, i);
		//LogLine("\nElement #%i:",(int)i);CFShow(element);LogNL();

	// Get usage page
		object = CFDictionaryGetValue( element, CFSTR(kIOHIDElementUsagePageKey) );
		if( object==0 || CFGetTypeID(object) != CFNumberGetTypeID() ) continue;
		if( !CFNumberGetValue((CFNumberRef) object, kCFNumberLongType, &number) ) continue;
		usagePage = number;
		if( usagePage!=kHIDPage_GenericDesktop && usagePage!=kHIDPage_Button ) continue;

	// Get usage
		object = CFDictionaryGetValue( element, CFSTR(kIOHIDElementUsageKey) );
		if( object==0 || CFGetTypeID(object) != CFNumberGetTypeID() ) continue;
		if( !CFNumberGetValue((CFNumberRef) object, kCFNumberLongType, &number) ) continue;
		usage = number;

	// Get cookie
		object = CFDictionaryGetValue( element, CFSTR(kIOHIDElementCookieKey) );
		if( object==0 || CFGetTypeID(object) != CFNumberGetTypeID() ) continue;
		if( !CFNumberGetValue((CFNumberRef) object, kCFNumberLongType, &number) ) continue;
		cookie = (IOHIDElementCookie) number;

	// Check for x and y axis and application		
		XXLog( "usagePage = $%04X, usage = $%02X, cookie=$%08X", (int)usagePage, (int)usage, (ulong)cookie );
		if( usagePage == kHIDPage_GenericDesktop /* 1 */ )
		{
			switch( usage )
			{
			case kHIDUsage_GD_Pointer:	/* 1 */ 		XLog( "  pointer device\n" );			break;
			case kHIDUsage_GD_Mouse:	/* 2 */ 		XLog( "  mouse\n" );					return no/*is_js*/;
			case kHIDUsage_GD_Joystick:	/* 4 */ 		XLog( "  joystick\n" ); isJoystick=yes; break;
			case kHIDUsage_GD_GamePad:	/* 5 */ 		XLog( "  game pad\n" ); isJoystick=yes; break;
			case kHIDUsage_GD_Keyboard:	/* 6 */ 		XLog( "  keyboard\n" );					return no/*is_js*/;
			case kHIDUsage_GD_Keypad:	/* 7 */ 		XLog( "  keypad\n" );					return no/*is_js*/;
			case kHIDUsage_GD_MultiAxisController:/*8*/	XLog( "  multi-axis controller\n" );  isJoystick = yes;	 break;
			case kHIDUsage_GD_X:		/* $30 */ 		XLog( "  x-axis\n" );	cookies.x_axis = cookie; break;
			case kHIDUsage_GD_Y:		/* $31 */		XLog( "  y-axis\n" );	cookies.y_axis = cookie; break;
			default:									XXLogNL(); break;
			}
		}
	// Check for buttons
		else if( usagePage == kHIDPage_Button /* 9 */ )
		{
			switch( usage )
			{
			case kHIDUsage_Button_1: /* 1 */	cookies.button_1 = cookie; XLog( "  button 1\n" ); break;
			case kHIDUsage_Button_2: /* 2 */	cookies.button_2 = cookie; XLog( "  button 2\n" ); break;
			default:							XXLogNL(); break;
			}
		}
	}
	return isJoystick;
}



/* ----	Search the IO registry for all HID devices ----
*/
void FindJoysticks ( bool force )
{
	static bool firstcall = yes;
	if( !firstcall && !force ) return;
	firstcall = no;

	XLogNL();
	XLogIn("FindJoysticks");
	
	HIDJoysticks[0].Unplug();
	HIDJoysticks[1].Unplug();

	io_iterator_t     objectIterator = 0;
	const mach_port_t masterPort     = kIOMasterPortDefault;	// or from IOMasterPort()
	IOReturn		  ioReturnValue  = kIOReturnSuccess;

// What do we search?
// Create a matching dictionary to search the I/O Registry by class name for all HID class devices
	CFMutableDictionaryRef myMatchDictionary = IOServiceMatching( kIOHIDDeviceKey );
	//CFShow(myMatchDictionary);
 
// Search I/O Registry for matching devices
	ioReturnValue = IOServiceGetMatchingServices( masterPort, myMatchDictionary, &objectIterator ); 
	myMatchDictionary = NULL;		// the dictionary was consumed by IOServiceGetMatchingServices() 	
	if( !objectIterator ) { XLogLine( "FindJoysticks:objectIterator = 0" ); return; }
	XXASSERT( ioReturnValue == kIOReturnSuccess );

// Loop over all devices found:
	io_object_t hidDevice = 0;
	uint i=0,j=0;
	while( (hidDevice=IOIteratorNext(objectIterator)) )
	{	 
		j++;
		HIDJoysticks[i].Plugin(hidDevice); LogNL();
		if( HIDJoysticks[i].IsPluggedIn() && ++i==2 ) break;
	}

	XLogLine( "found %i device%s, thereof %i joystick%s\n", (int)j, j==1?"":"s", (int)i, i==1?"":"s" );

    // Release iterator. Don't need to release iterator objects.
	IOObjectRelease( objectIterator );
}


bool IsPluggedIn ( JoystickID j )
{
	XXXLogIn("IsPluggedIn(JoystickID %i)",(int)j);

	bool f;
	switch(j)
	{
	case hid_joystick1:	f = HIDJoysticks[0].IsPluggedIn(); break;
	case hid_joystick2:	f = HIDJoysticks[1].IsPluggedIn(); break;
	case kbd_joystick:	f = yes; break;
	default:			f = no;  break;
	}
	
	XXXLogLine("-> %s",f?"yes":"no");
	return f;
}


uchar GetState ( JoystickID j )
{
	XXXLogIn("GetState(JoystickID)");

	switch(j)
	{
	case hid_joystick1:	return HIDJoysticks[0].GetState();
	case hid_joystick2:	return HIDJoysticks[1].GetState();
	case kbd_joystick:	return KeyboardJoystickGetState();
	default:			return 0x00;
	}
}







