/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#import "globals.h"
#import "unix/pthreads.h"
#import "ZxInfo.h"
#import "HierObj.h"
#import "Z80 options.h"
#import "Ula.h"
#import "templates/SharedArray.h"
#import "MachineList.h"


typedef SharedArray<ushort> Mem;


// read access to front machine from the main event loop:
#define	RLOCK		XXXASSERT(MAIN_THREAD); Machine* m = front_machine(); if(m) {
#define	RUNLOCK		}

// quick modify front machine from the main event loop:
// runForSound is blocked while machine is locked!
#define	QBLOCK		XXXASSERT(MAIN_THREAD); Machine* m = front_machine(); if(m) { m->Lock();
#define	QUNBLOCK	m->Unlock(); }

// longish modify front machine from the main event loop:
// halts cpu temporarily to avoid blocking of runForSound
#define	LBLOCK		XXXASSERT(MAIN_THREAD); Machine* m = front_machine(); if(m) { m->HaltCpu(); m->Lock();
#define	LUNBLOCK	m->Unlock(); m->ResumeCpu(); }
#define	LNOBLOCK	}	/* for MenuController:doCloseTopWindow() */

// read & write access to front machine from other threads:
#define	BLOCK		Machine* m; { PLocker z(&mlock); m = front_machine(); if(m) m->Lock(); } if(m) {
#define	UNBLOCK		m->Unlock(); }



//	Machine.tape_options:
enum	
{	halt_cpu_after_block_loaded	= 1,
	instant_load_from_tape		= 2,
	instant_save_to_tape		= 4,
	autostart_tape				= 8
};




class Machine : public HierObj
{
	PLock			lock;					// data access lock

// general info
	ZxInfo const*	model;					// all model info
	cstr			filepath;				// Snapshot file path or nil

// memory
	Mem				rom;					// rom pages
	Mem				ram;					// ram pages
	bool			rom_dis;				// set from external item, e.g. IF2			

// patches:
	int				patches;				// count

// machine state
	int				cpu_state;				// 0 = running; 1..7 = halted
	int				tape_state;				// set during rom load/save routine
	
// options
	int				cpu_options;			// cpu_waitmap | cpu_crtc | cpu_breap_sp | cpu_break_rwx
	int				tape_options;

// instant tape loader handling:
	int				instant_load_fd;		// open file for reading
	cstr			instant_load_filename;	// for error messages
	FileType		instant_load_filetype;	// file type: see TapeData.h

	int				instant_save_fd;		// open file for writing
	cstr			instant_save_filename;	// for error messages
	FileType		instant_save_filetype;	// file type: see TapeData.h
 
// keyboard
	uint			keyboard_mode;

// virtual machine time:
	Frequency		cpu_cycles_per_second;

	Time			time_dsp_start;			// seconds-based timestamp of start of current dsp buffer 
	long			cc_dsp_start;			// cpu->cc at start of current dsp buffer

// autostart taperecorder:	
	int				stop_tape_count;		// ""
	bool			is_saving;				// ""

// internal use:
	void			run_for_sound	();
	void			set_cpu_clock_and_predivider( Time cc_per_sec, int uc_per_cc );
	void			clear_break_ptr	();
	void			set_title		();
	void			load_preferences();
	void			dealloc_items	();

	void			init_with_model	( uint model );		// designated initializer
	void			init_with_default_model();			// with model from prefs
	cstr			init_with_file	( cstr file_path );
	void			kill			();

	Machine&		operator=		( Machine const& );	// prohibit
					Machine			( Machine const& );	// prohibit
	void			Unlink			( );				// prohibit
	void			Link			( int, HierObj* );	// prohibit
	void			LinkBeforeMe	( HierObj* );		// prohibit
	void			LinkBehindMe	( HierObj* );		// prohibit
	void			LinkAsFirstChild( HierObj* );		// prohibit
	void			LinkAsLastChild	( HierObj* );		// prohibit
	void			MoveToFront		( );				// prohibit
	void			MoveToEnd		( );				// prohibit
	void			MoveFore		( );				// prohibit
	void			MoveBack		( );				// prohibit
	ZxItem*			FirstChild		( );				// prohibit

public:

// breakpoints
	int				breakpoints;			// User-Breakpoints
	ushort*			breakpoint[100];
	ushort*			break_ptr;				// breaking pointer after cpu_break_[rwx]
											// also set to pc for startup when sitting on an x break point

// access pointers for special items:
	Screen*						screen;					// ((not derived from class Item))
	Z80*						cpu;
	Ula*						ula;
	Mmu*						mmu;
	KempstonMouseInterface*		kempston_mouse_interface;
	KempstonJoystickInterface*	kempston_joystick_interface;
	SinclairZxInterface1*		sinclair_zx_interface1;
	SinclairZxInterface2*		sinclair_zx_interface2;
	SinclairZxPrinter*			sinclair_zx_printer;
	FullerBox*					fuller_box;
	ZonxBox*					zonx_box;
	ICTester*					ic_tester;
	ZxPlus2JoystickPorts*		zxplus2_js_ports;		// built-in JS ports of the +2, +3 and +2A/B


public:
					Machine				();
					Machine				(uint model);
					Machine				(cstr file_path);
					~Machine			();

// HierObj methods:

	ZxItem*			FirstItem			( )					{ return (ZxItem*)HierObj::FirstChild(); }
	void			LinkItemWithIO		( ZxItem* i )		{ HierObj::LinkAsFirstChild(i); }
	void			LinkItemWithoutIO	( ZxItem* i )		{ HierObj::LinkAsLastChild(i); }

	virtual bool	IsA					( int id ) const	{ return id==isa_Machine || HierObj::IsA(id); }
	
	
//	Load and Save:

	void			SetModel			(uint model);
	OSStatus		LoadFile			(cstr file_path);
	cstr			init_with_sna		(cstr path);
	cstr			init_with_z80		(cstr path);
	cstr			init_with_rom		(cstr path);
	cstr			init_with_scr		(cstr path);
	cstr			init_with_ass		(cstr srcfile);
	cstr			init_with_80		(cstr path);
	cstr			init_with_81		(cstr path);
	void			ReloadSnapshot		();

	OSStatus		Save_80				(cstr path)			throw();
	OSStatus		Save_81				(cstr path)			throw();
	OSStatus		SaveSNA				(cstr path)			throw();
	OSStatus		SaveZ80				(cstr path);
	OSStatus		SaveROM				(cstr path)			throw();
	OSStatus		SaveSCR				(cstr path)			throw();
	void			SaveFile			(cstr file_path);
	void			SaveScreenshot		(cstr file_path=NULL);
	void			RecordMovie			(cstr file_path=NULL);
	
	void			SaveZX80TapeBlock	(int fd, FileType)	throw(file_error, bad_alloc);
	void			SaveZX81TapeBlock	(int fd, FileType)	throw(file_error, bad_alloc);
	void			SaveZxspTapeBlock	(int fd, FileType)	throw(file_error, bad_alloc);

	void			LoadZX80TapeBlock	(int fd, FileType)	throw(file_error, bad_alloc, any_error);
	void			LoadZX81TapeBlock	(int fd, FileType)	throw(file_error, bad_alloc, any_error);
	void			LoadZxspTapeBlock	(int fd, FileType)	throw(file_error, bad_alloc);
	
	uint			GetModel			()			{ return model->model; }
	ZxInfo const*	GetInfo				()			{ return model; }

//	Z80 callbacks:

	void			OutputAtCycle		(long cc, ushort address, uchar byte);
	uchar			InputAtCycle		(long cc, ushort address);
	uchar			PatchAtCycle		(long cc, ushort pc, uchar instr);

	void			InfoFirstLocAtCycle			(long cc, ushort pc);
	bool			IllegalInstructionAtCycle	(long cc, ushort pc);
	uchar			InterruptAcknowledgedAtCycle(long cc, ushort pc);
	void			NmiAcknowledgedAtCycle		(long cc, ushort pc);
	void			InstructionRetiAtCycle		(long cc, ushort pc);
	void			InstructionRetnAtCycle		(long cc, ushort pc);

//	Items:

//	void			RegisterItem		(ZxItem*);
//	void			UnregisterItem		(ZxItem*);
	ZxItem*			FindItem			( int id )	{ return (ZxItem*)HierObj::FindChild(id); }

	Ay*				new_ay_for_this_machine	();
	Mmu*			new_mmu_for_this_machine();
	Z80*			GetCpu				()			{ return cpu; }
	Ay*				GetAy				()			{ return (Ay*)FindItem(isa_Ay); }	// TODO
	Ula*			GetUla				()			{ return ula; }
	ICTester*		GetICTester			()			{ return ic_tester; }
	Mmu*			GetMmu				()			{ return mmu; }

	Screen*			GetScreen			()			{ return screen; }
	Mem&			GetRom				()			{ return rom; }
	void			DisableInternalRom	()			{ rom_dis=yes; }
	void			EnableInternalRom	()			{ rom_dis=no; }
	bool			InternalRomDisabled	()			{ return rom_dis; }
	Mem&			GetRam				()			{ return ram; }
	ushort*			GetVideoRam			();

	void			ToggleInternalAy				();
	void			ToggleFullerBox					();
	void			ToggleZonxBox					();
	void			ToggleKempstonJoystickInterface	();
	void			ToggleKempstonMouseInterface	();
	void			ToggleSinclairZxInterface2		();
	void			ToggleSinclairZxPrinter			();
	void			ToggleSinclairZxInterface1		();

	void			toggle_ic_tester	();
	void			toggle_test_ic_none	();
	void			toggle_test_ic_7400	();
	void			toggle_test_ic_7402	();
	void			toggle_test_ic_7406	();
	void			toggle_test_ic_27128();


//	Machine state:

	bool			IsRunning			()			{ return (cpu_state&7)==0; }	
	bool			IsHalted			()			{ return (cpu_state&7)!=0; }	
	void			HaltCpu				(bool f)	{ if(f) HaltCpu(); else ResumeCpu(); }
	void			HaltCpu				()			{ PLocker z(&lock); cpu_state++; XXXASSERT(cpu_state&7); }
	void			ResumeCpu			()			{ PLocker z(&lock); if (cpu_state&7) { cpu_state--; clear_break_ptr(); } }

	void			Reset				();
	void			Nmi					();
	void			PowerOn				();
	void			Lock				()			{ lock.Lock(); }		// short-timed blocking only!
	void			Unlock				()			{ lock.Unlock(); }		// runForSound() hangs on this lock!
	bool			TryLock				()			{ return lock.TryLock(); }
	Machine*		Next				()			{ return (Machine*)HierObj::Next(); }
	Machine*		Prev				()			{ return (Machine*)HierObj::Prev(); }

//	Running the machine:

	void			RunForSound			();
	void			AutoStartStopTape	();

	void			StepOver			();
	void			StepIn				();
	void			StepOut				();
	void			RunCpuCycles		(long cc);

	void			SetBreakpoint		(ushort* brkPtr, int rwx_mode);
	void			ClearBreakpoint		(ushort* brkPtr, int rwx_mode);
	void			ClearAllBreakpoints	();

	void			RegisterRomPatch	( uint addr );
	void			UnregisterRomPatch	( uint addr );
	void			ClearAllPatches		();


//	Options:

	void			SavePreferencesNow		();
	cstr			GetFilePath				()		{ return filepath; }
	int				GetCpuOptions			()		{ return cpu_options; }

	void			SetKeyboardMode			(uint mode);
	uint			GetKeyboardMode			()		{ return keyboard_mode; }
		
	void			SetFullscreen			(bool yes_no);
	bool			IsFullscreen			();
	void			SetWindowZoom			(int factor);

	void			SetBeeperVolume			(Sample volume);
	Sample			GetBeeperVolume			();

	void			SetUseWaitmap			(bool yes_no);
	bool			UsesWaitmap				()		{ return cpu_options & cpu_waitmap; }

	void			SetCpuClockPredivider	(int new_uc_per_cc);		// keep ula clock,  change predivider => change cpu clock
	void			SetCpuCyclesPerSecond	(Time new_cc_per_sec);	// keep predivider, change ula clock, cpu clock, frame rate
	void			SetMachineOverdrive		(double ula_clock_factor);	// keep predivider, change ula clock, cpu clock, frame rate
	double			GetCpuCyclesPerSample	()		{ return cpu_cycles_per_second / samples_per_second; }
	double			GetCpuCyclesPerFrame	()		{ XXXASSERT(ula); return ula->GetCpuCyclesPerFrame(); }
	long			GetCurrentCpuCycle		();
	Frequency		GetCpuCyclesPerSecond	()		{ return cpu_cycles_per_second; }

	void			SetHaltCpuAfterLoad		(bool);
	bool			HaltsCpuAfterBlockLoaded()		{ return tape_options & halt_cpu_after_block_loaded; }

	void			SetAutostartTape		(bool);
	bool			AutostartsTape			()		{ return tape_options & autostart_tape; }

	void			SetInstantLoadFromTape	(cstr/*or NULL for close/off*/);	
	void			SetInstantSaveToTape	(cstr/*or NULL for close/off*/);	
	bool			InstantLoadsFromTape	()		{ return tape_options & instant_load_from_tape; }
	cstr			InstantLoadFilename		()		{ return instant_load_filename; /* may be NULL */ }	
	bool			InstantSavesToTape		()		{ return tape_options & instant_save_to_tape; }
	cstr			InstantSaveFilename		()		{ return instant_save_filename; /* may be NULL */ }	

	void			MoveToFrontInMList		();

	friend class Z80;
};


