/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define LOG		1
#define SAFE	3

#include "Mouse.h"
#include "Machine.h"
#include "DisplayList.h"
#include "Screen.h"
#include "utilities.h"


namespace Mouse
{
	int  hide_mouse_count	= 0;
	bool mouse_is_nailed	= no;
	bool mouse_is_grabbed	= no;

	bool	left_button		= 0;
	bool	right_button	= 0;
	bool	middle_button	= 0;

	uint	scaling			= 2;
	uint	scaled_x		= 0;
	uint	scaled_y		= 0;


void HandleCommandKey()
{
	if( !fullscreen_mode || display_list.Count()>1 )  ReleaseMouse();
}


/*	Handle Event
	special handling of events while mouse is grabbed
	to be called from MainApp:handleEvent:
	returns 'true' if handled
*/
bool HandleEvent( NSEvent* event )
{
	XXLogIn("Mouse:HandleEvent: %i",(int)[event type]);

	XXASSERT( mouse_is_grabbed );

	switch( [event type] )
	{
	case NSLeftMouseDown:	left_button  = 1; return yes;
	case NSRightMouseDown:	right_button = 1; return yes;
	case NSOtherMouseDown:	middle_button= 1; return yes;

	case NSLeftMouseUp:		left_button  = 0; return yes;
	case NSRightMouseUp:	right_button = 0; return yes;
	case NSOtherMouseUp:	middle_button= 0; return yes;

	case NSLeftMouseDragged:		
	case NSRightMouseDragged:		
	case NSOtherMouseDragged:		
	case NSMouseMoved:		
		CGMouseDelta dx, dy;
		CGGetLastMouseDelta( &dx, &dy );	// mouse movement since last mouseMoved event
		scaled_x += dx;
		scaled_y -= dy;
		return yes;	/* handled */

	case NSFlagsChanged:
		if( [event modifierFlags] & NSCommandKeyMask ) HandleCommandKey();
//	case NSKeyUp:
//	case NSKeyDown:	
//		return no;	/* not handled */

	default:
		return no;	/* not handled */
	}
}


/*	Grab mouse.
	Even if mouse is already grabbed, move mouse pointer to middle of screen.
*/
void GrabMouse( )
{	
	XLogIn("Mouse:GrabMouse");
	
	RLOCK;
		Screen* screen = m->GetScreen();
		[screen makeKeyWindow];
		
		if( !mouse_is_grabbed )
		{
			HideMouse();						// hide mouse
			NailMousePointer(yes);				// Nail it there
			[screen setAcceptsMouseMovedEvents:yes];	// we want mouseMoved events
			mouse_is_grabbed = yes;
		}

	// Window frame (biased to bottom/left corners)
		NSRect  b = [screen frame];

	// center of window (biased bottom/left)
		NSPoint p = { b.origin.x+b.size.width/2, b.origin.y+b.size.height/2 };	

	// All-enclosing workspace rect (biased to top/left corner)
		CGRect cgbox = fullscreen_mode 
					 ? [screen getFSDisplayRef]->GetDisplayBounds()
					 : CGDisplayBounds(kCGDirectMainDisplay);

	// center of window (biased top/left)
		CGPoint cgpnt = { p.x, cgbox.origin.y+cgbox.size.height - p.y };						

	// locate mouse to center of screen
		MoveMouseTo(cgpnt);

		XXLogLine("Workspace rect:     %s (biased to top/left corners)",BoundsStr(cgbox));
		XXLogLine("Screen window rect: %s (biased to bottom/left corners)",BoundsStr(b));
		XXLogLine("Move mouse to:      x=%i, y=%i (workspace coord)",(int)cgpnt.x,(int)cgpnt.y);

	RUNLOCK;
}

void ReleaseMouse( )
{
	XLogIn("Mouse:ReleaseMouse");

	if( !mouse_is_grabbed ) return;

	RLOCK;
		Screen* screen = m->GetScreen();
		NailMousePointer(no);
		ShowMouse();
		[screen setAcceptsMouseMovedEvents:no];
		mouse_is_grabbed = no;
	RUNLOCK;
}


}	// namespace




