/*	Copyright  (c)	Günter Woigk 2006 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define SAFE	3
#define LOG 	0

#import "utilities.h"
#import "NXPanel.h"
#import "NXTextField.h"
#import "NXWindowMenu.h"
#import "Machine.h"
#import "unix/tempmem.h"


static NXPanel* panel[20];			// list of all panels
static uint		panels=0;
static NSTimer* update_timer = NULL;


@implementation NXPanel


/*	Update Panel: 
	- call from main event loop
	- for regular panel update
	- called only if panel is visible
	- front machine m is locked for reading (RLOCK/RUNLOCK)

	If this method is not reimplemented, that is, if the original method is called, 
	then this method is no longer called for this panel.
*/
-(void)updatePanel:(Machine*)m
{
	XLogIn("NXPanel:updatePanel");	

	// NXPanel:updatePanel is called:
	// => updatePanel is not reimplemented.
	// => clear do_update flag.

	XLogLine("panel \"%s\" is removed from update list", [[self title] UTF8String]);
	do_update = no; 
}


/* ----	Update Interrupt for all NXPanels ----
		This routine is sheduled to be called from the main event loop
		Only one panel is updated per event.
		2008-05-15: also purges the mainthread tempmempool.
*/
+(void)updateAllPanels
{
	PurgeTempMem();

	for (uint n=0;n<panels;n++)					// find next visible panel
	{
		static uint i = 0; i = (i+1) % panels;	// next panel (round robin beyond interrupt calls)
		NXPanel*    p = panel[i];				// next panel
		BOOL        v = [p isVisible];			// visible?
		NSMenuItem* m = p->menu_item;			// it's window menu item

		if( m && [m state]!=v ) [m setState:v];	// if panel has menu item and item state is wrong: update item

		if(v)									// visible?
		{
			if (!m)								// add window menu item if not yet exists
			{				
				p->menu_item = m = [window_menu createMenuItemForPanel:p];
				[m setTarget:p];
				[m setState:v];
			}
			if(p->do_update)
			{
				RLOCK; [p updatePanel:m/*machine*/]; RUNLOCK;	// update panel (may be longish)
				return;							// therefore return after updating one panel
			}									// next panel will be updated on next interrupt
		}
	}
}


/* ----	Class Method: -----------------------
		Find Panel for Menu Item
*/
+(NXPanel*)panelForMenuItem:(id)item
{
	for (uint i=0; i<panels; i++)
	{
		NXPanel* p = panel[i];
		if( item == p->menu_item ) return p;
	}
	return nil;
}


- (void)dealloc 
{
	if(menu_item) [menu_item release];
	[super dealloc];
}


-(id)initWithContentRect:(NSRect)box styleMask:(uint)style backing:(NSBackingStoreType)backing defer:(BOOL)flag
{
	XLogIn("NXPanel:initWithContentRect");	
	self = [super initWithContentRect:box styleMask:style backing:backing defer:flag];
	if(self) 
	{
		menu_item = nil;
		edit_view = nil;
		do_update = yes;

		XXASSERT(panels<NELEM(panel));
		panel[panels++] = self;
	
		if (!update_timer) update_timer = 
		[NSTimer scheduledTimerWithTimeInterval:0.05/*sec*/ target:[NXPanel class] 
				selector:@selector(updateAllPanels) userInfo:(id)NULL repeats:yes];

		//[self setLevel:NSFloatingWindowLevel];
		//[self setBecomesKeyOnlyIfNeeded:yes];
		//[self setDelegate:self];
	}
	return self;
}


/* ----	Assign existing menu item for panel --------------------
		If possible simply connect 'menu_item' in Interface Builder
*/
-(void)setWindowMenuItem:(NSMenuItem*)item
{
	if(menu_item) [menu_item release];
	menu_item = item;
	[item retain];
	[item setAction:@selector(orderFront:)];
	[item setTarget:self];
}

-(void)createWindowMenuItem
{
	if(!menu_item) [self setWindowMenuItem:[window_menu createMenuItemForPanel:self]];
}

-(void)removeWindowMenuItem
{
	if(menu_item) [menu_item release];
	menu_item = nil;
}


/* Convenience Methods:
*/
-(void)showPanel { [self orderFront:self]; }
-(void)hidePanel { [self orderOut:self]; }


/* Support for Edit Views:
	The currently edited view is hilighted yellow
	and stored in 'edit_view' so that 
	the sub class'es 'updatePanel' can easily skip it.
*/
- (void)controlTextDidBeginEditing:(NSNotification*)note
{
	XLogIn("NXPanel:controlTextDidBeginEditing");
	edit_view = [note object];
	[edit_view setBackgroundColor:[NSColor yellowColor]];
}

- (void)controlTextDidEndEditing:(NSNotification*)note
{
	XLogIn("NXPanel:controlTextDidEndEditing");
	[edit_view setBackgroundColor:[NSColor whiteColor]];
	edit_view = nil;
}


/*	Enable/Disable all direct subviews of the content view
*/
-(void)enableSubviews:(bool)yesno
{
	XLogIn("NXPanel:enableSubviews:%s",onoffstr(yesno));

	NSArray* subviews = [[self contentView] subviews];
	for(uint i=0;i<[subviews count];i++)
	{ 
		id subview = [subviews objectAtIndex:i];
		if([subview respondsToSelector:@selector(setEnabled:)])
		{
			[subview setEnabled:yesno];
		}
	}
}


/* Forward all keys to the keyboard panel:
*/
-(void)keyDown:(NSEvent*)event	{ XXLogIn("NXPanel:keyDown"); [kbd_panel keyDown:event]; }
-(void)keyUp:(NSEvent*)event	{ XXLogIn("NXPanel:keyUp"); [kbd_panel keyUp:event]; }
-(void)flagsChanged:(NSEvent*)e	{ XXLogIn("NXPanel:flagsChanged"); [kbd_panel flagsChanged:e]; }

//- (void)windowWillClose:(NSNotification *)aNotification	{ XLogIn("NXPanel:windowWillClose"); }
//- (void)awakeFromNib	{ XLogIn("NXPanel:awakeFromNib"); [self setDelegate:self]; }

@end







