/*	Copyright  (c)	Günter Woigk 2007 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#define SAFE	3
#define LOG		1

#import "DisplayList.h"


/* ===============================================================
				List of all physical displays
=============================================================== */


DisplayList display_list;

DisplayList::DisplayList()
{
	XLogIn("new DisplayList");

	CGDisplayErr err = CGGetActiveDisplayList( MAX_DISPLAYS, display_ids, &num_displays );
	if( err == CGDisplayNoErr ) 
	{
		for( uint i=0; i<num_displays; i++ ) { displays[i] = new Display(i,display_ids[i]); }
	}
	else
	{
		LogLine( "new DisplayList: error = %i.", (int)err );
		LogLine( "new DisplayList: Fullscreen not available!" );
		num_displays = 0;
	}
}



/* ===============================================================
		Utilities ro read values from a CFDictionaryRef
=============================================================== */
 
int get_int( CFDictionaryRef values, CFStringRef key )
{
	CFNumberRef number_value = CFNumberRef( CFDictionaryGetValue(values,key) );
	if( !number_value ) return -1;		// -1 = error: key does not exist in dictionary
	long result;
	if( CFNumberGetValue(number_value, kCFNumberIntType, &result) ) return result;
	return -1;							// -1 = error: result cannot be returned in requested data type
}

inline 
double get_double( CFDictionaryRef values, CFStringRef key )
{
	CFNumberRef number_value = CFNumberRef( CFDictionaryGetValue(values,key) );
	if( !number_value ) return -1;
	double result;
	if( CFNumberGetValue(number_value, kCFNumberDoubleType, &result) ) return result;
	return -1;
}

float get_float( CFDictionaryRef values, CFStringRef key )
{
	CFNumberRef number_value = CFNumberRef( CFDictionaryGetValue(values,key) );
	if( !number_value ) return -1;
	float result;
	if( CFNumberGetValue(number_value, kCFNumberFloatType, &result) ) return result;
	return -1;
}

static 
void dump_display_mode_values( CFDictionaryRef values )
{
	Log( " Display Mode: %i:",	   get_int(values,kCGDisplayMode) );
	Log( " bounds = %i x %i,",	   get_int(values,kCGDisplayWidth),        get_int  (values,kCGDisplayHeight) );
	Log( " bpp = %i, hz = %.2f\n", get_int(values,kCGDisplayBitsPerPixel), get_float(values,kCGDisplayRefreshRate) );
}



/* ===============================================================
	class Display: representatives for each physical display
=============================================================== */


Display::Display ( uint idx, CGDirectDisplayID disp )		
{
	myDisplayIndex		= idx;
	myDisplay			= disp;
	myDisplayModeList	= CGDisplayAvailableModes( disp );		
	myDisplayModeCount	= CFArrayGetCount( myDisplayModeList );
	myOpenGLDisplayID	= CGDisplayIDToOpenGLDisplayMask( disp );

	if(XXLOG) DumpAllDisplayModes(); 	
}


// set w,h,bpp,hz for display
// return TRUE for success
bool Display::SetDisplayMode( float width, float height, int bpp, float hz )
{
	XLogIn( "Display:SetDisplayMode: %.0f x %.0f with %i bpp at %.2f hz", width, height, bpp, hz );

	CFDictionaryRef mode_values = CGDisplayBestModeForParametersAndRefreshRate( 
			myDisplay, bpp, size_t(width), size_t(height), hz, NULL );
	
	if(XLOG) dump_display_mode_values( mode_values );

	CGDisplayErr err = CGDisplaySwitchToMode( myDisplay, mode_values );
	if(err) ShowAlert( "Setting display mode returned error: %i\n",(int)err );
	return err == CGDisplayNoErr;
}


float Display::GetRefreshRate( )
{
	CFDictionaryRef my_mode = GetCurrentDisplayMode();
	return get_float( my_mode, kCGDisplayRefreshRate );	
}


bool Display::IsTV( )
{
	CFDictionaryRef my_mode = GetCurrentDisplayMode();
	int result = get_int( my_mode, kCGDisplayModeIsTelevisionOutput );	
	return result == 1;		// not 0 (NO) and not -1 (UNSET)
}


// dump w,h,bpp,hz of current display mode of display
void Display::DumpCurrentDisplayMode( )
{
	CFDictionaryRef mode_values = CGDisplayCurrentMode(myDisplay);
	dump_display_mode_values( mode_values ); 
}


// dump w,h,bpp,hz of all possible display modes of display[idx]
void Display::DumpAllDisplayModes( )
{
	LogIn( "Display %i (OpenGL mask: %x): %i display modes", 
		 (int)myDisplayIndex, 
		 (int)myOpenGLDisplayID, 
		 (int)myDisplayModeCount );
	
	for( uint i=0; i<myDisplayModeCount; i++ )
	{
		dump_display_mode_values( (CFDictionaryRef)CFArrayGetValueAtIndex(myDisplayModeList,i) );
	}
}


// capture display for exclusive direct use 
// return TRUE for success
bool Display::CaptureDisplay(  )
{
	if( CGDisplayIsCaptured(myDisplay) ) 
	{ XLogLine("CaptureDisplay: display is already captured"); return false; }
	CGDisplayErr err = CGDisplayCapture(myDisplay);
	if(err) XLogLine("CaptureDisplay: error = %i",(int)err);
	return err == CGDisplayNoErr;
} 


// release exclusive use for display
// return TRUE for success
bool Display::ReleaseDisplay( )
{
	if( !CGDisplayIsCaptured(myDisplay) ) 
	{ XLogLine("ReleaseDisplay: display is not captured"); return false; }
	CGDisplayErr err = CGDisplayRelease(myDisplay);
	if(err) XLogLine("ReleaseDisplay: error = %i",(int)err);
	return err == CGDisplayNoErr;
}



