/*	Copyright  (c)	Günter Woigk 2002 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

@class Screen: 
	Maintain a Specci screen in an OpenGLView
	
Notes on locks and threads:
	See file "Locks and Semaphores.txt"
*/


#define	SAFE	3
#define	LOG		1

#import <OpenGL/OpenGL.h>
#import <OpenGL/gl.h>
#import <OpenGL/glu.h>
#import <OpenGL/glext.h>
#import <OpenGL/CGLRenderers.h>

#import "unix/tempmem.h"
#import "utilities.h"
#import "Screen.h"
#import "Machine.h"
#import "MenuController.h"
#import "Z80.h"
#import "NXTextField.h"
#import "KbdPanel.h"
#import "DisplayList.h"
#import "Preferences.h"
#import "KempstonMouseInterfacePanel.h"
#import "Mouse.h"
#import "GifEncoder.h"
#import "ZxPixelmap.h"
#import "Colormap.h"
 


// Colors:

float const h = 0.8;
static float i2r[16]={0,0,h,h,0,0,h,h,0,0,1,1,0,0,1,1};
static float i2g[16]={0,0,0,0,h,h,h,h,0,0,0,0,1,1,1,1};
static float i2b[16]={0,h,0,h,0,h,0,h,0,1,0,1,0,1,0,1};


// One black pixel:

static uchar my_black_pixel = 0;


/*	search for flashing attributes in screen
	screen must be normalized,
	else flashing attributes with fore == back color also return 'yes'
*/
inline
bool
has_flashing( ushort* screen )
{
	ushort* p = screen + 192*32;			// 6 kB
	while( --p >= screen && *(short*)p>=0 ) {} 
	return p>=screen;
}



/*	Normalize screen:

	-1- Clear flashing bit and all pixels if foreground == background color.
	-2- Invert pixels of flashing attr_pixels in flashphase 1

	After normalizing the flashphase info is no longer required for rendering the pixels.
	Normalizing leads to different attr_pixel values for flashing pixels when the flashphase toggles:
	They are automatically detected as 'dirty' when comparing a new screen with the previous sceen.

	Note: in a forced redraw in updateThread the lower portion of the screen is taken from the 
	previous already normalized screen with inverted flashing pixels in flashphase 1.
	Ignoring this results in flashing of these pixels with every forced screen update.
	To fix this, call normalize for the old screen again to double toggle the old pixels 
	before merging them with the new screen.

	The flashing bits must be preserved for making flashing screenshots of haltet machines.
	
	Other methods:
	
	- Clear flashing bits in one flashphase and preserve it in the other:
	Then a forced screen update can set the ula_flashphase to the bit-preserving flashphase.
	Won't work: the old screen for the first forced redraw comes from a normal screen update.
	
	- Don't handle flashing here:
	The flashphase info is no longer contained in the screen_pixel values and must be passed 
	to many functions separately. Flashing screen_pixel values do not change when the flashphase 
	changes and must be marked as dirty separately. 
*/
static
void
normalize_screen( ushort* screen, bool flashphase )
{
	ushort* p = screen + 192*32;				
	
	if( flashphase )	// flashphase 1 => invert flashing pixels
	{	
		while( --p >= screen )
		{
			ushort ap = *p;						// attr_pixel value
			if( ((ap^(ap>>3))&0x0700) == 0 )	// foreground color == background color?
			{ *p = ap & 0x7f00; }				// => clear flashing bit and pixels
			else if( (short)ap<0 )  			// flashing?
			{ *p = ap ^ 0x00ff; }				// => invert pixels
		}
	}
	else				// flashphase 0 => only check for fore==back color
	{	
		while( --p >= screen )
		{
			ushort ap = *p;						// attr + pixel byte
			if( ((ap^(ap>>3))&0x0700) == 0 )	// foreground color == background color?
			{ *p = ap & 0x7f00; }				// => clear flashing bit and pixels
		}
	}
	
	
}


/*	for GIF image:
	Renumber colors in image to reduce color table size:
	overwrites pmap[] and cmap[] with new matching indexes/colors.
	clears transp_color if no transparent pixels are found.
*/
void 
ReduceColors ( Pixelmap& pmap, Colormap& cmap )
{
	XXLogIn  ( "ReduceColors(Pixelmap,Colormap)" );
	XXLogLine( "old colors: %i", (int)cmap.UsedColors() );
	if( XXLOG && cmap.HasTranspColor() ) LogLine( "old transp: %i",(int)cmap.TranspColor() );
	XXLogLine( "cmap: %s",cmap.CmapStr() );
	
	Colormap my_cmap;				// new cmap
	uchar my_conv[256]; memset(my_conv,0xff,256);	// old color index -> new color index

	uptr p  = pmap.Pixels();		// running pointer
	int  w  = pmap.W();
	int  h  = pmap.H();
	int  dy = pmap.DY(); XXXTRAP(dy<w);
	int  n  = 0;					// new cmap size

	for( int y=0; y<h; y++, p+=dy )
	for( int x=0; x<w; x++ )
	{
		int c = p[x];				// c = old color index
		int k = my_conv[c];			// k = new color index
		if( k>=n )
		{
			XXASSERT(c<cmap.UsedColors());
			my_conv[c] = k = n++;
			my_cmap.AddColor(cmap[c]);

			XXLogLine( "  pixel[x=%i,y=%i] = %i -> %i, rgb=(%hhu,%hhu,%hhu)", 
				(int)(pmap.X1()+x), (int)(pmap.Y1()+y), (int)c,(int)k, cmap[c][0],cmap[c][1],cmap[c][2] );
		}
		p[x] = k;
	}

	XXXASSERT( n <= cmap.UsedColors() );
	XXXASSERT( n == my_cmap.UsedColors() );

	if( cmap.HasTranspColor() ) 
	{
		int my_transp = my_conv[cmap.TranspColor()];
		if( my_transp<n )
		{
			my_cmap.SetTranspColor( my_transp );
			XXLogLine( "new transp: %i",(int)my_cmap.TranspColor() );
		}
		else
		{
			XXLogLine( "no transparent pixels!" );
		}
	}

	cmap = my_cmap;
	XXLogLine( "new colors: %i", (int)cmap.UsedColors() );
	XXLogLine( "cmap: %s",cmap.CmapStr() );
}




/* ==== IMPLEMENTATION ==============================
*/
@implementation Screen


-(bool)isFullscreen				{ return fullscreen; }
-(bool)isNotDirty				{ return !new_data_valid; }
-(bool)isDirty					{ return new_data_valid; }
-(bool)drawsInnerScreen			{ return draw_inner_screen; }

-(void)updateAll				{ update_all = yes; }
-(void)savePreferencesNow		{ if(!fullscreen) [self saveFrameUsingName:@"main_window"]; }
-(DisplayRef)getFSDisplayRef	{ return fs_display; }

-(void)update					// Replacement handler
{								// a window is sent an update event every pass through the event loop 
	XXLogIn("Screen:update");	// e.g. with every key press
								// in fullscreen we get an update with every mouse move
//	[super update];	// sends NSWindowDidUpdateNotification message to the default notification center
					// sends windowDidUpdate message to this screen
} 

// Replacement handlers: forward all keys to the keyboard panel:
-(BOOL)acceptsFirstResponder		{ return YES; /* accept key events */ }		
-(void)keyDown:(NSEvent*)event		{ XXLogIn("Screen:keyDown");		[kbd_panel keyDown:event]; }
-(void)keyUp:(NSEvent*)event		{ XXLogIn("Screen:keyUp");			[kbd_panel keyUp:event]; }
-(void)flagsChanged:(NSEvent*)event	{ XXLogIn("Screen:flagsChanged");	[kbd_panel flagsChanged:event]; }


// Notifications
-(void)windowDidResignKey:(NSNotification*)n	{ XLogIn("Screen:windowDidResignKey");  }
-(void)windowDidResignMain:(NSNotification*)n	{ XLogIn("Screen:windowDidResignMain"); }

-(void)windowDidBecomeKey:(NSNotification*)n	
{
	XLogIn("Screen:windowDidBecomeKey");  
	machine->MoveToFrontInMList(); 
}

-(void)windowDidBecomeMain:(NSNotification*)n	
{
	XLogIn("Screen:windowDidBecomeMain"); 
	machine->MoveToFrontInMList();
	[kbd_panel allKeysUp]; 
}

//	- (void) windowWillMiniaturize:(NSNotification*)n	{ XXLogIn("Screen:windowWillMiniaturize"); }
//	- (void) windowDidMiniaturize:(NSNotification*)n	{ XXLogIn("Screen:windowDidMiniaturize"); }
//	- (void) windowDidDeminiaturize:(NSNotification*)n	{ XXLogIn("Screen:windowDidDeminiaturize"); }
	- (void) windowDidExpose:(NSNotification*)n			{ XXLogIn("Screen:windowDidExpose"); update_all=true; }
//	- (void) windowDidChangeScreen:(NSNotification*)n	{ XXLogIn("Screen:windowDidChangeScreen"); }
//	- (void) windowWillMove:(NSNotification*)n			{ XXLogIn("Screen:windowWillMove"); }
//	- (NSRect)windowWillUseStandardFrame:(NSWindow*)sender defaultFrame:(NSRect)defaultFrame
//	- (BOOL) windowShouldZoom:(NSWindow*)sender toFrame:(NSRect)newFrame { return yes; }	  // allow zooming
//	- (BOOL) windowShouldClose:(id)sender								 { return yes; }	  // allow closing
//	- (NSSize)windowWillResize:(NSWindow*)sender toSize:(NSSize)newFrame { return newFrame; } // allow resizing
//	- (void) windowDidUpdate:(NSNotification*)n			{ XXLogIn("Screen:windowDidUpdate"); }
//	- (void) windowDidMove:(NSNotification*)n			{ XXLogIn("Screen:windowDidMove"); }
	- (void) windowDidResize:(NSNotification*)n			{ XXLogIn("Screen:windowDidResize"); update_all=true;}

- (void)windowWillClose:(NSNotification*)n			
{
	XLogIn("Screen:windowWillClose");	
	delete machine;			// ~machine() will subsequently closeAndDispose this window
}


-(int)width
{
	return fullscreen ? fs_display->GetDisplayWidth()
					  : (int)[[self contentView] frame].size.width;
}

-(int)height
{
	return fullscreen ? fs_display->GetDisplayHeight()
					  : (int)[[self contentView] frame].size.height;
}

-(NSSize)size
{
	return fullscreen ? NSMakeSize([self width],[self height])
					  : [[self contentView] frame].size;
}

-(int)getZoom
{
	int zoom = Min( [self width]/256, [self height]/192 ); Limit(1,zoom,4);
	return zoom;
}


-(void)setZoom:(int)factor
{
	if(fullscreen) [self setFullscreen:off];
	
// I believe that it is not required to lock this with 'lock'		2006-12-02 kio
// it is probably locked by 'lockFocusIfCanDraw'.

	NSRect box = [self frame]; 
	box = [self contentRectForFrameRect:box];
	NSSize newsize = { (32+2*4)*8*factor, (24+2*3)*8*factor };

	box.origin.x = int( box.origin.x + (box.size.width  - newsize.width)  /2   );
	box.origin.y = int( box.origin.y + (box.size.height - newsize.height) *4/4 );

	box.size = newsize;
	box = [self frameRectForContentRect:box];
	[self setFrame:box display:yes];

	update_all = yes;
}


-(void)setDrawInnerScreen:(bool)onoff
{
	if( onoff!=draw_inner_screen )
	{
		draw_inner_screen = onoff;
		update_all  = yes;
	}
}


//	handle openGL error: when switching back from fullscreen there is often an error. 
-(void)checkOpenGLError
{
	GLint err = glGetError(); 
	if( err!=GL_NO_ERROR ) 
	{
		XLogLine( "OpenGL error: $%04x", (int)err ); 
		WaitDelay( 0.5 );		// give Mac OSX time to do whatever is blocking
		update_all = true;		// drawing failed => redraw all required
	} 
}


/* ================================================================================
		Gif file handling
		save a screenshot or record movie
   ================================================================================
*/

-(void)saveGif:(cstr)path mode:(int)new_state
{
	PLocker z(lock);

	if( gif.state!=gif_none ) 
	{
		if( gif.state==gif_recording ) gif.state = gif_close_file;
	}
	else
	{
		if(!gif.encoder) gif.encoder = new GifEncoder;
		gif.encoder->OpenFile( path );
		gif.state = new_state;
	}
}


-(OSStatus)recordMovie:(cstr)path	
{
	XLogIn("Screen:recordMovie");
	[self saveGif:path mode:gif_record_movie];
	return ok;
}


-(OSStatus)saveScreenshot:(cstr)path	
{
	XLogIn("Screen:saveScreenshot");
	[self saveGif:path mode:gif_screenshot];
	return ok;
}


-(void)saveGifFrame
{
	XXLogIn("Screen:saveGifFrame");

	switch( gif.state )
	{
	case gif_screenshot: /*	Save screenshot as Gif image.
							-	border incl. border effects
							-	attr+pixels as displayed (not just a screen$ representation)
							-	animated gif for flashing pixels	
							note: Gif viewers, esp. open source and freeware stuff like ImageMagick or Firefox,
							are totally broken in respect to rendering and animating gifs,
							so that we must stick to the bare basics as common denominator. */
	{
		XXXASSERT( gif.encoder && gif.encoder->ImageInProgress() );
	// Calculate pixelmaps:
		normalize_screen( new_attr_pixel,1 );				// toggle flashing pixels
		ZxPixelmap screen1( new_attr_pixel, new_border, ula_top_lines, ula_row_bytes, gif_h_border, gif_v_border );
		normalize_screen( new_attr_pixel,1 );				// toggle flashing pixels
		ZxPixelmap screen2( new_attr_pixel ); 
		screen2.SetToDiff( screen1, zx_transp_color );

	// Calculate Colormaps and total number of colors in image:
		Colormap cmap1 = zx_cmap; ReduceColors(screen1,cmap1); //XXLogLine("cmap1: %s",cmap1.CmapStr());
		Colormap cmap2 = zx_cmap; ReduceColors(screen2,cmap2); //XXLogLine("cmap2: %s",cmap2.CmapStr());
		int total_colors = cmap2.UsedColors();
		for( int i=0; i<cmap1.UsedColors(); i++ ) { total_colors += cmap2.FindColor(cmap1[i])==Colormap::not_found; }

	// Write to file:
		gif.encoder->WriteScreenDescriptor( screen1.W(), screen1.H(), total_colors );
		screen1.OffsetAddresses(gif_h_border,gif_v_border);
		screen2.OffsetAddresses(gif_h_border,gif_v_border);

		if( screen2.IsNotEmpty() )		// not empty? => flashing
		{
			gif.encoder->WriteLoopingAnimationExtension();
			gif.encoder->WriteGraphicControlBlock( 32/*delay*/ );
			gif.encoder->WriteImage(screen1,cmap1);
			gif.encoder->WriteGraphicControlBlock( 32/*delay*/, cmap2.TranspColor() );
			gif.encoder->WriteImage(screen2,cmap2);
		}
		else
		{
			gif.encoder->WriteImage(screen1,cmap1);
		}

		gif.encoder->CloseFile();
		gif.state = gif_none;
		break;
	}
	case gif_record_movie:	/* start movie recording: */
	{
		XXXASSERT( gif.encoder && gif.encoder->ImageInProgress() );
		XXXASSERT( !gif.pixelmap && !gif.othermap );
		gif.pixelmap = new ZxPixelmap( new_attr_pixel, new_border, ula_top_lines, ula_row_bytes, gif_h_border, gif_v_border );
		gif.othermap = new ZxPixelmap(); 	// nur anlegen		
		gif.frame_count = 0;
		gif.encoder->WriteScreenDescriptor( gif.pixelmap->W(), gif.pixelmap->H(), zx_cmap );
		gif.encoder->WriteLoopingAnimationExtension();
		gif.state = gif_recording;
		gif.commandkey_up = no;
		break;
	}
	case gif_recording:		// record next frame:
	{
		if( ~modifier_keys & NSCommandKeyMask ) gif.commandkey_up = yes;	// cmd key up?
		else if( gif.commandkey_up ) gif.state = gif_close_file;		// cmd key down again?
	}
	case gif_close_file:	// record last frame and close file
	{
		XXXASSERT( gif.encoder && gif.encoder->ImageInProgress() );
		XXXASSERT( gif.pixelmap && gif.othermap );	// image pending to write & scratch

		int n=0; for(int i=0;i<32*192;i++) n += new_attr_pixel[i] != old_attr_pixel[i];
		if(n) 
		XXLogLine("%i attr_pixel values changed.", n);

		gif.frame_count++;
		gif.othermap->SetToScreenDiff( new_attr_pixel, old_attr_pixel );
		
		if( gif.state==gif_close_file || gif.othermap->IsNotEmpty() )
		{
			Colormap cmap = zx_cmap;
			ReduceColors( *gif.pixelmap, cmap );
			gif.encoder->WriteGraphicControlBlock( gif.frame_count*2, cmap.TranspColor() );
			gif.pixelmap->OffsetAddresses(gif_h_border,gif_v_border);
			gif.encoder->WriteImage( *gif.pixelmap,cmap );
			gif.pixelmap->OffsetAddresses(-gif_h_border,-gif_v_border);

			if( gif.state==gif_close_file )
			{
				XLogLine("Screen:saveGifFrame: closing file");
				gif.encoder->CloseFile();		// => finish recording
				delete gif.pixelmap; gif.pixelmap = NULL;
				delete gif.othermap; gif.othermap = NULL;
				gif.state = gif_none;
			}
			else
			{
				gif.frame_count = 0;
				Swap( gif.pixelmap, gif.othermap );
			}
		}
		break;
	}
	default: IERR();
	}		// switch
}



/* ================================================================================
		Do the actual work
		draw frame for ZX Spectrum style ULA/CRTC
   ================================================================================
*/
-(void)drawFrame: (int)w: (int)h
{
	XXXLogIn("Screen:drawFrame");

// draw screen

	int zoom  = Min( w/256, h/192 ); Limit(1,zoom,4);		// screen zoom
	int hzoom = zoom*8;										// new_border hor. zoom
	int x0    = (w-256*zoom) / 2;
	int y0    = (h-192*zoom) / 2 / zoom * zoom;

	//int ula_top_lines	= machine->ula->GetLinesBeforeScreen();
	//int ula_bot_lines	= machine->ula->GetLinesAfterScreen();
	//int ula_row_bytes	= machine->ula->GetUlaCyclesPerLine()/16;
	int ula_side_bytes	= ula_row_bytes-32;

	int scr_top_lines	= y0/zoom;						if (scr_top_lines>ula_top_lines*2/3)	scr_top_lines = ula_top_lines*2/3;
	int scr_bot_lines	= (h-y0-zoom*192+zoom-1)/zoom;	if (scr_bot_lines>ula_bot_lines*2/3)	scr_bot_lines = ula_bot_lines*2/3;
	int scr_left_bytes  = x0/hzoom;						if (scr_left_bytes>ula_side_bytes/3)	scr_left_bytes = ula_side_bytes/3;
	int scr_right_bytes = (w-x0-zoom*256+hzoom-1)/hzoom;if (scr_right_bytes>ula_side_bytes/3)	scr_right_bytes = ula_side_bytes/3;											
// update all after geometry change:

	if (update_all)
	{
		update_all = false;
		if(XXLOG)
		{
			LogLine("Screen:drawFrame:update_all:");
			LogLine("     w,  h  = %i, %i",w,h);
			LogLine("     x0, y0 = %i, %i",x0,y0);
			LogLine("     Zoom   = %i",zoom);
			LogLine("Ula: top/bottom lines = %i,%i",ula_top_lines, ula_bot_lines);
			LogLine("     row/side   bytes = %i,%i",ula_row_bytes, ula_side_bytes);
			LogLine("Scr: top/bottom lines = %i,%i",scr_top_lines, scr_bot_lines);
			LogLine("     left/right bytes = %i,%i",scr_left_bytes, scr_right_bytes);
		}

	// update viewport  ((imageable area))
		glViewport(0, 0, w, h);

	// setup indexed color conversion maps
		glPixelMapfv(GL_PIXEL_MAP_I_TO_R,16,i2r);
		glPixelMapfv(GL_PIXEL_MAP_I_TO_G,16,i2g);
		glPixelMapfv(GL_PIXEL_MAP_I_TO_B,16,i2b);

	// setup projection matrix
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glOrtho( 0,w, h,0, -1,1 );

	// setup new_pixels unpacking, transfer, mapping & rasterization
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glPixelStorei(GL_UNPACK_ALIGNMENT,1);

	// paint it black
	// displayed == new_border+new_pixels+new_attr == screen
		memset(old_border,0,ula_row_bytes*(ula_top_lines+192+ula_bot_lines));
		memset(old_attr_pixel,0,(192*32)*2);		// set all attr to black+black
		memset(pixels,0,sizeof(pixels));
		glPixelZoom( w, -h );
		glRasterPos2i( 0, 0 );
		glDrawPixels( 1, 1, GL_COLOR_INDEX,GL_UNSIGNED_BYTE, &my_black_pixel );
	}

// draw border
	{
	// calculate dirty range

		int bda = ula_top_lines-scr_top_lines;
		int bde = ula_top_lines+192+scr_bot_lines;
		
		while( bda<bde )
		{
			long* o = (long*)(old_border+ula_row_bytes*bda-(scr_left_bytes+1));
			long* n = (long*)(new_border+ula_row_bytes*bda-(scr_left_bytes+1));
			long* e = o + (scr_left_bytes+1+32+scr_right_bytes+3) /4;
			
			while( o<e && *n==*o ) { n++;o++; };
			if(o<e) break; else bda++;
		}

		while( bda<bde )
		{
			long* o = (long*)(old_border+ula_row_bytes*(bde-1)-(scr_left_bytes+1));
			long* n = (long*)(new_border+ula_row_bytes*(bde-1)-(scr_left_bytes+1));
			long* e = o + (scr_left_bytes+1+32+scr_right_bytes+3) /4;
			
			while( o<e && *n==*o ) { n++;o++; };
			if(o<e) break; else bde--;
		}
		
	// draw border

		if( bda<bde )			// anything to draw?
		{
		// fractional part of leftmost new_border byte
			if( x0 > scr_left_bytes*hzoom && x0 < ula_side_bytes/3*hzoom )	
			{
				glPixelZoom( x0-scr_left_bytes*hzoom, -zoom );
				glPixelStorei(GL_UNPACK_ROW_LENGTH,ula_row_bytes);	

				glRasterPos2i( 0, y0+(bda-ula_top_lines)*zoom );
				glDrawPixels( 1, bde-bda, GL_COLOR_INDEX,GL_UNSIGNED_BYTE, new_border + ula_row_bytes*bda -scr_left_bytes-1 );
			}

			glPixelZoom(hzoom,-zoom);
			glPixelStorei(GL_UNPACK_ROW_LENGTH,ula_row_bytes);

		// inner screen drawing enabled => only draw outer border
			if( draw_inner_screen )	
			{
				if(bda<ula_top_lines)
				{		
					glRasterPos2i( x0-scr_left_bytes*hzoom, y0+(bda-ula_top_lines)*zoom );
					glDrawPixels( scr_left_bytes+32+scr_right_bytes, ula_top_lines-bda, GL_COLOR_INDEX,GL_UNSIGNED_BYTE, 
								  new_border + ula_row_bytes*bda -scr_left_bytes);
					bda = ula_top_lines;
				}

				if( bde>ula_top_lines+192 )
				{		
					glRasterPos2i( x0-scr_left_bytes*hzoom, y0+192*zoom );
					glDrawPixels( scr_left_bytes+32+scr_right_bytes,bde-(ula_top_lines+192), GL_COLOR_INDEX,GL_UNSIGNED_BYTE,
						new_border + ula_row_bytes * (ula_top_lines+192)-scr_left_bytes);
					bde = ula_top_lines+192;
				}

				if( bde>bda )
				{
					if( scr_left_bytes )
					{		
						glRasterPos2i( x0-scr_left_bytes*hzoom, y0+(bda-ula_top_lines)*zoom );
						glDrawPixels( scr_left_bytes,bde-bda, GL_COLOR_INDEX,GL_UNSIGNED_BYTE, new_border + ula_row_bytes*bda -scr_left_bytes );
					}
					if( scr_right_bytes )
					{		
						glRasterPos2i( x0+32*hzoom, y0+(bda-ula_top_lines)*zoom );
						glDrawPixels( scr_right_bytes, bde-bda, GL_COLOR_INDEX,GL_UNSIGNED_BYTE, new_border + ula_row_bytes*bda +32 );
					}
				}		
			}
			
		// inner screen drawing is disabled for 'fast mode' or debugging:
			else		
			{
				glRasterPos2i( x0-scr_left_bytes*hzoom, y0+(bda-ula_top_lines)*zoom );
				glDrawPixels ( scr_left_bytes+32+scr_right_bytes, bde-bda, GL_COLOR_INDEX,GL_UNSIGNED_BYTE, new_border + ula_row_bytes*bda -scr_left_bytes);
			}
		}
	}


// draw screen 

	if( draw_inner_screen )
	{
	// normalize new attr+pixels:
		normalize_screen( new_attr_pixel, ula_flashphase );

	// update pixels[] and calculate dirty[]:
		ulong dirty[192];
		ushort const* oap = old_attr_pixel;
		ushort const* nap = new_attr_pixel;
		uptr z = pixels;								// running destination pointer in pixels[]
		
		for( int y=0; y<192; y++ )
		{
			ulong d=0;			

			for( int x=0; x<32; x++, nap++, oap++ )
			{
				ushort ap = *nap;						// new attr + pixel byte
				if( ap==*oap ) { z+=8; continue; }		// no change

			// changed:
				d |= (1<<x);							// mark dirty
			//	uchar f = a2f[ short(ap)>>8];			// foreground color for pixels==1
			//	uchar b = a2b[ushort(ap)>>8];			// background color for pixels==0

				uchar f = a2f(ap);			// foreground color for pixels==1
				uchar b = a2b(ap);			// background color for pixels==0

			//	int fi = short(ap)>>8;
			//	int bi = ushort(ap)>>8;
			//	uchar f = a2f[fi];			// foreground color for pixels==1
			//	uchar b = a2b[bi];			// background color for pixels==0

				for( ushort m=0x0080; m; m>>=1 )		// update pixels
				{
					*z++ = ap&m ? f : b; 
				}
			}
			dirty[y] = d;
		}
		
		XXTRAP( z   != pixels         + 256*192 );
		XXTRAP( oap != old_attr_pixel +  32*192 );
		XXTRAP( nap != new_attr_pixel +  32*192 );
		
	// draw pixels[]
		glPixelZoom(zoom,-zoom);
		glPixelStorei(GL_UNPACK_ROW_LENGTH,256);
		int ya,ye=0;
		while(ye<192)
		{	
			ulong d=0;
			ya = ye;    while(ya<192&&!dirty[ya]) { ya++; } if (ya>=192) break;
			ye = ya; r: while(ye<192&& dirty[ye]) { d|=dirty[ye]; ye++; }
			if(ye<191&&dirty[ye+1]) { ye++; goto r; }	// melt across single unchanged row

			d=~d; 
			int xa=0;	while(d&(1<<xa)) xa++; xa = xa*8;	
			int xe=31;	while(d&(1<<xe)) xe--; xe = xe*8 +8;	
			
			glRasterPos2i(x0+xa*zoom,y0+ya*zoom);
			glDrawPixels(xe-xa,ye-ya,GL_COLOR_INDEX,GL_UNSIGNED_BYTE,pixels+ya*256+xa);
		}
	}
}






/* -----------------------------------------------------------------
					The Worker Thread
   -------------------------------------------------------------- */


/* ---- transfer data for next frame ----
		normally only called from audio interrupt
		but may also be called from main event loop
		data supply is mutual exclusive by the 'lock'
*/
-(void)nextFrame
{
	XXXLogIn("Screen:nextFrame");	

	if( new_data_valid ) { gif.frame_count++; return; }	// old frame still in work

	lock->Lock();
		Ula* ula = machine->GetUla(); XXXTRAP(!ula);
		Swap( new_attr_pixel, ula->GetAttrPixel() );
		Swap( new_border,     ula->GetBorder()    );
		ula_top_lines		= ula->GetLinesBeforeScreen();
		ula_bot_lines		= ula->GetLinesAfterScreen();
		ula_row_bytes		= ula->GetOctetsPerLine(); 
		ula_flashphase		= ula->GetFlashPhase();
		new_data_valid		= yes;
	lock->Unlock();

	sema->Unlock();
}


-(void)updateThread:(id)dummy
{
	XLogLine("Screen:updateThread: starting");		// nicht LogIn() wg. Thread  ((indent would persisit))

	NSAutoreleasePool* pool = [[NSAutoreleasePool alloc] init];

	Machine* old_fm	= nil;				// front machine
	int width=0, height=0;				// width and height of drawing port
	
	{	XXXASSERT(gl_context);
		GLint f = [preferences getBoolForKey:"fullscreen_mode_synchronize_with_vbl"];
		CGLSetParameter( (CGLContextObj)[gl_context CGLContextObj]/*CGLGetCurrentContext()*/, kCGLCPSwapInterval, &f );
	}
	
// ---- SCREEN REFRESH LOOP ----

	update_all = yes;						// first time run: draw all pixels
aa:	Swap( old_attr_pixel, new_attr_pixel );	// swap buffers if they were drawn
	Swap( old_border,     new_border     );
a:	new_data_valid = no;					// => accept new frame 
	PurgeTempMem();

// adjust thread priority
	if( front_machine()!=old_fm )
	{
		old_fm = front_machine();
		[NSThread setThreadPriority:old_fm==machine?0.5:0.3];
	}

// wait for work
	sema->TryLock(0.1);		// timeout: 0.1 sec.
	XXXASSERT(machine);
	if(termi) goto x;		// => dispose screen & terminate self
	if( !fullscreen && ![self isVisible] ) { XXLogLine("Screen:updateThread: not visible"); goto a; }

	if( !new_data_valid )	// timeout?
	{
		XXLogLine("Screen:updateThread: timeout"); 

	// forced redraw with video beam indicator:
		PLocker z(lock);	// protect access to 'Screen.machine' and 'new_*' buffers
		if(termi) goto x;	// cleared by Machine.~Machine() via Screen.closeAndDispose() right now! => exit
							// else: machine can't be deleted because Screen.closeAndDispose() will block on 'lock'
		if(!machine->TryLock()) { XLogLine("Screen:updateThread: cannot lock"); goto a; }	// protect access to machine data 
							// note: Machine.~Machine() must not block 'run_lock'!
							// only TryLock, because machine.lock is 'outer' compared to screen lock!
							// if failed, just don't force update now and hope for better luck after 0.1 sec.

		// now: machine is locked. 
		// we have both, machine.lock and screen.lock, though aquired in wrong sequence

		if( !new_data_valid )	// not supplied in a race condition right now?
		{
			Ula* ula = machine->GetUla(); XXXASSERT(ula);
			long cc = machine->GetCurrentCpuCycle();
			ula->UpdateBorderUpToCycle(cc);
			ula->UpdateScreenUpToCycle(cc);
			long n = ula->GetCurrentScreenIndex(); Limit(0L,n,192*32L);
			memcpy( new_attr_pixel,   ula->GetAttrPixel(), sizeof(*new_attr_pixel) * n          ); 
			normalize_screen(old_attr_pixel,ula_flashphase);	// double toggle toggled pixels
			memcpy( new_attr_pixel+n, old_attr_pixel+n,    sizeof(*new_attr_pixel) * (192*32-n) );
			long m = ula->GetCurrentBorderIndex(); Limit(0L,m,ula->GetOctetsPerFrame());
			memcpy( new_border,	  ula->GetBorder(), sizeof(*new_border) * m ); 
			memcpy( new_border+m, old_border+m,     sizeof(*new_border) * (ula->GetOctetsPerFrame()-m) );
			//ula_top_lines	 = ula->GetLinesBeforeScreen();		in init() sinnvoll vorbesetzt, falls noch nie nextFrame() 
			//ula_bot_lines	 = ula->GetLinesAfterScreen();		""
			//ula_row_bytes	 = ula->GetUlaCyclesPerLine()/16;	""
			//ula_flashphase = ula->GetFlashPhase();  // upper and lower halfes may have different flash phases ((ignored))
			ula->AddVideoBeamIndicator( new_border, new_attr_pixel ); 
			new_data_valid = true;
			//if(machine==front_machine()) add_frame_hit(1);	TODO oder ignorieren
		}
		
		machine->Unlock();
	}
	
// prepare, draw, release:	
	{
		NSOpenGLView* content_view = [self contentView];
		XXXASSERT( content_view );	
		if(![content_view lockFocusIfCanDraw]) { LogLine("Screen:updateThread: cannot draw"); goto a; /* skip frame */ }

		// set the context to which openGL calls are directed
			XXXASSERT(gl_context);
			[gl_context makeCurrentContext];	

		// update NSOpenGLView after view moved or resized:
			if(update_all) 
			{
				[gl_context update];
				width  = [self width];
				height = [self height];
			}

		// **DOIT!**
			[self drawFrame:width:height];	
			[self checkOpenGLError];		// nach stopFullscreen() gibt's häufig einen Fehler!
			glFlush();						// force actual execution of all buffered commands
			//glFinish();					// also blocks until done
			[self checkOpenGLError];		
			//[gl_context flushBuffer];		// if NSOpenGLPixelFormat contains NSOpenGLPFADoubleBuffer 

		[content_view unlockFocus];
	}

// screenshot, record movie:
	if( gif.state!=gif_none ) [self saveGifFrame];

// Loop
	goto aa;

// Exit:
x:	[pool release];
	XLogLine("Screen:updateThread: terminated");	
	termi = no;		// flag: we are done.
}


-(void)startUpdateThread
{
	termi = no;
	[NSThread detachNewThreadSelector:@selector(updateThread:) toTarget:self withObject:nil];
}

-(void)stopUpdateThread
{
	termi = yes;	// inhibit forced redraw in updateThread after waiting for sema timeout. 
	sema->Unlock();	// => updateThread: kill Screen & terminate self
	while(termi) WaitDelay(0.01);
}


-(void)startWindow
{
	XLogIn("Screen:startWindow");	

	NSOpenGLPixelFormatAttribute attr[] = 
	{
		NSOpenGLPFANoRecovery,										//	
	//	NSOpenGLPFADoubleBuffer,									//	needs [gl_ctx flushBuffer] not glFlush()
	//	NSOpenGLPFAAccelerated,										//	may fail! 
		NSOpenGLPFAColorSize,	(NSOpenGLPixelFormatAttribute) 8,	//	bitsPerPixel: kio:32, Apple:24	
	//	NSOpenGLPFADepthSize,	(NSOpenGLPixelFormatAttribute) 0,	//	Tiefenpuffer: kio:32, Apple: 8
	//	NSOpenGLPFAAlphaSize,   (NSOpenGLPixelFormatAttribute) 0,	//	Apple: 8
	//	NSOpenGLPFAStencilSize, (NSOpenGLPixelFormatAttribute) 0,	//	Apple: 0
	//	NSOpenGLPFAAccumSize,   (NSOpenGLPixelFormatAttribute) 0,	//	Apple: 0
		NSOpenGLPFAWindow,											//	Apple
		(NSOpenGLPixelFormatAttribute) 0
	};

	NSOpenGLPixelFormat* fmt = [[[NSOpenGLPixelFormat alloc] initWithAttributes:attr] autorelease];	
	ASSERT(fmt);
	NSOpenGLView* cv = [[NSOpenGLView alloc] initWithFrame:[self frame] pixelFormat:fmt];
	XXXASSERT(cv); 
	[self setContentView:[cv autorelease]];
	gl_context = [cv openGLContext];
	XXXASSERT(gl_context);

//	machine->HaltCpu();
	[self makeKeyAndOrderFront:self];	// -> callback windowDidBecomeKey -> MoveToFrontInMList -> lock mlock hangs
//	machine->ResumeCpu();				// => statt dessen m->HaltCpu() das nutzt nix weil run_for_sound trotzdem erst mal mlock lockt
	[self startUpdateThread];
}


-(void)stopWindow
{
	[self stopUpdateThread];
	[self setContentView:[[[NSView alloc] initWithFrame:[self frame]] autorelease]];
	gl_context = NULL;
}


/*	Start Fullscreen Mode
	return:	fullscreen = true		ok
			fullscreen = false		failed
*/
-(void)startFullscreen
{
	XLogIn("Screen:startFullscreen");	

	fullscreen = [preferences getBoolForKey:"fullscreen_mode_enabled"];
	if(!fullscreen) return;			// generally disabled

	XXTRAP( fullscreen_mode );		// another instance runs in fullscreen mode!  
									// this cannot happen: no access to menu in non-front mmachines

// get my display:
	fullscreen = display_list.Count() >= 1;
	if(!fullscreen) return;
	uint fs_display_idx = [preferences getIntForKey:"fullscreen_mode_display_id"];	
	if( fs_display_idx >= display_list.Count() ) { fs_display_idx = 0/*main screen*/; }
	fs_display = display_list[fs_display_idx];

// Capture display:
	fullscreen = fs_display->CaptureDisplay();
	if(!fullscreen) return;

//	save window frame for restoring after fullscreen:
	if(fs_old_frame) [fs_old_frame release];
	fs_old_frame = [[self stringWithSavedFrame] retain];

// Set display mode
	uint width			= [preferences getFullscreenModeWidth];
	uint height			= [preferences getFullscreenModeHeight];
	uint colorbits		= [preferences getFullscreenModeColorDepth];
	uint refreshrate	= (int)[preferences getFullscreenModeRefreshRate];

	fullscreen = fs_display->SetDisplayMode( width, height, colorbits, refreshrate );
	if(!fullscreen) return;
	fullscreen_mode = yes;		// global flag for New/Open/Save et.al.

// Create  this "pixel format attribute" thing for openGL:
// this must be done _after_ setting display mode!
	long ogl_display_id = fs_display->GetOpenGLDisplayID();

	NSOpenGLPixelFormatAttribute attr[] =
	{
		NSOpenGLPFANoRecovery,
	//	NSOpenGLPFADoubleBuffer,									//	=> needs [gl_context flushBuffer] not glFlush()
		NSOpenGLPFAColorSize,	(NSOpenGLPixelFormatAttribute) 8,	//	bitsPerPixel: kio:32, Apple:24	
	//	NSOpenGLPFAAlphaSize,   (NSOpenGLPixelFormatAttribute) 0,	//	Apple: 8
	//	NSOpenGLPFADepthSize,	(NSOpenGLPixelFormatAttribute) 0,	//	Apple: 8
	//	NSOpenGLPFAStencilSize, (NSOpenGLPixelFormatAttribute) 0,	//	Apple: 0
	//	NSOpenGLPFAAccumSize,   (NSOpenGLPixelFormatAttribute) 0,	//	Apple: 0
		NSOpenGLPFAFullScreen,
		NSOpenGLPFAScreenMask,  (NSOpenGLPixelFormatAttribute) ogl_display_id,
	//	NSOpenGLPFAAccelerated,										//	may fail!
		(NSOpenGLPixelFormatAttribute) 0
	};

// create this "pixel format" for openGL:
	NSOpenGLPixelFormat* fmt = [ [[NSOpenGLPixelFormat alloc] initWithAttributes:attr] autorelease ];	
	fullscreen = fmt!=NULL;
	if(!fullscreen) return;

// create drawing context for openGL:
// this is independent from my window. It could be created and used without any window!
	gl_context = [ [NSOpenGLContext alloc] initWithFormat:fmt shareContext:nil ];
	fullscreen = gl_context!=NULL;
	if(!fullscreen) return;
	[gl_context setFullScreen];

// move window behind fullscreen:
//	this is needed for OS X 10.3.9, because in 10.3.9 you click "through" the fullscreen window
//	and hit whatever is below the mouse pointer, evtl. moving the app into the background.
//	the current window frame was stored above in NSString* fs_old_frame
	CGRect box = fs_display->GetDisplayBounds();
	[self setFrame:*(NSRect*)&box display:NO];
	XLogLine("Display bounds: x=%i, y=%i, w=%i, h=%i",(int)box.origin.x,(int)box.origin.y,(int)box.size.width,(int)box.size.height);

// if only 1 display hide mouse
// and make shure mouse is not over menu or window title bar for clicking in 10.3.9:
	if( machine->kempston_mouse_interface ) { Mouse::GrabMouse(); }
	if( display_list.Count()==1 ) { Mouse::HideMouse(); }

// and finally start update thread:
	[self startUpdateThread];
}


-(void)stopFullscreen
{
	[self stopUpdateThread];
	fullscreen_mode = fullscreen = no;
	[gl_context release]; 
	gl_context = NULL;
	DisplayList::RestoreDisplayModes();
	fs_display->ReleaseDisplay(); 
	if(fs_old_frame) [self setFrameFromString:fs_old_frame];
	if( Mouse::MouseIsHidden() ) Mouse::ShowMouse(); 
	Mouse::ReleaseMouse();
}


/*	Start or Stop Fullscreen Mode
	Note: this function is called with RLOCK / RUNLOCK only 
		  from openSnapshotWithDialog and saveSnapshotWithDialog
*/
-(void)setFullscreen:(bool)f	
{
	XLogIn("Screen:setFullscreen: %s",onoffstr(f));
	if( f==fullscreen ) return;

	if( f ) 
	{
		[self stopWindow]; 
		[self startFullscreen]; 
		if(!fullscreen) [self startWindow]; 
	}
	else
	{ 
		[self stopFullscreen]; 
		[self startWindow]; 
	}
}



/* ----	replacement handler ----
		deinitialize & dealloc
*/
-(void)dealloc
{
	XLogIn("Screen:dealloc");
	delete gif.encoder; 
	delete gif.pixelmap; 
	delete gif.othermap;
	delete lock;
	delete sema;
	delete[] new_attr_pixel; 
	delete[] new_border; 
	delete[] old_border;
	delete[] old_attr_pixel;	
	if(fs_old_frame) [fs_old_frame release];
	[super dealloc];
}


/* ---- dispose screen and terminate updateThread --------------
		called from Machine:~Machine().
*/
-(void)closeAndDispose
{
	XLogIn("Screen:closeAndDispose");	

	if(fullscreen) [self stopFullscreen]; else [self stopWindow];
	[self autorelease];		// wg.  windowWillClose -> delete machine -> closeAndDispose 
}


/*	Designated Initializer
	Setup for windowed mode and start worker thread
*/
-(id) initWithMachine:(Machine*)m
{
	XLogIn("Screen:initWithMachine");

// init window:
	int w = 2*8*(32+4+4), x = 100;
	int h = 2*8*(24+3+3), y = 75;
	NSRect box = {{x,y},{w,h}};						// {x,y,w,h} gets overwritten by preferences

	uint const style = NSTitledWindowMask | NSClosableWindowMask | NSMiniaturizableWindowMask | NSResizableWindowMask;
//	uint const style = NSBorderlessWindowMask;		// Apple: borderless used for both, FS and Win
				
	NSBackingStoreType								// Apple: NSBackingStoreBuffered used for both, FS and Win
	backingType	= NSBackingStoreBuffered;			// Kio: sonst kein Alphakanal (-> Schatten, runde Ecken)

	self = [super initWithContentRect:box styleMask:style backing:backingType defer:NO];
	if(!self) return self;
	
	[self setReleasedWhenClosed:NO];
	[self setContentMinSize:NSMakeSize(256.0,192.0)];
	[self setResizeIncrements:NSMakeSize(2.0,2.0)];
	[self setFrameUsingName:@"main_window"];
	[self setDelegate:self];

// init locks:
	lock = new PLock();
	sema = new PSema("Screen.sema"); 

// init data members:
	machine				= m;
	draw_inner_screen	= yes;		// also draw inner screen (and not only border)
	fullscreen			= no;
	fs_display			= NULL;
	gl_context			= NULL;
	fs_old_frame		= NULL;
	
	gif.state			= gif_none;
	gif.encoder			= NULL;
	gif.pixelmap		= NULL;
	gif.othermap		= NULL;

	new_attr_pixel		= new ushort[192*32];
	old_attr_pixel		= new ushort[192*32];
	new_border			= new uchar[MAX_ULA_CLOCKS_PER_LINE/16*(MAX_LINES_BEFORE_SCREEN+192+MAX_LINES_AFTER_SCREEN+1)];
	old_border			= new uchar[MAX_ULA_CLOCKS_PER_LINE/16*(MAX_LINES_BEFORE_SCREEN+192+MAX_LINES_AFTER_SCREEN+1)];
	new_data_valid		= no;
	sema->Lock();					// no new_data_valid

// sinnvoll vorbesetzen, sonst krachts wenn 1. forced redraw vor erstem regulären nextFrame()
	ula_top_lines	= 64;			// ula->GetLinesBeforeScreen();
	ula_bot_lines	= 56;			// ula->GetLinesAfterScreen();
	ula_row_bytes	= 56;			// ula->GetUlaCyclesPerLine()/16;
	ula_flashphase	= 0;

// start worker thread:
	[self startWindow];
	return self;
}


@end






