/*	Copyright  (c)	Günter Woigk 2007 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	Create GIF files from ZX Spectrum screens
*/


#define	SAFE	3
#define	LOG		1

#import "ZxPixelmap.h"
#import "globals.h"



/* ====================================================================
   global ZX Spectrum color table:
		8 x normal brightness
		8 x bright
*/
cComp F = 255;			// "bright": full brightness
cComp h = 255*4/5;		// "normal": reduced brightness
static 
Comp zx_colors[] = 
{
	0,0,0, 0,0,h, h,0,0, h,0,h, 0,h,0, 0,h,h, h,h,0, h,h,h,		// r,g,b
	0,0,0, 0,0,F, F,0,0, F,0,F, 0,F,0, 0,F,F, F,F,0, F,F,F 
};
cColormap zx_cmap( zx_colors,16,zx_transp_color );



/* ====================================================================
	conversion tables:
	attribute byte -> pixel color
	
	accessing the tables with bit 7 (flashing bit) set:

		foreground color = a2f[ (signed char)attr_byte ]
		background color = a2b[ (unsigned char)attr_byte ]
*/
#if 0	// das wird nicht in dieser Reihenfolge angelegt!
  uchar	a2b[128];		// attr&0x7F -> background color
  uchar	a2f[128];		// attr&0x7F -> foreground color
#else
  uchar	a2c[256];
#endif

inline 
void InitColortable()	
{
	XLogIn("InitColortable");
	assert(zx_transp_color==8);				// use 'bright black' for transparency

	for ( int n=0; n<128; n++ ) 
	{
		char b = (n>>3)&0x0f;				// background pixel color
		a2c[n] = b==8 ? 0 : b;				// used 'black' instead of 'bright black'
		char f = (n&0x07)+((n>>3)&0x08);	// foreground pixel color
		a2c[n+128] = f==8 ? 0 : f;			// used 'black' instead of 'bright black'
	}
}
ON_INIT( InitColortable() );




// #####################################################################################


void
ZxPixelmap::SetToScreen ( ushort screen[] )
{
	XXASSERT( dy >= 256 );

	int x = Max( 0,   X1()/8 );				// copy limits into local variables
	int y = Max( 0,   Y1() );				// scale x values from pixel coord to attr_pixel index
	int w = Min( 32, (X2()+7)/8 ) - x;		// and extend p1.x and p2.x to cover full attr_pixels
	int h = Min( 192, Y2() ) - y;			// and shrink to screen$ pixels rect if frame includes border
	int o = dy - w*8;

	ushort* q = screen + x + y*32;			// running source pointer in screen[]
	uchar*  z = PixelPtr(x*8,y);			// running destination pointer in pixels[]

	for( uptr e=z+h*dy; z<e; q+=32, z+=o ) 
	for( int x=0; x<w; x++ )
	{
		ushort ap = q[x];					// attr<<8 + pixels
		uchar f = a2f(ap);					// foreground color for pixels==1
		uchar b = a2b(ap);					// background color for pixels==0
		for( ushort m=0x80; m; m>>=1 )
		{
			*z++ = ap&m ? f : b; 
		}
	}
}



/*	Init pixmap with ZX Spectrum screen
	full screen, no border, x/y are set to 0/0
*/
ZxPixelmap::ZxPixelmap ( ushort screen[] )
:
	Pixelmap( 0,0,256,192 )					// allocate full-size pixel memory
{		
	SetToScreen( screen );
}



/*	Init pixmap with ZX Spectrum screen
	partial screen, no border
	x/y/w/h are set acc. to bbox
*/
ZxPixelmap::ZxPixelmap ( ushort screen[], cBox& bbox )
:
	Pixelmap( 0,0,256,192 )					// allocate full-size pixel memory
{	
	box = bbox;								// and shrink frame as requested
	
//	XXLIMIT ( X2(), 256 );					// requested frame must not exceed allocated pixel memory
//	XXLIMIT ( Y2(), 192 );
//	XXLIMIT ( X1(), X2() );
//	XXLIMIT ( Y1(), Y2() );

	SetToScreen( screen );
}



/*	Init pixmap with ZX Spectrum screen and border
	x/y are set to 0/0
	w/h are set to screen_size + 2*border
	default border is 32/24 pixels
	ula_top_lines is needed in conjunction with v_border to find first row for pixmap in border[]
	ula_row_bytes is needed to step from row to row in border[]
*/
ZxPixelmap::ZxPixelmap ( ushort screen[], uchar border[], int ula_top_lines, int ula_row_bytes, int h_border, int v_border )
:
	Pixelmap(-h_border,-v_border,256+2*h_border,192+2*v_border)		// allocate extra-large
{	
	XXASSERT( h_border%8==0 );
	XXLIMIT ( h_border,48 );
	XXLIMIT ( v_border,36 );
	
// Border:
	uptr z = data;	// pixel[-h_border,-v_border]
	uptr q = border + ula_top_lines*ula_row_bytes - v_border*ula_row_bytes - h_border/8;
	int  w = W();
	int  h = H();

	for( int y=0; y<h; y++ )
	{
		if( q<border ) { memset(z,0/*black*/,w); z+=dy; q+=ula_row_bytes; continue; }	// should rarely ever be required
		for( int x=0; x<h_border; x+=8 ) { XXXTRAP(*q>7); memset(z,*q,8); z+=8; q+=1; }	// left border
		if( uint(y-v_border) < 192u ) { q += 32; z += 256; } else						// skip center border if inside screen$
		for( int x=0; x<256; x+=8 )      { XXXTRAP(*q>7); memset(z,*q,8); z+=8; q+=1; }	// draw center boder if above/below screen$
		for( int x=0; x<h_border; x+=8 ) { XXXTRAP(*q>7); memset(z,*q,8); z+=8; q+=1; }	// right border
		q += ula_row_bytes - w/8;														// skip horizontal flyback
		//z += dy-w;																	// skip to next pixel row
	}

// Screen:
	SetToScreen( screen );
}



/*	Set my pixels to contain the subframe of new_attr_pixel
	which has changed since old_attr_pixels.
	inside the update box unchanged pixels are set to zx_transp_color.
	old and new screen$s should be normalized prior to call.
	my pixelmap must be large enough and positioned to be resized to {0,0,256,192}.
*/
void ZxPixelmap::SetToScreenDiff ( ushort nju[], ushort old[] )
{
	XXLogIn("ZxPixelmap:SetToScreenDiff");
	XXASSERT(dy>=256&&dy<=320);
	XXASSERT(old!=nju);
	XXASSERT(PixelPtr(0,0)>=data);
	XXASSERT(PixelPtr(255,191)<data+(192+24)*dy);

	int n=0; for(int i=0;i<32*192;i++) n += nju[i] != old[i];
	if(n) 
	XXLogLine("%i attr_pixel values changed!", n);

// find first and last row:

	ushort o = old[0]; old[0] = ~nju[0];	// stopper
	int a = 0, e = 192*32-1;
	while( old[e]==nju[e] ) e--;			// find end  ((inner limit))
	old[0] = o;								// undo stopper
	if( e==0 &&								// stopped by stopper?
		old[0]==nju[0] ) { SetSize(0,0); return; }	// first pixel identical too? => set to empty pixelmap
	while( old[a]==nju[a] ) a++;			// find start

	int y1 = a/32*32;						// start of first dirty row 
	int y2 = e/32*32+31;					// end of last dirty row

// find first and last column:	

	for(;;y1++)
	{
		for( a=y1; a<=y2 && old[a]==nju[a]; a+=32 ) {}
		if( a<=y2 ) break;
	}

	for(;;y2--)
	{
		for( e=y2; e>=y1 && old[e]==nju[e]; e-=32 ) {}
		if( e>=y1 ) break;
	}

	int x1 = y1 % 32;	y1 /= 32;	
	int x2 = y2 % 32;	y2 /= 32;
	
	SetFrame( Loc(x1*8,y1), Loc(x2*8+8,y2+1) );
	XXLogLine("Bbox of changes: p1=%i,%i; sz=%i,%i",x1*8,y1,x2*8,y2);

// paint pixels:

	uptr z = PixelPtr(x1*8,y1);					// running destination pointer in pixels[]
	int dz = dy - (x2-x1+1)*8;
	ushort* np = nju + y1*32;
	ushort* op = old + y1*32;

	for( int y=y1; y<=y2; y++, z+=dz, np+=32, op+=32 )
	{		
		memset( z, zx_transp_color, dy-dz );

		for( int x=x1; x<=x2; x++ )
		{
			ushort ap = np[x];						// new attr + pixel byte
			if( ap==op[x] )	{ z+=8; continue; }		// no change?

		// attr_pixel changed:
			uchar f = a2f(ap);					// foreground color for pixels==1
			uchar b = a2b(ap);					// background color for pixels==0

			for( ushort m=0x0080; m; m>>=1 )	// update pixels
			{
				*z++ = ap&m ? f : b; 
			}
		}
	}
}



#if 0
/*	Init pixmap with flashing region in ZX Spectrum screen.
	non-flashing pixels are set to zx_transp_color
	flashing pixels are set to their foreground/background color counterpart
	note: the attr+pixel pairs should be normalized prior to working with them:
		  normalize: keep flashing attribute except if foreground color == background color.
					 optionally invert flashing pixels for the 'inverting' flashphase.
*/
void ZxPixelmap::InitWithFlashing ( ushort attr_pixel[] )
{
	XLogIn("ZxPixelmap::ZxInitWithFlashing");

	if(XXXSAFE) 
	{
	// assert that transparent color index is unused:
		for(int i=0;i<256;i++) { ASSERT(a2b[i]!=zx_transp_color); }	// checks a2b[] for i<128 and a2f[] for i>=128
	// assert that flashing pixels exist
		int i; for(i=0;i<32*192;i++) { if( attr_pixel[i]&0x8000 ) break; }
		ASSERT( i < 32*192 );
	}

	normalize_screen(attr_pixel,1);

// calc affected box:
	int xmin,xmax,ymin,ymax;
	ushort* p = attr_pixel;
	char row[192], col[32]; memset(row,0,sizeof(row)); memset(col,0,sizeof(col));
	for( int y=0;y<192;y++ ) for( int x=0;x<32;x++ ) { if( *p++ & 0x8000 ) { row[y]=col[x]=1; } }
	
	for( xmin=0;  col[xmin]==0; xmin++ ) {} // must terminate because at least one attr byte flashes!
	for( ymin=0;  row[ymin]==0; ymin++ ) {}
	for( xmax=31; col[xmax]==0; xmax-- ) {} 
	for( ymax=191;row[ymax]==0; ymax-- ) {}
	XXLogLine("flashing rect: [%i..x..%i] x [%i..y..%i]", xmin*8,xmax*8+7, ymin,ymax);


	this->x = xmin<<3;
	this->y = ymin;
	this->w = (xmax-xmin+1)<<3; dy = w;
	this->h = (ymax-ymin+1);

	delete[] data; data = pixels = new uchar[w*h];
	memset( pixels, transp_color, w*h );

	ushort* q = attr_pixel + ymin*32;			// running source pointer in screen[]
	uchar*  z = pixels;							// running destination pointer in pixels[]
		
	for( uptr e=z+w*h; z<e; q+=32 ) 
	for( int x=xmin; x<=xmax; x++ )
	{
		uint ap = q[x];							// attr<<8 + pixels

		if( (short)ap<0 )	// flashing?
		{
			uchar f = a2f(ap);					// foreground color for pixels==0
			uchar b = a2b(ap);					// background color for pixels==1
			for( uint m=0x80; m; m>>=1 )
			{
				*z++ = ap&m ? b : f; 
			}
		}
		else				// not flashing
		{
			z += 8;
		}
	}
}									 
#endif

