/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	buffer<uchar>		rles Daten
	samples_per_second		Samplerate für diesen Block
 
	Der Puffer enthält abwechselnd eine Lauflängenangabe für die High-Phase und die Low-Phase.
	Der Puffer startet mit der Low-Phase, ggf. mit Lauflänge '0'.
	Mehrere Phasen gleicher Auslenkung können aneinandergehängt werden, 
	wenn man für das gegenteilige Sample dazwischen als Lauflänge 0 einträgt.

	pilot_start			Sample-Position des Beginns des Nutzsignals (idR. Pilot-Ton)
	data_start			Sample-Position des Beginns der Nutzdaten (idR. nach dem Sync-Bit)
	data_end			Sample-Position des Endes   des Nutzsignals

tzx:	

	der erste Puls ist "low"
	der erste Puls eines Standard-Tap-Blocks ist "low"
	nach einem Puls wird der Level für den nächsten Puls invertiert. (logisch)
	nach einem Direct Recording Block (ID $15) wird der Level für den nächsten Puls *nicht* invertiert.
	eine Pause von 0 ms nach einem Block wird komplett ignoriert.
	eine Pause > 0 ms bedeutet einen Puls von 1 ms mit dem aktuellen (invertierten) Level 
						und danach den Rest der Zeit mit Level "low"
						und der Level für den nächsten Puls wird *nicht* invertiert (bleibt "low").
					<=>	Puls mit aktuellem (invertiertem) Level und angegebener Länge 
						und Abfallautomatik bei der Wiedergabe
						und Level für den nächsten Puls auf "low" setzen.

Play() == Read():

	buffer.Pos()		nächstes zu lesendes rles-Byte
	total_samples		insgesamt im Buffer gespeicherte Samples
	time_akku			noch zu spielende Zeiten "last_sample" bis wieder ein Byte gelesen werden muss.
	last_sample			zuletzt gespieltes Sample. Für Signalabschwächung gegen 0 in Pausen.

	aktuelle Phase	  = buffer.Pos() & 1  ?  "low"  :  "high" 		
	Gesamtspielzeit   =	total_samples/samples_per_second

Record() = Write():

	buffer.Pos()		Position für nächstes zu schreibendes rles-Byte
	time_akku			Summierer für Zeiten "high" oder "low" bis wieder ein Byte geschrieben werden muss.
	last_sample			zuletzt gelesenes Sample
	total_samples		invalid

	aktuelle Phase    = buffer.Pos() & 1  ?  "high"  :  "low" 	

Notes:
	
	Beim Speichern von time_akku ändert sich die Laufzeit wegen der Sampleraten-Anpassung. 
	Damit ändert sich in dem Moment die exakte aktuelle Position und Gesamtspiellänge.
	Nach Schreiben eines n-Sample-Puffers mit Write() erhöht sich die aktuelle Position deshalb nicht exakt um n Samples!
*/


#define SAFE	3
#define	LOG 	2

#import	<fcntl.h>
#import	<unistd.h>
#import	<math.h>
#import <zlib.h>
#import	"unix/file_utilities.h"
#import	"cstrings/cstrings.h"
#import	"RlesData.h"
#import	"DSP.h"
#import "AlertPanel.h"


/*	settings:
*/
#define	PLAYBACK_VOLUME	0.3		// <-- das geht auch noch durch die PlayThrough-Dämpfung


/* ---- ZX Spectrum tape loader timing ----
		generell: 0-Bits werden mit 2 kHz, 1-Bits mit 1 kHz gespeichert.
		Pilot-Phasen sind etwas länger als 1-Bits,
		das Sync-Bit ist etwas kürzer als ein 0-Bit.
		Signale können beim Lesen invertiert sein.
		Durch schlechte Aufnahmen verschieben sich Taktflanken z.T. gewaltig.
		
Messwerte am emulierten ZX Spectrum +128 @ 44.1 kHz:
		Pilot	   27 Samples
		Sync	8 + 9 Samples
		0-Bit	   11 Samples
		1-Bit	   22 Samples
		
lt. RamSoft "ZX Spectrum Loaders Guide":
		Sync	667+735 cpu cycles  ~ 2623+2381 Hz
		0-Bit		855 cpu cycles  ~ 2047 Hz
		1-Bit	   1710 cpu cycles  ~ 1023 Hz
		Pilot	   2168 cpu cycles  ~  807 Hz
*/
Frequency const	cc_per_second	= 3500000.0;	// cpu cycles per second
ulong  const	cc_per_pilot	= 2168;			// cpu cycles per pilot half phase
ulong  const	cc_per_bit_0	=  855;			// cpu cycles per 0-bit half phase
ulong  const	cc_per_bit_1	= 1710;			// cpu cycles per 1-bit half phase
ulong  const	cc_per_sync1	=  667;			// cpu cycles for 1st sync phase
ulong  const	cc_per_sync2	=  735;			// cpu cycles for 2nd sync phase


/*	Samplefrequenz für RLES:
	Die Phasen des normalen ZX Spectrum Tapeloader-Signal sollen möglichst immer in ein Nibble passen.
	Damit ergibt sich eine max. Samplefrequenz von 25 kHz für computergenerierte Signale.
	Für natürliche Signale wg. der teils massiven Nulldurchgangsverschiebung nur ca. 21 kHz.
	Die Samplefrequenz sollte außerdem ein ganzzahliger Teil der Hardware-Samplerate sein.
*/
inline Frequency BestRlesSampleRate() throw() { return samples_per_second; }


/*	Berechnung der aktuellen Bandposition:
*/
Time RlesData::CurrentPosition() const
{
	return recording ? CurrentSamplePos()/samples_per_second - time_akku
					 : CurrentSamplePos()/samples_per_second + time_akku;
}


/*	Berechnung der Gesamtblocklänge:
	Ann.: Bei Aufnahme wird immer der gesamte Block überschrieben, d.h. vorher gelöscht.
	Damit ist dann die Gesamtlänge immer == aktuelle Bandposition.
*/
Time RlesData::TotalPlaytime() const
{
	return recording ? CurrentSamplePos()/samples_per_second + time_akku
					 : TotalSamples()/samples_per_second;
}


/*	determine 'current phase'  ('high' or 'low')
	this is the phase for the next rles byte to write	(recording)
	or      the phase of  the last rles byte read		(playback)
*/	

#define NEXT_PHASE				(buffer.Pos() & 1)
#define	LAST_PHASE				(!(buffer.Pos() & 1))

#define	CURRENT_PLAYBACK_PHASE	LAST_PHASE
#define CURRENT_RECORD_PHASE	NEXT_PHASE

#define	NEXT_PHASE_IS_HIGH		NEXT_PHASE
#define	NEXT_PHASE_IS_LOW		LAST_PHASE



/*	return whether this DataType can import/export certain file formats:
	static functions.
*/
bool RlesData::CanRead ( FileType ft ) 
{
	switch(ft)
	{
	case zx80_file:
	case o_file:
	case zx81_file:
	case p_file:
	case tap_file:
	case rles_file:
	case pzx_file:
	case tzx_file:	return yes;
	default:		return no;
	}
}

bool RlesData::CanWrite ( FileType ft ) 
{
	switch(ft)
	{
	case tap_file:
	case rles_file:	return yes;
	default:		return no;
	}
}



/*	initialization
*/
void RlesData::init ()
{
	buffer.Purge();
	samples_per_second	 = BestRlesSampleRate();
	time_akku		 = 0.0;
	last_sample		 = 0.0;
	data_start		 = 0;
	data_end		 = 0;
}


/*	initialization from other instance
*/
void RlesData::init ( TapeData const& q )
{
	switch (q.GetDataType())
	{
	default:		IERR();
	case bytes:		TODO();		// RlesData::init(ByteData)
	case shorts:	TODO();		// RlesData::init(ShortData)
	
	case rles:
		{
			RlesData const& Q = (RlesData const&)q;
			buffer			  = Q.buffer;
			samples_per_second	  = Q.samples_per_second;
			time_akku		  = Q.time_akku;
			last_sample		  = Q.last_sample;
			data_start		  = Q.data_start;	
			data_end		  = Q.data_end;
		}	break;
	}
}


/*	destruction
*/
void RlesData::kill()
{
}


/*	clear all data
*/	
void RlesData::Purge ()
{
	kill(); 
	TapeData::Purge();
	init();
}


/*	store rles bytes into buffer 
*/
inline void RlesData::store_byte ( uchar c ) throw(bad_alloc)
{
	buffer.Write(c);
}

inline void RlesData::store_bytes ( uchar c, ulong n ) throw(bad_alloc)
{
	buffer.Write(c,n); 
}

inline void RlesData::store_bytes ( uptr  bu, ulong n ) throw(bad_alloc)
{
	buffer.Write(bu,n); 
}

inline void RlesData::store_pair  ( uchar c0, uchar c1 ) throw(bad_alloc)
{
	buffer.Write(c0); 
	buffer.Write(c1); 
}

inline void RlesData::store_pairs ( uchar c0, uchar c1, ulong n ) throw(bad_alloc)	
{
	while( n-- ) { store_pair(c0,c1); } 
}


/*	store single pulse into buffer
	stores 1+2*n bytes  
	=> always toggles next phase  "high" <-> "low"
*/
inline void RlesData::store_samples( float f ) throw(bad_alloc)	
{
	XXXLogIn( "RlesData:store_samples(%.1f)", f );

	long n = long( f+0.45 );
	XXXASSERT(n>=0);

	while(n>255) { store_pair(255,0); n-=255; }	// most times none
	store_byte(n);								// may be 0
}

inline void RlesData::store_time( Time t ) throw(bad_alloc)	
{
	XXXLogIn("RlesData:store_time(%f)", t );
	store_samples( t * samples_per_second );
}

inline void RlesData::store_cc( ulong n ) throw(bad_alloc)	
{
	XXXLogIn( "RlesData:store_cc(%lu)", n );
	store_samples( n * samples_per_second / cc_per_second );
}


/*	store tzx-style pause:
	if pause == 0.0 do nothing  
	else
	tzx:  store ≥1ms opposite level, padding pause with 'low' level to mimic silence level read on ula port.
	tzx:  don't toggle after pause, next phase level remains 'low'.

	kio:  the silence level depends on model!!
	kio:  => store xlong pulse with opposite level. Play() will reduce level to 0.0. Silence level managed by machine::input().
	kio:     force next phase level to 'low'. (as required by tzx 'pause' spec.)
*/
void RlesData::store_tzx_pause ( Time pause )
{
	XXLogIn( "RlesData:store_tzx_pause( %f sec )", (float)pause );

	if(pause>0.0)								// pause = 0.0  =>  do nothing.	
	{
		data_end = buffer.Pos();
		store_time(pause);						// store pulse with current (==toggled) level
		if(NEXT_PHASE_IS_HIGH) store_byte(0);	// force next pulse level to 'low'. 
	}
}


/*	store symbol from tzx block $19 into buffer:
	symbol = 
	{	byte				initial_phase: 0=toggle, 1=keep, 2=low, 3=high
		short_z[max_cnt]	pulse length (cc), sequences shorter than max_cnt are delimited with 0x0000
	}
*/
void RlesData::store_tzx19_symbol ( uptr p, ulong max_cnt )
{
	XXXLogIn( "RlesData:store_tzx19_symbol" );
	XXXASSERT(max_cnt <= 256 );
	
// initial phase:
	uint initial_phase = *p++;
	switch( initial_phase&3 )		
	{
	case 2: /* set to 0 */
	case 3: /* set to 1 */
		if( NEXT_PHASE == (initial_phase&1) ) break;
	case 1: /* don't toggle, keep signal level, extend pulse */
		store_byte(0);
	case 0: /* toggle, default */
		break;
	}

// pulses:
// delimited by 0x0000; max 256:
	for( uint i=0; i<max_cnt; i++ )
	{
		ushort cc_per_pulse = p[0] + 256* p[1]; p+=2; if( !cc_per_pulse ) break;
		store_cc( cc_per_pulse ); 
	}	
}


/*	Store a pure tone:
	E.g. store pilot tone of a block
	Can also be used to store a single pulse

	note: try to reproduce the overall frequency as exact as possible.
		  some games (Tomahawk) seem to be very sensible to the overall timing.
*/
void RlesData::store_pure_tone( ulong pulses, Time secs_per_pulse )
{
	XXLogIn( "RlesData:store_pure_tone" );

	float samples_per_pulse = secs_per_pulse * samples_per_second;
	XXXASSERT( samples_per_pulse <= 255.0 );

	XXLogLine( "pulses: %u",			pulses );						// num. pulses
	XXLogLine( "cc/pulse: %u",			uint(cc_per_second * secs_per_pulse + 0.5) );	// cc per pulse
	XXLogLine( "samples/pulse: %.1f",	samples_per_pulse );			// samples per pulse
	XXLogLine( "freq.:%.1f Hz",			float(0.5 / secs_per_pulse) );	// freq.

	float rem = 0.0;
	while( pulses-- )
	{
		float fsamples = samples_per_pulse+rem;
		uint  samples  = uint(fsamples+0.5);
		rem = fsamples-samples;
		store_byte(samples);
	}
}


/*	Store pure data:
	Store data of tape block into buffer
*/
void RlesData::store_pure_data
	(	uptr	bu,
		ulong	bits_in_bu,				
		Time	secs_per_bit0,			// duration of one halfphase
		Time	secs_per_bit1	)		// duration of one halfphase
{	
	XXLogIn( "RlesData:store_pure_data" );

	float samples_per_bit0 = secs_per_bit0  * samples_per_second;	XXXASSERT(samples_per_bit0<=255);
	float samples_per_bit1 = secs_per_bit1  * samples_per_second;	XXXASSERT(samples_per_bit1<=255);

	XXLogLine( "bits: %u = %lu bytes + %lu bits", bits_in_bu,bits_in_bu/8l, bits_in_bu%8l );
	XXLogLine( "samples/pulse[0]: %.1f", samples_per_bit0 );
	XXLogLine( "samples/pulse[1]: %.1f", samples_per_bit1 );

	buffer.Grow( buffer.Pos() + bits_in_bu*2 );

	if( !data_start ) 
	{
		data_start = buffer.Pos();
		// TODO: setting block info should be done by caller!
		if( !major_block_info )							// does not work if block data was saved in pieces 
		{ SetBlockInfoFromData( bits_in_bu/8, bu ); }	// and first piece too short for info
	}

	float rem = 0.0, fsamples, samples_per_bit;
	ulong samples;

	for( long bits = bits_in_bu; bits>0; bits-=8 )
	{
		char byte = *bu++;
		int  end  =  8 - Min(8l,bits);
		for( int b=7; b>=end; b-- )
		{						
			samples_per_bit = (byte>>b)&1 ? samples_per_bit1 : samples_per_bit0;

			fsamples = samples_per_bit+rem;
			samples  = ulong(fsamples+0.5);	store_byte(samples);
			rem      = fsamples-samples;

			fsamples = samples_per_bit+rem;
			samples  = ulong(fsamples+0.5);	store_byte(samples);
			rem      = fsamples-samples;
		}
	}
	
	XXXASSERT( buffer.Pos()==buffer.Size() );
}


/*	Store one block of data:
	out:	stored data starts with high signal pase (pilot)
			if pause > 0.0:	stored data ends with high signal phase 
							time_hi = high signal phase for pause
			if pause = 0.0:	stored data ends with low signal phase 
							time_hi,lo flushed to buffer
*/
void RlesData::store_data_block 
	(	uchar*	bu,
		ulong	bits_in_bu,
		Time	seconds_per_pilot,		// total duration of pilot tune
		Time	seconds_per_silence,	// total duration of trailing silence
		Time	secs_per_pilot_pulse,	// duration of one halfphase
		Time	secs_per_sync1_pulse,	// duration of one halfphase
		Time	secs_per_sync2_pulse,	// duration of one halfphase
		Time	secs_per_bit0_pulse,	// duration of one halfphase
		Time	secs_per_bit1_pulse	)	// duration of one halfphase
{
	XXLogIn("RlesData::store_data_block");
	XXLogLine("%i bytes + %i bits",				int(bits_in_bu/8), int(bits_in_bu%8));
	XXLogLine("sample rate: %i",				(int)samples_per_second);
	XXLogLine("pilot: %f samples/pulse, %f sec, %u pulses",	
												float(samples_per_second*secs_per_pilot_pulse), float(seconds_per_pilot), 
												uint( seconds_per_pilot / secs_per_pilot_pulse +0.5 ) );
	XXLogLine("sync1: %f samples/pulse",		float(samples_per_second*secs_per_sync1_pulse) );
	XXLogLine("sync2: %f samples/pulse",		float(samples_per_second*secs_per_sync2_pulse) );
	XXLogLine("bit0:  %f samples/pulse",		float(samples_per_second*secs_per_bit0_pulse) );
	XXLogLine("bit1:  %f samples/pulse",		float(samples_per_second*secs_per_bit1_pulse) );
	XXLogLine("pause: %f samples, %f sec",		float(samples_per_second*seconds_per_silence), (float)seconds_per_silence );

// pilot
	ulong pulses_per_pilot = ulong( seconds_per_pilot  / secs_per_pilot_pulse +0.5 );
	store_pure_tone( pulses_per_pilot, secs_per_pilot_pulse );

// sync
	store_time(secs_per_sync1_pulse);
	store_time(secs_per_sync2_pulse);
	
// data
	store_pure_data( bu, bits_in_bu, secs_per_bit0_pulse, secs_per_bit1_pulse );

// silence
	store_tzx_pause( seconds_per_silence );
}


/*	Store one block of data from .tap file
	standard rom routine timing
*/
void RlesData::store_data_block 
	(	uchar*	bu,
		ulong	bits_in_bu,		// must include at least one trailing bit!
		Time	pause )			// total duration of trailing silence
{
	XXLogIn("RlesData::store_data_block(rom)");

		Time secs_per_pilot	= cc_per_pilot / cc_per_second;		// duration of one halfphase
		Time secs_per_sync1	= cc_per_sync1 / cc_per_second;		// duration of one halfphase
		Time secs_per_sync2	= cc_per_sync2 / cc_per_second;		// duration of one halfphase
		Time secs_per_bit0	= cc_per_bit_0 / cc_per_second;		// duration of one halfphase
		Time secs_per_bit1	= cc_per_bit_1 / cc_per_second;		// duration of one halfphase

		Time pilot = (bu[0]&0x80 ? 3220 : 8064) * secs_per_pilot;	// pilot pulses: 8064 Header, 3220 Data

	store_data_block( 
		bu,
		bits_in_bu,
		pilot,				// total duration of preceeding pilot tune
		pause,				// total duration of trailing silence
		secs_per_pilot,		// duration of one halfphase
		secs_per_sync1,		// duration of one halfphase
		secs_per_sync2,		// duration of one halfphase
		secs_per_bit0,		// duration of one halfphase
		secs_per_bit1 );
}




/*	Record():
	store runlength for time_akku into buffer 
	toggles next phase  "high" <-> "low"		((stores 1+2*n bytes))
  note: the time_akku is used to record sampled sound
		before any function which stores rles bytes directly is used,
		the time_akku must be flushed, which will toggle the current phase.
		Append a 0 byte to toggle back the phase if required.
*/
inline void RlesData::store_akku()
{
	store_time(time_akku);
	time_akku = 0.0;
}


/*	Record():
	add time to time_akku
	if time = 0.0 then nothing is done.					(=> easy merging of split long phases)
	if requested phase does not match current phase,
	then the akku is first stored into the buffer,
	toggling the current phase.
*/
inline void RlesData::append_time ( Time t, bool requested_phase )
{
    if(t>0.0)
    {
        if( CURRENT_RECORD_PHASE != requested_phase ) store_akku();
        time_akku += t;
    }
}


/*	Play():
	read 'time_akku' from buffer
	merges 'split' run length bytes
	setzt last_sample entsprechend Phase
	wechselt deshalb immer zur umgekehrten Phase  <=>  liest  1  oder  1+2*n  Bytes
	außer am Buffer-Ende 
	oder wenn die aktuelle Phase 0 Samples lang wäre  ((sollte nur am Block-Anfang auftreten))
*/
void RlesData::read_akku()
{
	if( buffer.AtEnd() )
	{
		time_akku = 0.0;
	}
	else
	{
		ulong samples; 
		do{ samples = buffer.Read(); } while( 0 && samples==0 && buffer.Avail(1) );

		while( buffer.Avail(2) && buffer.Peek()==0 )
		{
			buffer.Skip(); // Pos()++;
			samples += buffer.Read();
		}

		last_sample = buffer.Pos()&1 ? -PLAYBACK_VOLUME : +PLAYBACK_VOLUME;
		time_akku = samples / samples_per_second;
		XXXLog("%i ",(int)samples);
	}
}


/*	Read (Play) samples from block buffer
	liest mit read_akku() Samples aus dem Buffer
	füllt Zielpuffer mit last_sample
	Schwächt dabei last_sample immer weiter ab. (-> Null-Level je nach Modell in machine::input())
	Samplefrequenz des Zielpuffers: samples_per_second  (core audio sample rate)
	Samples werden zu den Zieldaten addiert! (gemixt)
*/
long RlesData::Play ( Sample* data, long cnt )		
{
	if (recording) { stop_recording(); return 0; }

	Sample* p=data;
	
	for(;;)
	{
		long samples = long( time_akku*samples_per_second +0.5 );
		
		if(cnt<=samples)
		{
			time_akku -= cnt / samples_per_second;	// kann negativ werden!
			while(cnt--) { *p++ += last_sample; last_sample *= 0.998; }	// geht in ca. 2200 Samples auf 1.2% zurück
			return p-data;
		}
		else
		{
			cnt -= samples;
			while(samples--) { *p++ += last_sample; last_sample *= 0.998; }		
			read_akku();
			if( time_akku<=0.0 && buffer.AtEnd() ) return p-data;		// end of buffer
		}
	}
}


/*	Record (Write) to tape:
	TODO: detection of data_start
*/
void RlesData::Record ( Sample const* data, long cnt, Signal_Type signal_type )		
{
	if(!recording) start_recording();

	if( signal_type != data_signal )	// silence or noise
	{
		if( !data_end )					// first silence block?
		{
			if( !data_start ) { XLogLine("RlesData:Record(): data_start=0  !!!"); }
			store_akku();				// store pending time_akku and toggle NEXT_PHASE			
			data_end = buffer.Pos();	
		}

		append_time( cnt / samples_per_second, CURRENT_RECORD_PHASE );
		last_sample = 0.0;
		return;
	}

// data signal:

	Sample const* p = data;
	Sample last = last_sample;
	Time const secs_per_tick = 1.0 / samples_per_second;
	Time akku = time_akku; 

	uint samples_per_pilot		= int( cc_per_pilot / cc_per_second * samples_per_second +0.5 );
	uint samples_per_sync1		= int( cc_per_sync1 / cc_per_second * samples_per_second +0.5 );
	uint samples_per_sync2		= int( cc_per_sync2 / cc_per_second * samples_per_second +0.5 );
	uint samples_per_sync1_max	= samples_per_sync1 + (samples_per_pilot-samples_per_sync1)/4;	// Absolut limit
	uint samples_per_sync12_max	= (samples_per_sync1+samples_per_sync2) *3/2;		// Absolut limit for sync1+sync2

	for(;;)
	{
		if( CURRENT_RECORD_PHASE )	while( cnt && *p>=0.0/*-0.01*/ ) { cnt--; last = *p++; akku += secs_per_tick; }
		else						while( cnt && *p<=0.0/*+0.01*/ ) { cnt--; last = *p++; akku += secs_per_tick; }

		if(cnt==0) break;

		Time z = last / (last-*p) * secs_per_tick;	// exakte Pos. des Nulldurchgangs ermitteln und Sample anteilig zuschlagen
		XXXASSERT(z>=0.0);
		akku += z;
		store_time( akku );
		akku = -z;							// < 0 !!!

		if( !data_start && buffer.Pos()>=256 
			&& buffer[buffer.Pos()-2]<=samples_per_sync1_max 
			&& (uint)buffer[buffer.Pos()-2]+(uint)buffer[buffer.Pos()-1]<=samples_per_sync12_max )
		{ data_start=buffer.Pos(); }
	}

	time_akku   = akku;				// remember pending time
	last_sample = last;				// remember last sample
}


/*	Record (Write) to tape:
	TODO: detection of data_start
*/
void RlesData::Record ( StereoSample const* data, long cnt, Signal_Type signal_type )		
{
	if(!recording) start_recording();

	if( signal_type != data_signal )	// silence or noise
	{
		if( !data_end )					// first silence block?
		{
			if( !data_start ) { XLogLine("RlesData:Record(): data_start=0  !!!"); }
			store_akku();				// store pending time_akku and toggle NEXT_PHASE			
			data_end = buffer.Pos();	
		}

		append_time( cnt / samples_per_second, CURRENT_RECORD_PHASE );
		last_sample = 0.0;
		return;
	}

// data signal:

	StereoSample const* p = data;
	Sample last = last_sample;
	double const secs_per_tick = 1.0 / samples_per_second;
	Time akku = time_akku; 

	uint samples_per_pilot		= int( cc_per_pilot / cc_per_second * samples_per_second +0.5 );
	uint samples_per_sync1		= int( cc_per_sync1 / cc_per_second * samples_per_second +0.5 );
	uint samples_per_sync2		= int( cc_per_sync2 / cc_per_second * samples_per_second +0.5 );
	uint samples_per_sync1_max	= samples_per_sync1 + (samples_per_pilot-samples_per_sync1)/4;	// Absolut limit
	uint samples_per_sync12_max	= (samples_per_sync1+samples_per_sync2) *3/2;		// Absolut limit for sync1+sync2

	for(;;)
	{
		if( CURRENT_RECORD_PHASE )	while( cnt && *p>=0.0/*-0.01*/ ) { cnt--; last = (p++)->Mono(); akku += secs_per_tick; }
		else						while( cnt && *p<=0.0/*+0.01*/ ) { cnt--; last = (p++)->Mono(); akku += secs_per_tick; }

		if(cnt==0) break;

		Time z = last / (last-p->Mono()) * secs_per_tick;	// exakte Pos. des Nulldurchgangs ermitteln und Sample anteilig zuschlagen
		XXXASSERT(z>=0.0);
		akku += z;
		store_time( akku );
		akku = -z;							// < 0 !!!

		if( !data_start && buffer.Pos()>=256 
			&& buffer[buffer.Pos()-2]<=samples_per_sync1_max 
			&& (uint)buffer[buffer.Pos()-2]+(uint)buffer[buffer.Pos()-1]<=samples_per_sync12_max )
		{ data_start=buffer.Pos(); }
	}

	time_akku   = akku;				// remember pending time
	last_sample = last;				// remember last sample
}


/*	stop playing or recording
*/
void RlesData::stop_recording()
{
	XXXASSERT(recording);

	if(time_akku>0.0) store_akku();
	buffer.Resize( buffer.Pos() );
	buffer.ShrinkToFit();
	recording = no;
}

void RlesData::Stop()
{
	if(recording) stop_recording();			// not really required
}


/*	start recording
*/
void RlesData::start_recording()
{
	XXXASSERT(!recording);

	buffer.Purge();
	buffer.Allocate(1 MB);
	//buffer.Rewind();
	samples_per_second	= BestRlesSampleRate();
	data_start		= 0;
	data_end		= 0;
	time_akku		= 0.0;
	last_sample		= 0.0;

	recording		= yes;
	modified		= yes;
	ClearMajorBlockInfo();
}

void RlesData::StartRecording()
{
	// do nothing. recording starts when first samples are written
	// so silence can be skipped without already flushing the current contents.
}

void RlesData::StartPlaying()
{
	if(recording) stop_recording();			// not really required
}



/*	Seek start of block
*/
void RlesData::SeekStart()
{
	if(recording) stop_recording(); 
	
	buffer.Rewind();
	time_akku		= 0.0;
	last_sample		= 0.0;
}


/*	Seek end of block
*/
void RlesData::SeekEnd()
{
	if(recording) stop_recording(); 

	buffer.Seek( buffer.Size() );
	time_akku		= 0.0;
	last_sample		= 0.0;
}


/*	seek tape position
*/
void RlesData::SeekPosition	( Time new_position )
{
	if( new_position<=1e-9 ) { SeekStart(); return; } 
	if( recording ) stop_recording();
	ulong new_sample_pos = ulong( new_position * samples_per_second );
	if( new_sample_pos>=TotalSamples() ) { SeekEnd(); return; }
	time_akku		= 0.0;
	last_sample		= 0.0;

// günstigste Startposition wählen: (current pos, start oder end of buffer)
	if( new_sample_pos<CurrentSamplePos()/2 )				  
	{
		buffer.Rewind(); 
	} 
	else if( new_sample_pos>(CurrentSamplePos()+TotalSamples())/2 ) 
	{
		buffer.Seek( buffer.Size() ); 
	}

// mit 'current_sample_pos' an 'new_sample_pos' ranfahren:
	while( CurrentSamplePos()<new_sample_pos )
	{
		XXXASSERT(buffer.Pos()<buffer.Size());
		buffer.Skip();
	}

	while( CurrentSamplePos()>new_sample_pos )
	{
		XXXASSERT(buffer.Pos()>0);
		buffer.rSkip();
	}

//	buffer.Ptr() zeigt jetzt auf das gesuchte rles byte
//	current_sample_pos entspricht dem 1. sample im rles byte
//	new_sample_pos entspricht dem sample im rles byte, in dem die zielzeit liegt
	time_akku = buffer.Read() / samples_per_second;

// advance current_position to new_position:
	time_akku -= new_position - CurrentSamplePos() / samples_per_second;
}


/*	read data from file:
	return:
		errno=0, return bytes stored > 0:	ok,  block contains data
		errno=0, return bytes stored = 0:	eof, block empty
		errno>0, return -1:					error
*/
long RlesData::ReadFromFile ( int fd, FileType ft )
{
	XLogIn("\nRlesData:ReadFromFile");

	Purge();
	ClearError();			

	off_t filesize = FileLength(fd);	 if( filesize<0 ) return -1;	// error
	off_t oldpos = lseek(fd,0,SEEK_CUR); if( oldpos<0 )   return -1;	// errno set
	if( oldpos==filesize ) return 0;	// end of data
	XXLogLine( "fpos = %lu", (ulong)oldpos );

	try
	{
		buffer.Purge();
		buffer.Allocate( 1 MB ); 

		switch( ft )
		{
		default:		
			IERR();	

		case o_file:	
		case zx80_file:	
			if( oldpos>0 ) return 0;	// end of data: .80 / .o files contain only 1 block => ignore trash at and of file
			read_zx80_file(fd);	
			break;

		case p_file:	
		case zx81_file:	
			if( oldpos>0 ) return 0;	// end of data: .81 / .p files contain only 1 block => ignore trash at and of file
			read_zx81_file(fd);	
			break;

		case tap_file:	read_tap_file(fd);	break;	// return if ok, else throw error
		case rles_file:	read_rles_file(fd);	break;
		case tzx_file:	read_tzx_file(fd);	break;
		case pzx_file:	read_pzx_file(fd);	break;
		case aiff_file:	read_aiff_file(fd);	break;
		case wav_file:	read_wav_file(fd);	break;
		}

		if( data_end < data_start || data_end==buffer.Pos() )		// block w/o trailing pause: can only happen at eof or file error
		{								// at eof: bogus tzx file!
			store_tzx_pause( 0.001 );
			XLogLine( "last block has no trailing silence! (fixed)" );
		}
	}
	catch( file_error err )	{ ForceError(err.Error(),err.Text()); }
	catch( bad_alloc ) { ForceError(outofmemory); }

//	data_bytes = (data_end-data_start)/16;

	buffer.Resize( buffer.Pos() );
	buffer.ShrinkToFit();							// note: may throw bad_alloc
	SeekStart();
	
	if( NoError() )
	{
		return buffer.Size();
		
//		off_t newpos = lseek( fd,0,SEEK_CUR ); 
//		if( newpos>=0 ) return buffer.Size() ? newpos-oldpos /*ok*/ : 0 /*eof*/;
	}
	
	ShowAlert( "RlesData:ReadFromFile: %s\n", ErrorStr() );
	return -1;						// error
}


/*	write data to file:
	must not change current_position
*/
long RlesData::WriteToFile ( int fd, FileType ft )
{
	switch( ft )
	{
	default:		IERR();	
	case tap_file:	write_tap_file(fd);		break;
	case rles_file:	write_rles_file(fd);	break;
	case tzx_file:	write_tzx_file(fd);		break;
	case pzx_file:	write_pzx_file(fd);		break;
	case aiff_file:	write_aiff_file(fd);	break;
	case wav_file:	write_wav_file(fd);		break;
	}
	return 0;	// TODO: RlesData::WriteToFile()
}



////////////////////////////////////////////////////////////////////////////
///////////////// .80, .81, .o & .p file import & export ///////////////////
////////////////////////////////////////////////////////////////////////////

#define zx81lo	uchar(samples_per_second*520/cc_per_second)
#define zx81hi	uchar(samples_per_second*530/cc_per_second)
#define zx81ooo	uchar(samples_per_second*1.3e-3)

inline void RlesData::store_zx81_bit( bool b )
{
	store_pairs( zx81hi, zx81lo, b?8:3 );
	store_pair(  zx81hi, zx81ooo );
}

void RlesData::store_zx81_byte( uchar c )
{
	for( int i=7; i>=0; i-- ) store_zx81_bit( (c>>i)&1 );
}

void RlesData::store_zx81_data( uchar bu[], uint len )
{
	for( uint i=0; i<len; i++ ) store_zx81_byte( bu[i] );
}
	

/*	read zx80 tape file: "*.80" or "*.o"

	ZX80 Cassette File Structure
		x seconds    your voice, saying "filename" (optional)
		x seconds    video noise
		5 seconds    silence
		LEN bytes    data, loaded to address 4000h, LEN=(400Ah)-4000h.
		x seconds    silence / video noise

		ZX80 files do not have filenames
		ZX80 files cannot be autostarted. 
		The data is loaded to address $4000++
		The data contains the whole system area, basic program, VARS.
		Video memory is NOT included in ZX80 files.		
		the last byte of a (clean) file should be $80 (the last byte of VARS)
		The system area should contain proper data. 
		$400A	defines the end address (used to calculate the file length). 
		$4028++	may be misused for whatever purpose.

	.80 and .o files:	include only the data, loaded to $4000++
	.o files:			typically there is some garbage at the file end
*/
void RlesData::read_zx80_file ( int fd ) throw(file_error,bad_alloc)
{
	uchar sysvars[0x28]; 
	read_data( fd, sysvars, 0x28 );
	int end = Peek2Z( sysvars+0x0A );	
	int len = end - 0x4028;				if(len<=0) throw(file_error("sysvar $400A: illegal load end address"));
	uchar bu[len];
	read_data( fd, bu, len );
		
	SetMajorBlockInfo( "ZX80 file" );
	
	append_time(0.01,0);	// set init. level := low
	store_akku();
	
	data_start = buffer.Pos();

	store_zx81_data( sysvars,0x28 );
	store_zx81_data( bu, len );

	data_end = buffer.Pos();

	append_time(1.00,0);	// trailing pause, low level
	store_akku();
}


/*	read zx81 tape file: "*.81" or "*.p"

	ZX81 Cassette File Structure
		x seconds    your voice, saying "filename" (optional)
		x seconds    video noise
		5 seconds    silence
		1-127 bytes  filename (bit7 set in last char)
		LEN bytes    data, loaded to address 4009h, LEN=(4014h)-4009h.
		1 pulse      video retrace signal (only if display was enabled)
		x seconds    silence / video noise

		The data contains system area, basic program, video memory, VARS.
		the last byte of a (clean) file should be $80 (the last byte of VARS)

		$4014	defines the end address (used to calculate the file length)
		$4029	points to the next executed (autostarted) BASIC line
		$403B	indicates if program runs in SLOW or FAST mode (bit 6)
		$403C++	may be misused for whatever purpose, 
		video memory must contain 25 HALT opcodes if the file was saved in SLOW mode.

		Files should usually not exceed 16 KBytes. 
		The memory detection procedure in both ZX80 and ZX81 BIOS stops after 16 KBytes (at $8000), 
		and initializes the stack pointer at that address, even if more memory is installed. 
		Thus loading files of 16K or more would destroy the stack area, 
		unless a separate loader has previously moved the stack area to another location.
		However, most ZXes don't have more than 16K RAM, so bigger files won't work on most computers anyways.

	.81 and .p files:	include only the data, loaded to $4009++
	.p files:			typically there is some garbage at the file end
	.p81 files:			start with the 1..127 bytes filename, last byte ORed with $80
*/
void RlesData::read_zx81_file ( int fd ) throw(file_error,bad_alloc)
{
	uchar sysvars[0x3c-0x09]; 
	read_data( fd, sysvars, 0x3c-0x09 );
	int end = Peek2Z( sysvars+0x14-0x09 );	
	int len = end - 0x403c;					if(len<=0) throw(file_error("sysvar $4014: illegal load end address"));
	uchar bu[len];
	read_data( fd, bu, len );
	
	uchar noname[] = { 0x00|0x80 };			// name = " "  ((one space))
	
	SetMajorBlockInfo( "ZX81 file" );
	
	append_time(0.01,0);	// set init. level := low
	store_akku();
	
	store_zx81_data( noname, sizeof(noname) );

	data_start = buffer.Pos();

	store_zx81_data( sysvars,0x3c-0x09 );
	store_zx81_data( bu, len );

	data_end = buffer.Pos();

	append_time(1.00,0);	// trailing pause, low level
	store_akku();
}



////////////////////////////////////////////////////////////////////////////
/////////////////////// .tzx file import & export //////////////////////////
////////////////////////////////////////////////////////////////////////////



// for tzx file ID $19:
	struct TzxPRLE
	{
		uchar		symbol_idx;		
		ushort		repetitions;		// Intel byte order =>  use Peek2Z()
	};



/*	read data from file:
	in:	fd = file
		delim = expected final chunk:
			$00 pause/eof
			$22 group end
			$25 loop end
			$27 return from sequence.

	return:	
		errno cleared
		bytes stored > 0:	ok,  block contains data
		bytes stored = 0:	eof, block empty

	error:
		throw error:	file_error  or  bad_alloc
		errno set
*/
void RlesData::read_tzx_file ( int fd, uchar delim )		throw(file_error,bad_alloc)
{	
	uchar id;
	
a:	try 
	{ 
		id = read_char(fd);	
	} 
	catch( file_error err ) 
	{	
		if( err.Error()!=endoffile )	throw( err );
		if( delim!=0 )					throw( file_error( Using("tzx file: chunk $%2X missing",(uint)delim) ) );
		ClearError();	// clear eof error
		return;			// eof reached, nothing read.
	}

	switch(id)
	{
	case 0x10:	/*	Standard Speed Data Block (as in TAP files)
					This block must be replayed with the standard ROM values for all loading speed variables.
					+00 short	Pause after this block in milliseconds		[1000ms]
					+02 short	Length of following data		as in .TAP file
					+04 ff.		Data							as in .TAP file
					total chunk length = [2,3]+4		*/
		{	XLogIn("$10 standard speed data block");
			ulong ms_pause	= read_ushort_z(fd); 
			ulong len		= read_ushort_z(fd);	 if(len<2) len+=0x10000;
			uchar bu[len+1];  read_bytes(fd,bu,len); bu[len]=0x00; 

			XLogLine("bytes = %i",int(len-2));
			XLogLine("pause = %lu msec",ms_pause);
			if(NEXT_PHASE) XLogLine("inverted signal");
			store_data_block( bu, len*8/*+(ms_pause!=0)*/, ms_pause*0.001 );
			if( ms_pause ) break; else goto a;
		}

	case 0x11:	/*	Turbo Loading Data Block
					Similar to the normal TAP block but with additional info on the timings: 
					The same tape encoding is used as for the standard speed data block. 
					+00 short	Length of PILOT pulse						[2168cc]
					+02 short	Length of SYNC First pulse					[667cc]
					+04 short	Length of SYNC Second pulse					[735cc]
					+06 short	Length of ZERO bit pulse					[855cc]
					+08 short	Length of ONE bit pulse						[1710cc]

					+0A short	Length of PILOT tone (in pulses)			[Header: 8064 pulses, Data: 3220 pulses]
					+0C byte	Used bits in last byte						[8]
					+0D short	Pause after this block in milliseconds		[1000ms]
					+0F 3-byte	Length of following data
					+12 ff.		Data; format is as for TAP (MSb first)
					total chunk length = [0F,10,11]+12				*/
		{	XLogIn("$11 turbo loading block");
			Time pilot		= read_ushort_z(fd) / cc_per_second;	
			Time sync1		= read_ushort_z(fd) / cc_per_second;	
			Time sync2		= read_ushort_z(fd) / cc_per_second;	
			Time bit0		= read_ushort_z(fd) / cc_per_second;	
			Time bit1		= read_ushort_z(fd) / cc_per_second;	
			Time pilot_time  = read_ushort_z(fd) * pilot;
			uint last_bits		= read_uchar(fd);
			ulong ms_pause		= read_ushort_z(fd);
			ulong len			= read_u3byte_z(fd);

			XLogLine("bytes = %i",int(len-2));
			XLogLine("pause = %lu msec",ms_pause);
			if(NEXT_PHASE) XLogLine("inverted signal");

			uchar bu[len];	  read_bytes(fd,bu,len);

			store_data_block( bu, len*8-8+last_bits, pilot_time, ms_pause*0.001, pilot, sync1, sync2, bit0, bit1 );
			if( ms_pause ) break; else goto a;
		}
	
	case 0x12:	/*	Pure Tone
					Produce a tone which is basically the same as a pilot tone. 
					+00 short	Length of pulse in T-States
					+02 short	Number of pulses
					total chunk length = 4			*/
		{	XLogIn("$12 pure tone");
			ulong cc_per_pulse = read_ushort_z(fd);
			uint  pulses       = read_ushort_z(fd);
			if( pulses==0 ) throw(file_error("tzx file: ID $12: number of pulses = 0"));

			XLogLine( "%u pulses %s", pulses,pulses&1?"(odd)":"" );
			XLogLine( "cc/pulse: %lu", cc_per_pulse );									// cc per pulse
			XLogLine( "samples/pulse: %.1f", cc_per_pulse * samples_per_second/cc_per_second );	// samples per pulse
			XLogLine( "freq: %.3f Hz", cc_per_second / cc_per_pulse * 0.5 );			// freq = 0.5 per sec
			XLogLine( "total time: %.3f sec", pulses * cc_per_pulse / cc_per_second );	// total time [sec]

			store_pure_tone( pulses, cc_per_pulse / cc_per_second );
			goto a;
		}

	case 0x13:	/*	Sequence of Pulses of Different Lengths
					Up to 255 pulses with individual length can be stored in this block.
					+00 byte	Number of pulses
					+01 short	Length of first pulse in T-States
					+03 short	Length of second pulse...
						etc.
					total chunk length = [00]*2+1		*/
		{	XLogIn("$13 sequence of pulses of different length");
			uint   pulses	 = read_uchar(fd);		
			ushort bu[pulses]; read_data(fd,bu,pulses);

			if( pulses==0 ) throw(file_error("tzx file: ID $13: number of pulses = 0"));
			XLogLine( "%i pulses %s", (int)pulses,pulses&1?"(odd)":"" );

			for( uint i=0; i<pulses; i++ ) store_cc( Peek2Z(bu+i) );
			goto a;
		}

	case 0x14:	/*	Pure Data Block
					The same as in the turbo loading data block, except that it has no pilot or sync pulses.
					+00 short	Length of ZERO bit pulse
					+02 short	Length of ONE bit pulse
					+04 byte	Used bits in LAST Byte
					+05 short	Pause after this block in milliseconds (ms)
					+07 3-byte	Length of following data
					+0A ff.		Data (MSb first)
					total chunk length = [07,08,09]+0A		*/
		{
			XLogIn("$14 pure data block");
			Time bit0	= read_ushort_z(fd) / cc_per_second;	
			Time bit1	= read_ushort_z(fd) / cc_per_second;	
			uint last_bits	= read_uchar(fd);
			ulong ms_pause	= read_ushort_z(fd);	
			ulong len		= read_u3byte_z(fd);	
			uchar bu[len];	  read_bytes(fd,bu,len);

			XLogLine("bytes = %i",int(len));
			XLogLine("pause = %lu msec",ms_pause);
			if(NEXT_PHASE) XLogLine("inverted signal");
			
			store_pure_data( bu, len*8-8+last_bits, bit0, bit1 );
			store_tzx_pause( ms_pause*0.001 );
			if( ms_pause ) break; else goto a;
		}

	case 0x15:	/*	Direct Recording
					Each sample value is represented by one bit only: 0 for low, 1 for high. 
					The preferred sampling frequencies are 22050 (158 T states) or 44100 Hz (79 T states/sample). 
					+00 short	Number of T states per sample (bit of data)
					+02 short	Pause after this block in milliseconds (ms)
					+04 byte	Used bits (samples) in last byte of data (1..8)
					+05 3-byte	Data length
					+08 ff.		Data. Each bit represents one sample. MSb is played first.
					total chunk length = [05,06,07]+8			*/
		{
			XLogIn("$15 direct recording");
			uint cc_per_sample	= read_ushort_z(fd);	
			ulong ms_pause		= read_ushort_z(fd);	
			uint last_bits		= read_uchar(fd);
			ulong len			= read_u3byte_z(fd);			
			uchar bu[len];		  read_bytes(fd,bu,len);

			if( buffer.Pos()==0 ) samples_per_second = cc_per_second / cc_per_sample;
			else if( samples_per_second!=cc_per_second/cc_per_sample ) XLogLine("resampling...");

			uint e = 8;
			for( ulong i=0; i<len; i++ )
			{
				char byte = bu[i];
				if(i==len-1) e = last_bits;
				else if(byte==(char)0xff) { append_time_hi(8/samples_per_second); continue; } 
				else if(byte==(char)0x00) { append_time_lo(8/samples_per_second); continue; } 

			b:	uint n, b=0;
				n=0; while( (b<e) && byte<0 ) { byte*=2; b++; n++; } 
				if(n) append_time_hi(n/samples_per_second);
				n=0; while( (b<e) && byte>=0 ) { byte*=2; b++; n++; } 
				if(n) append_time_lo(n/samples_per_second);
				if(b<e) goto b;
			}

			store_akku();						// toggles phase after storing
			if( ms_pause ) XLogLine("pause = %lu msec",ms_pause);
			if( ms_pause ) { store_tzx_pause( ms_pause*0.001 ); break; }
			else { store_byte(0); goto a; }		// toggle back to last bit state
		}

	case 0x16:	//	C64 ROM Type Data Block		***deprecated***
				//	total chunk length = [00,01,02,03]		  
		{	XLogIn("$16 C64 ROM tape data block");
			throw( file_error("tzx file: ID $16: C64_ROM_TAPE_DATA_BLOCK not supported.") );
		}
		
	case 0x17:	//	C64 Turbo Tape Data Block	***deprecated***
				//	total chunk length = [00,01,02,03]		  
		{	XLogIn("$17 C64 turbo tape data block");
			throw( file_error("tzx file: ID $17: C64_TURBO_TAPE_DATA_BLOCK not supported.") );
		}
		
	case 0x18:	/*	CSW (compressed square wave) recording				
					+00	long	Length of block (w/o these 4 bytes)
					+04	short	pause after this block [msec]
					+06	3byte	sampling rate [1/sec]
					+09	byte	compression type: 1=RLE, 2=Z-RLE
					+0A	long	number of stored pulses (after decompression)
					+0E	data[]	CSW data or 'z' compressed CSW data
					total chunk length = [00,01,02,03] + 4
				*/
		{	XLogIn("$18 CSW compressed direct recording");

			ulong  len				= read_ulong_z(fd) -10;	XLogLine( "bytes = %lu",		len );
			ulong  ms_pause			= read_ushort_z(fd);	XLogLine( "pause = %lu msec",	ms_pause );
			Frequency samples_per_s	= read_u3byte_z(fd);	if( buffer.Pos()==0 ) samples_per_second = samples_per_s;
															XLogLine( "samples/sec = %f %s", samples_per_s,
																	  samples_per_s==samples_per_second ? "":"... resampling!" );
			uchar compression_type	= read_uchar(fd);		XLogLine( "%sRLES data", compression_type==1?"":"z-compressed ");
			ulong total_pulses		= read_ulong_z(fd);		XLogLine( "total pulses = %lu",	total_pulses );			
			uptr bu=0, zbu=0;

			try
			{
				bu = new uchar[len];
				read_bytes( fd, bu, len );		// data from file
				
				switch( compression_type )
				{
				default:
					throw( file_error("tzx file: ID $18: bad compression type") );				

				case 2:	// compressed:
						// uncompress bu[len] (compressed)  -->  bu[len] (uncompressed)
					{
						ulong zlen = total_pulses*2;		// *2: data may contain x-long pulses which are 5 bytes each
					c:	zbu = new uchar[zlen];
						int err = uncompress( zbu, &zlen, bu, len );		// zlib.h						
						switch(err)
						{
						case Z_OK:			// success
							delete[]bu; bu=zbu; zbu=0; len=zlen; 
							break;
						case Z_BUF_ERROR:	// dest[] too small
							delete[] zbu;
							zlen += zlen/2;
							goto c;
						case Z_MEM_ERROR:	throw(bad_alloc());
						case Z_DATA_ERROR:	throw(file_error("tzx file: ID $18: uncompress: corrupted data"));
						default:			throw(file_error("tzx file: ID $18: uncompress: unknown error"));
						} 
					}
					
				case 1:	// uncompressed:
					{
						if( len != total_pulses ) { XLogLine("*** len!=total_pulses"); }
						if( total_pulses > len )  { throw( file_error("tzx file: ID $18: bad size or count") ); }

						bool rf = samples_per_s != samples_per_second;		// resampling required ?

						for( ulong i=0; i<len; i++ )
						{					
							ulong n = bu[i];
							if( n ) 
							{
								if(rf) goto r; 
								store_byte(n); 
							}
							else
							{
								if( i+5>len ) throw(file_error("tzx file: ID $18: bad final pulse"));
								n = Peek4Z( (ulong*)(bu+i+1) ); i += 4;
								if( n/samples_per_s > 30.0 ) throw( file_error("tzx file: ID $18: bad pulse length") );
							r:	store_time( n / samples_per_s );
							}
						}
					}
				}
			}
			catch( file_error e )     { delete[]bu; delete[]zbu; throw(e); }
			catch( bad_alloc e ) { delete[]bu; delete[]zbu; throw(e); }

			delete[]bu; delete[]zbu; 
			store_tzx_pause( ms_pause*0.001 );
			if( ms_pause ) break; else goto a;
		}
			
	case 0x19:	/*	Generalized data block
					+00	long	len						Length of block (w/o these 4 bytes)
					+04 short	ms_pause				pause after block [ms]

					+06 long	p_data_cnt				Number of elements in pilot/sync data block (can be 0)
					+0A	byte	p_symdef_max_pulses		Maximum number of pulses per pilot/sync symbol
					+0B	byte	p_symdef_cnt			Number of elements in pilot/sync symbol definition table (0=256)

					+0C long	d_data_cnt				Number of elements in data data block (can be 0)
					+10 byte	d_symdef_max_pulses		Maximum number of pulses per data symbol
					+11 byte	d_symdef_cnt			Number of elements in data symbol definition table (0=256)
					+12ff:

			only if p_data_cnt>0:
						{	byte,							initial_phase: 0=toggle, 1=keep, 2=low, 3=high
							short							pulse length [cc]
						}		p_symdef[p_symdef_max_pulses*p_symdef_cnt]	symbol definition table

						{	byte,							symbol index
							short							repetitions
						}		p_data[p_data_cnt]		Pilot and sync data stream

			only if d_data_cnt>0:
						{	byte,							initial_phase: 0=toggle, 1=keep, 2=low, 3=high
							short							pulse length [cc]
						}		d_symdef[d_symdef_max_pulses*d_symdef_cnt]	symbol definition table

						byte[]	d_data[d_data_cnt]		Data data stream: symdef idx

					total chunk length = [00,01,02,03] + 4		*/
		{
			XLogIn("$19 Generalized data block");

			skip_bytes(fd,4);	// block length
			ulong ms_pause		= read_ushort_z(fd);	XLogLine( "pause = %lu msec",	ms_pause );

			ulong p_data_cnt	= read_ulong_z(fd);		XLogLine( "Number of elements in pilot/sync data stream = %lu",	p_data_cnt );
			ulong p_symdef_max_pulses = read_uchar(fd);	XLogLine( "Maximum number of pulses per symbol = %lu", p_symdef_max_pulses );
			ulong p_symdef_cnt	= read_uchar(fd);		if( p_symdef_cnt==0 ) p_symdef_cnt = 256;
														XLogLine( "Size of symbol definition table = %lu", p_symdef_cnt );

			ulong d_data_cnt	= read_ulong_z(fd);		XLogLine( "Number of elements in content data stream = %lu",	d_data_cnt );
			ulong d_symdef_max_pulses = read_uchar(fd);	XLogLine( "Maximum number of pulses per symbol = %lu", d_symdef_max_pulses );
			ulong d_symdef_cnt	= read_uchar(fd);		if( d_symdef_cnt==0 ) d_symdef_cnt = 256;
														XLogLine( "Size of symbol definition table = %lu", d_symdef_cnt );

			if( p_data_cnt )
			{
			// read symbol definition table:
				uchar symdef          [ p_symdef_cnt * (1+p_symdef_max_pulses*2) ];
				read_bytes( fd, symdef, p_symdef_cnt * (1+p_symdef_max_pulses*2) );

			// read data stream:
				uchar data          [ p_data_cnt*3 ];
				read_bytes( fd, data, p_data_cnt*3 );

			// store data stream:
				for ( uint i=0; i<p_data_cnt; i++ )
				{
					uint symbol  = data[3*i+0]; 
					uint rep_cnt = data[3*i+1] + 256 * data[3*i+2];		// caveat: byte oder and missaligned short
					XXLogLine("symbol: idx=%i, repetitions=%i",(int)symbol,(int)rep_cnt);
					if( symbol>=p_symdef_cnt ) throw( file_error("tzx file: ID $18: symbol>p_symbols") );
					while(rep_cnt--) { store_tzx19_symbol( &symdef[symbol*(1+p_symdef_max_pulses*2)], p_symdef_max_pulses ); }
				}
			}

			if( d_data_cnt )
			{
			// read symbol definition table:
				uchar symdef          [ d_symdef_cnt * (1+d_symdef_max_pulses*2) ];
				read_bytes( fd, symdef, d_symdef_cnt * (1+d_symdef_max_pulses*2) );

			// read data stream:
				ulong NB = (ulong)ceil(log2(d_symdef_cnt));	XLogLine("bits per symbol in content data stream = %lu", NB);
				ulong DS = (ulong)ceil(NB*d_data_cnt/8);	XLogLine("size of content data stream = %lu bytes", DS);
 				uchar data          [ DS ];
				read_bytes( fd, data, DS );

			// store data stream:
				uptr p = data;
				int bit = 7;
				for ( uint i=0; i<d_data_cnt; i++ )
				{
					uint symbol = 0;
					for( ulong i=0; i<NB; i++ ) { symbol += symbol + ((*p>>bit)&1); if(!bit--) { bit=7; p++; } }

					if( symbol>=d_symdef_cnt ) throw( file_error("tzx file: ID $18: symbol>d_symbols") );
					store_tzx19_symbol( &symdef[symbol*(1+d_symdef_max_pulses*2)], d_symdef_max_pulses );
				}
			}
			
			store_tzx_pause( ms_pause*0.001 );
			if( ms_pause ) break; else goto a;
		}
	
	case 0x20:	/*	Pause (Silence) or `Stop the Tape' Command
					This will make a silence (low amplitude level (0)) for a given time in milliseconds. 
					If the value is 0 then the emulator or utility should (in effect) STOP THE TAPE, 
					+00 short	Value of the pause in milliseconds (ms)
					total chunk length = 2				*/
		{
			XLogIn("$20 pause");
			ulong ms_pause = read_ushort_z(fd);		XLogLine( "pause = %lu msec", ms_pause );

			if( !ms_pause ) 
			{ 
				XLogLine("stop the tape!");
				if( buffer.Pos()==0 ) SetMajorBlockInfo("Stop the tape!");
				ms_pause = 5000; 
			}
			
			store_tzx_pause( ms_pause*0.001 );
			break;	// return ok
		}
					
	case 0x21:	/*	Group Start
					This block marks the start of a group of blocks which are to be treated as one single (composite) block. 
					This is very handy for tapes that use protected blocks like Speedlock 1 (which has around 64 pure tone blocks) 
					or Bleeploads (which may well have over 160 custom loading blocks). 
					You can also give the group a name (example 'Speedlock 1 Block').

					For each group start block, there must be a group end block. Nesting of groups is not allowed.
					+00 byte	Length of the Group Name
					+01 ff.		Group name in ASCII (please keep it under 30 characters long)
					total chunk length = [00]+01			*/
		{
			XLogIn("$21 group start");
			
			uint cnt = read_uchar(fd);	
			if( cnt )
			{
				char z[cnt+1]; z[cnt]=0;
				read_bytes(fd,z,cnt);		
				if(z[0]) 
				{
					XLogLine( "name = %s", z );
					SetMajorBlockInfo(z);
				}
			}

			read_tzx_file( fd, 0x22 /*group end*/ );
			if( buffer.Pos()>data_end /*trailing pause?*/ ) break; else goto a;
		}

	case 0x22:	/*	Group End
					This block type has no body! 
					total chunk length = 0			*/
		{
			XLogIn("$22 group end");
			
			if( delim==0x22 ) return;
			throw( file_error("tzx file: ID $22: GROUP_END unexpected") );
		}

	case 0x23:	/*	Jump To Block			TODO
					This block will enable you to jump from one block to another within the file. 
					The value is a signed short; eg. Some examples: 
					Jump 0 = 'Loop Forever' - this should never happen
					Jump 1 = 'Go to the next block' - it is like NOP in assembler ;)
					Jump 2 = 'Skip one block'
					Jump -1 = 'Go to the previous block' 

					All blocks are included in the block count !
					+00 short	Relative jump value
					total chunk length = 2			*/
		{
			XLogIn("$23 jump to block");
			throw(file_error("tzx file: ID $23: JUMP_TO_BLOCK not yet supported"));
		}

	case 0x24:	/*	Loop Start
					If you have a sequence of identical blocks, or of identical groups of blocks, 
					you can use this block to tell how many times they should be repeated.
					For simplicity reasons don't nest Loop blocks!
					+00 short	Number of repetitions (greater than 1)
					total chunk length = 2			*/
		{	XLogIn("$24 loop start");
			XXXASSERT(time_akku==0.0);

			ulong loop_count = read_ushort_z(fd);			
			if( loop_count==0 ) throw(file_error("tzx file: ID $24: loop count = 0")); 
			ulong loop_start = buffer.Pos();
			//ulong loop_start_sample_pos = CurrentSamplePos();

			read_tzx_file( fd, 0x25 /*loop end*/ );
			XXXASSERT(time_akku==0.0);
			
			ulong block_size = buffer.Pos() - loop_start;
			if( block_size==0 ) throw(file_error("tzx file: ID $25: loop contains no data")); 			
			if( block_size&1 ) { XLogLine("odd number of pulses in repeated block"); }

			float total_required = loop_count * float(block_size);
			if( total_required>16 MB ) throw(file_error("tzx file: ID 25$: total loop size exceeds any sane limit")); // sanity test
			buffer.Allocate( loop_start + (ulong)total_required );

			XLogLine( "single block size = %lu bytes", block_size );
			XLogLine( "repetitions       = %lu",       loop_count );
			XLogLine( "total loop size   = %lu bytes", ulong(total_required) );

			while( --loop_count )
			{
				buffer.Write( buffer.Data()+loop_start, block_size );
			}

			
			//cuptr q = buffer.Data() + loop_start;
			//uptr  z = const_cast<uptr>(buffer.Data()) + buffer.Pos();		// TODO: 
			//cuptr e = buffer.Data() + buffer.Size();

			//XLog("single block size = %lu bytes\n", ulong(z-q) );
			//XLog("repetitions       = %lu\n", ulong(loop_count) );
			//XLog("total loop size   = %lu bytes\n", ulong(total_required) );

			//while( z!=e ) { *z++ = *q++; }
			//buffer.Seek( buffer.Size() );			
			goto a;
		}

	case 0x25:	/*	Loop End
					This block has no body! 
					total chunk length = 0			*/
		{
			XLogIn("$25 loop end");
			if( delim==0x25 ) return;
			throw(file_error("tzx file: ID $25: unexpected LOOP_END"));
		}

	case 0x26:	/*	Call Sequence			TODO
					This block is analogue of the CALL Subroutine statement. 
					It basically executes a sequence of blocks that are somewhere else and then goes back to the next block. 
					Because more than one call can be normally used you can include a list of sequences to be called. 
					The 'nesting' of call blocks is also not allowed for the same reasons. 
					You can, of course, use the CALL blocks in the LOOP sequences and vice versa.
					The value is relative for the obvious reasons. See 'Jump To Block' for reference on the values.
					+00 short	Number of Calls to be made
					+02 short	Call 1 block number (relative signed offset)
						ff.		Next Call block number
					total chunk length: [00,01]*02+02			*/
		{
			XLogIn("$26 call sequence");
			throw( file_error("tzx file: ID $26: CALL_SEQUENCE not yet supported") );
		}

	case 0x27:	/*	Return From Called Sequence
					This block has no body. 
					total chunk length = 0			*/
		{
			XLogIn("$27 return from sequence");
			if( delim==0x27 ) return;
			throw( file_error("tzx file: ID $27: unexpected RETURN_FROM_SEQUENCE") );
		}

	case 0x28:	/*	Select Block
					This block is useful when the tape consists of two or more separately-loadable parts. 
					With this block, the user can select one of the parts and the emulator will start loading from that block. 
					The emulator has to show a menu with the selections when it encounters this block. 
					All offsets ar relative signed words.
					+00 short	Length of the whole block (without these two bytes)
					+02 byte	Number of Selections
					+03 short	Relative Offset of Selection 1
					+05 byte	Length of the Description 1 Text
					+06 ff.		Description 1 Text in ASCII
								Please use only one line per Selection and limit it to 30 characters!
					ff. ff.		Next Selection Offset
					total chunk length: [00,01]+02		*/
		{
			XLogIn("$28 select block: not supported. ignored.");
			uint cnt = read_ushort_z(fd);	
			skip_bytes(fd,cnt);	
			goto a;
		}

	case 0x2A:	/*	Stop Tape if in 48K Mode
					When this block is encountered, the tape should stop if the machine is an 48K Spectrum.
					This block is to be used for multiloading games that load one level at a time in 48K mode, 
					but load the entire tape at once if in 128K mode.
					This block has no body, but follows the extension rule:
					+00 long	Length of the whole block (0)
					total chunk length = 4			*/
		{
			XLogIn("$2A stop tape if in 48k mode");
			skip_bytes(fd,4);		

			if( buffer.Pos()==0 ) SetMajorBlockInfo("Stop the tape if in 48k mode!");
			store_tzx_pause( 3.0 /*Pause*/ );
			break;	// return ok
		}

	case 0x2B:	/*	Set signal level
					+00	long	length of block (without these 4 bytes) = 1
					+04	byte	signal level: 0 or 1
					total chunk length = [00,01,02,03] + 1  =  5
				*/
		{
			XLogIn("$2B set signal level 0/1");
			uchar bu[5]; uchar& next_phase = bu[4];
			read_bytes(fd,bu,5);

			XXXASSERT(time_akku==0.0);
			if( NEXT_PHASE != next_phase ) store_byte(0);	// store zero-length pulse to toggle NEXT_PHASE
			goto a;
		}

	case 0x30:	/*	Text Description
					This is meant to identify parts of the tape. 
					This description is not guaranteed to be shown while the tape is playing, 
					but can be read while browsing the tape or changing the tape pointer.
					The description can be up to 255 characters long but please keep it down to 
					about 30 so the programs can show it in one line.
					Please use 'Archive Info' block for title, authors, publisher,...
					+00 byte	Length of the Text
					+01 ff.		Text in ASCII
					total chunk length: [00]+01		*/
		{
			XLogIn("$30 block info");
			uint cnt = read_uchar(fd);	
			char z[cnt+1]; z[cnt]=0;
			read_bytes(fd,z,cnt);		
			if( SameStr(LowerStr(LeftStr(z,8)),"created ") )
			{ XLogLine("abusive usage (skipped)"); }			// "Created with Ramsoft MakeTZX" or similar
			else SetMajorBlockInfo(z);
			goto a;
		}

	case 0x31:	/*	Message Block 
					Display a message for a given time. 
					This should not stop the tape and it should not make silence. 
					If the time is 0 then the emulator should wait for the user to press a key.
					It is not guaranteed that the emulator will actually display the message.
					+00	byte	Time for which the message should be displayed in seconds (s)
					+01 byte	Length of the message
					+02 ff.		Message that should be displayed (in ASCII)
								- Stick to a maximum of 30 chars per line
								- Use single 0x0D to separate lines.
								- Stick to a maximum of 8 lines.
					total chunk length = [01]+2			*/
		{
			XLogIn("$31 'subtitle' block (ignored)");
			uchar bu[2]; uchar& cnt = bu[1];
			read_bytes(fd,bu,2);
			skip_bytes(fd,cnt);			
			goto a;
		}

	case 0x32:	/*	Archive Info			TODO
					Use this block at the beginning of the tape to identify the title of the game, author, publisher, year of publication, price (including the currency), type of software (Arcade Adventure, Puzzle, Word Processor, ...), protection scheme it uses (Speedlock 1, Alkatraz, ...) and its origin (Original, Budget re-release, ...), etc. This block is built in a way that allows easy future expansion. 
					The block consists of a series of text strings. Each text has its identification number and then the ASCII text.
					If all texts in the tape are in English language then you don't have to supply the 'Language' field.

					The information about what hardware the tape uses is in the 'Hardware Type' block, so no need for it here.
					+00 short	Length of the block (without these two bytes)
					+02 byte	Number of text strings
					+03 ff.		Text strings:
						+00 byte	Text Identification byte:  
									00 - Full Title
									01 - Software House / Publisher
									02 - Author(s)
									03 - Year of Publication
									04 - Language
									05 - Game/Utility Type
									06 - Price
									07 - Protection Scheme / Loader
									08 - Origin
									FF - Comment(s)
						+01 byte	Length of text
						+02 ff.		Text in ASCII format
						...			Next Text
					total chunk length = [00,01]+2		*/
		{
			XLogIn("$32 archive info: not yet implemented");
			uint cnt = read_ushort_z(fd);	
			skip_bytes(fd,cnt);				
			goto a;
		}

	case 0x33:	/*	Hardware Type		TODO
					This tells what hardware the programs on this tape requires.  (only 100% shure infos should be listed) 
					00 byte		Number of machines and hardware types for which info is supplied
					01 ff.		List of machines and hardware:
						  00 byte	Hardware type
						  01 byte	Hardware ID
						  02 byte	Value
								00 - The tape works on this machine or with this hardware
								01 - The tape works on this machine and requires it's special hardware
								02 - The tape works on this machine but it doesn't use it's special hardware
								03 - The tape does NOT run on this machine or with this hardware
						  .. .  Next Machine/Hardware Info (if any)

					The list of types and IDs is somewhat large, and may be found at the end of the format description.
					total chunk length: [00]*3+1			*/
		{
			XLogIn("$33 hardware type: not yet implemented");
			uint cnt = read_uchar(fd);	
			skip_bytes(fd,cnt*3);			
			goto a;
		}

	case 0x34:	/*	Emulation Info
				// ***deprecated*** in version 1.20
					misc. info, e.g. r-register emu, ldir emu, hires color, etc.
					total chunk length = 8			*/
		{
			XLogIn("$34 emulation info: ignored.");
			skip_bytes(fd,8);			
			goto a;
		}

	case 0x35:	/*	Custom Info Block
				// ***deprecated*** in version 1.20
					This block can be used to save any information you want. 
					+00 16 bytes	Identification string (in ASCII)
					+10 long		Length of the custom info
					+14 ff.			Custom info
					total chunk length = [$10,$11,$12,$13]+$14		*/
		{
			XLogIn("$35 custom data block: deprecated, ignored.");
			skip_bytes(fd,16);				
			ulong len = read_ulong_z(fd);	
			skip_bytes(fd,len);				
			goto a;
		}
	
	case 'Z':	// header of concatenated file
		{
			XLogIn("$5A file header (skipped)");
			skip_bytes(fd,9);
			goto a;
		}
	
	case 0x40:	/* snapshot block				
				// ***deprecated*** in version 1.20
					+00	1 byte		snapshot type: 0=z80, 1=sna
					+01 3 bytes		N = length of snapshot following
					+04 N bytes		snapshot
					total chunk length =		*/
		{
			XLogIn( "$40 snapshot" );
			throw(file_error("tzx file: ID $40: SNAPSHOT deprecated, not supported."));
		}
					
	default:
		{
			XLogIn( "unknown chunk ID $%2X: ignored",(int)id );
			ulong len = read_ulong_z(fd);	
			skip_bytes(fd,len);
			goto a;
		}
	}

// block end after pause:

	XXXASSERT( NoError() );
	XXXASSERT( id!=0x22 && id!=0x25 && id!=0x27 );
	
	
	if( delim!=0 ) 
	{
		goto a;
	//	if( delim==0x22 /*group end*/ ) goto a;
	//	throw( file_error( Using("tzx file: block $%2X missing, found $%2X",uint(delim),uint(id)) ) );
	}
}


void RlesData::write_tzx_file ( int fd )	throw(file_error,bad_alloc)	{ TODO();
}




////////////////////////////////////////////////////////////////////////////
/////////////////////// .pzx file import & export //////////////////////////
////////////////////////////////////////////////////////////////////////////

void RlesData::write_pzx_file ( int fd )	throw(file_error,bad_alloc)	{ TODO();
}


void RlesData::read_pzx_file ( int fd )	throw(file_error,bad_alloc)	
{
	long id;
	long blen;
	uptr bu = NULL;

	XLogIn("RlesData::read_pzx_file");
	
a:	try 
	{ 
		id   = read_long_x(fd);	
		blen = read_long_z(fd);	if( blen>16 MB ) throw(file_error("pzx: ill. blen"));
	} 
	catch( file_error err ) 
	{	
		if( err.Error()!=endoffile ) throw( err );
		ClearError();				// clear eof error
		return;						// eof reached, nothing read.
	}

	try
	{
		delete[] bu; bu=NULL;
		bu = new uchar[blen+1+99];		// may throw bad_alloc
		read_bytes(fd,bu,blen);			// may throw file_error

		ptr q = (ptr)bu;				// read pointer
		ptr e = (ptr)bu+blen;			// end of data

		switch(id)
		{
		case 'ZXTP':	// old PZX header up to version 0.3 (draft)
		{
			XLogIn("ZXTP block");
		
			if(blen<2) throw(file_error("pzx-file: ZXTP: ill. blen"));
			uchar vh=*q++;		// major version number
			uchar vl=*q++;		// minor version number
			XLogLine("version %hhu.%hhu (draft)", vh,vl );
			throw( file_error("old pzx file encoded using a draft version pre 1.0. no longer supported.") );
		}
		
		case 'PZXT':	// PZX header block:
		{				//		dc.b	major version number
						//		dc.b	minor version number
						//		dc.s	info  dictionary
						//
			XLogIn("PZXT block");
		
			if(blen<2) throw(file_error("pzx-file: PZXT: ill. blen"));
			uchar vh=*q++;		// major version number
			uchar vl=*q++;		// minor version number
			XLogLine("version %hhu.%hhu", vh,vl );
			if( vh!=1 ) throw(file_error(Using("version %hhu.%hhu of pzx file is not supported.",vh,vl)));

			// validate strings:
			*e = 0;						// enforce string delimiter for last string
			for( ptr p=e; p>=q; p-- ) { if(*p&&*(uptr)p<' ') throw(file_error("pzx-file: PZXT: ill. char")); }

			// first string: tape name
			int n = strlen(q); if(n>80) throw(file_error("pzx-file: PZXT: ill. strlen"));
			if(n) 
			{
				XLogLine( "Title:\t%s",q );
				SetMajorBlockInfo(q); 
			}
			q+=n+1;

			// misc. info:
				//	Tag        TZX ID   Info
				//  ---------- ------ - ------------------------
				//	Publisher  [0x01] - Software house/publisher
				//	Author     [0x02] - Author(s)
				//	Year       [0x03] - Year of publication
				//	Language   [0x04] - Language
				//	Type       [0x05] - Game/utility type
				//	Price      [0x06] - Original price
				//	Protection [0x07] - Protection scheme/loader
				//	Origin     [0x08] - Origin
				//	Comment    [0xFF] - Comment(s)
				//	Some keys, like Author or Comment, may be used more than once.
				//
			while( e-q>2 )
			{
				cstr s1 = q; 
				int  n1 = strlen(s1);	q+=n1+1; if(e-q<1) goto a;	// unmatched entry: ignore
				cstr s2 = q; 				
				int  n2 = strlen(s2);	q+=n2+1; 
				XLogLine( "%s:\t%s", s1, s2 );
			}
		}	goto a;
		
		case 'PULS':	// arbitrary repetitions of pulses
		{
			XLogIn("PULS block");

			// first pulse is low:
			if( CURRENT_RECORD_PHASE != 0/*requested_phase*/ ) 
			{
				XLogLine("toggle initial phase");
				store_byte(0);
			}

			while( q<e )
			{
			// decode duration and repeat count acc. to code sample in specs:
				ulong cnt = 1, dur = Peek2Z(q); q+=2; 
				if( dur > 0x8000  ) { cnt = dur & 0x7FFF; dur = Peek2Z(q); q+=2; }
				if( dur >= 0x8000 ) { dur = ((dur & 0x7FFF)<<16) + Peek2Z(q); q+=2; }
				if(q>e) { XLogLine("incomplete pulse"); goto a; }
				XXLogLine("%u x %u cc", int(cnt), int(dur) );
			// store pulses:
				if(dur!=0) { while(cnt--) { store_cc(dur); } } 
				else if(cnt&1) store_byte(0);	// if cnt==1 and dur==0 then toggle phase
			}
		}	goto a;
			
		case 'DATA':
		{
			XLogIn("DATA block");

			ulong   bits = Peek4Z(q);	q += 4;			// total bits in data
			bool    initial_phase = (long)bits<0; bits &= 0x7fffffff;
			XLogLine("%i bytes + %i bits", int(bits/8),int(bits%8));
			uint    tail = Peek2Z(q);	q += 2;			// trailing pulse
			uint    cnt0 = Peek1Z(q);	q += 1;			// pulses for bit==0 
			uint    cnt1 = Peek1Z(q);	q += 1;			// pulses for bit==1
			ushort*	dur0 = (ushort*)q;	q += 2*cnt0;	// table of pulse durations for bit 0
			ushort*	dur1 = (ushort*)q;	q += 2*cnt1;	// table of pulse durations for bit 1

		#ifndef _LITTLE_ENDIAN
			for( uint i=0; i<cnt0; i++ ) dur0[i] = Peek2Z(dur0+i);	
			for( uint i=0; i<cnt1; i++ ) dur1[i] = Peek2Z(dur1+i);	
		#endif			

			XXLogLine("pulses for bit 0: %i", int(cnt0));
			if(XXLOG) { for(uint i=0;i<cnt0;i++) Log("%i ",int(dur0[i])); Log(" cc\n"); }
			XXLogLine("pulses for bit 1: %i", int(cnt1));
			if(XXLOG) { for(uint i=0;i<cnt1;i++) Log("%i ",int(dur1[i])); Log(" cc\n"); }
			XXLog("length of tail pulse: %i cc", int(tail));
						
			if( q + (bits+7)/8 > (ptr)bu+blen ) throw(file_error("pzx:DATA: block size mismatch"));
			
			// first pulse polarity:
			if( CURRENT_RECORD_PHASE != initial_phase ) 
			{
				XLogLine("toggle initial phase");
				store_byte(0);
			}

			if( !data_start ) data_start = buffer.Pos();

			if( !major_block_info ) SetBlockInfoFromData( bits/8, (uptr)q );
		
			while( bits )
			{
				char byte=*q++;
				uint n=Min(8ul,bits); bits-=n;
				while( n-- )
				{
					uint    cnt = byte<0 ? cnt1 : cnt0; 
					ushort* dur = byte<0 ? dur1 : dur0;
					byte+=byte;					
					for( uint i=0; i<cnt; i++ ) { if(dur[i]) store_cc(dur[i]); else store_byte(0); }
				}
			}
			
			if(tail) store_cc(tail);
		}	goto a;
			
		case 'PAUS':
		{	
			XLogIn("PAUS block");

			ulong dur  = Peek4Z(q); q+=4;
			bool initial_phase = (long)dur<0; dur &= 0x7fffffff;
			XLogLine( "%.3f", float(dur/3500000.0) );
			if( q>(ptr)bu+blen ) throw(file_error("pzx-file: PAUS: block truncated"));

			data_end = buffer.Pos();
			
			if( CURRENT_RECORD_PHASE != initial_phase ) store_byte(0);
			if(dur) store_cc(dur);
		}	break;		// block end
		
		case 'BRWS':
		{
			XLogIn("BRWS block");

			str s = SubStr( (ptr)q, (ptr)bu+blen );
			for( ptr p=strchr(s,0); --p>=s; ) { if(*(uptr)p<' ') *p=' '; }	// replace ctl codes
			XLogLine( "%s",s );
			if( !major_block_info ) SetMajorBlockInfo(s);
		}	goto a;
		
		case 'STOP':
		{
			XLogIn("STOP block");

			uint flag = Peek2Z(q); q+=2;
			if( buffer.Pos()==0 ) 
			{
				SetMajorBlockInfo( flag==0 ? "Stop the tape" : "Stop tape if loading into 48k model");
				store_time(5.0);
			}
		}	break;

		default:
			XLogLine( "unknown block type $%8lx", id );
			goto a;

		}		// switch(id)
	}		// try
	catch( file_error e ) { delete[] bu; throw(e); }
	catch( bad_alloc e  ) { delete[] bu; throw(e); }
	
	delete[] bu;
	return;				// ok
}






////////////////////////////////////////////////////////////////////////////
/////////////////////// .rles file import & export /////////////////////////
////////////////////////////////////////////////////////////////////////////

/*
	; info text for display: (multiple, optional)
		dc.m	"info"
		dc.l	length					; total length of data following
		dc.m	"info text",$00			; UTF-8 with c-style delimiter
		dc.s	padding					; optional some padding

	; Raw data: (multiple)
		dc.m	"rles"
		dc.l	length					; total length of data [bytes] following
		dc.l	sampling_frequency		; frequency in hertz
		dc.s	rles_data				; one high and low phase per byte

	; Header of a file concatenated by the user:
		dc.m	"RlesTape"				; file magic
		dc.m	"1.0", $00				; major + minor version
*/

void RlesData::read_rles_file ( int fd )	throw(file_error,bad_alloc)
{
	ulong hdr,len;
	
	XLogIn("RlesData::read_rles_file");

a:	try
	{
		hdr = read_ulong_z(fd); 
	} 
	catch( file_error err ) 
	{
		if( Error()!=endoffile ) throw(err);
		ClearError(); 
		return;
	}
	
	len = read_ulong_z(fd); 

	switch(hdr)
	{
	case 'Rles':			// skip concatenated file's header
		{					// note: we could check the version number..
			XLogIn("concatenated file header (skipped)");
			skip_bytes(fd,4); 
			goto a;
		}

	case 'info':
		if(len<100)
		{
			XLogIn("block info");
			char bu[len+1]; bu[len]=0;
			read_bytes(fd,bu,len);
			SetMajorBlockInfo(bu);
			goto a;
		}
		SetError("rles file: broken 'info' chunk");
		break;

	case 'rles':
		{
			XLogIn("rles data block");

			if(len<4 || len>1 MB)
			{ SetError("rles file: broken 'rles' chunk"); break; }

			samples_per_second = read_ulong_z(fd); len-=4;
			XLogLine( "rles compressed audio, f=%lu Hz, %lu samples", (ulong)samples_per_second, (ulong)len );

			if( samples_per_second<8000 || samples_per_second>100000 )
			{ SetError("rles file: sample frequency out of range"); break; }

			uchar bu[len]; read_bytes( fd, bu, len );
			buffer.Grow(len*2);

			uptr p=bu;
			uptr e=bu+len;
			while(p<e)
			{
				uchar byte = *p++;
				append_time_hi( byte&0x0f ? (byte>>4) : (byte>>4)*15 );
				append_time_lo( byte&0xf0 ? (byte&15) : (byte&15)*15 );
			}
			store_akku();		// flush time_akku
			
			break;	// ok
		}
	}
}


void RlesData::write_rles_file ( int fd )		throw(file_error,bad_alloc)
{
	if(buffer.Size()==0) { ClearError(); return; }

	if(MajorBlockInfo())
	{
		write_ulong_z(fd,'info');				
		ulong len = strlen(MajorBlockInfo())+1;
		write_ulong_z(fd,len);					
		write_bytes(fd,MajorBlockInfo(),len);			
	}

    TODO();		// RlesData::write_rles_file
    
	write_ulong_z(fd,'rles');					
	write_ulong_z(fd,buffer.Size()+4);			
	write_ulong_z(fd,(ulong)samples_per_second);	
	write_bytes(fd,buffer.Data(),buffer.Size());
}



////////////////////////////////////////////////////////////////////////////
/////////////////////// .aiff file import & export /////////////////////////
////////////////////////////////////////////////////////////////////////////


/*	read from AIFF file
	return if ok, else throw error
*/
void RlesData::read_aiff_file ( int fd )		throw(file_error,bad_alloc) {	TODO();
}


void RlesData::write_aiff_file ( int fd )		throw(file_error,bad_alloc) {	TODO();
}




////////////////////////////////////////////////////////////////////////////
/////////////////////// .wav file import & export //////////////////////////
////////////////////////////////////////////////////////////////////////////


/*	read from WAV file
	return if ok, else throw error
*/
void RlesData::read_wav_file ( int fd )		throw(file_error,bad_alloc) {	TODO();
}


void RlesData::write_wav_file ( int fd )		throw(file_error,bad_alloc) {	TODO();
}


	
////////////////////////////////////////////////////////////////////////////
/////////////////////// .tap file import & export //////////////////////////
////////////////////////////////////////////////////////////////////////////

/*	read one block of data from .tap file
	return if ok, else throw error
*/
void RlesData::read_tap_file ( int fd )		throw(file_error,bad_alloc)
{	
	ulong len = read_ushort_z(fd);	if(len<2) len+=0x10000;
	uchar bu[len+1];  	read_bytes(fd,bu,len); bu[len]=0x00; 
	store_data_block( bu, len*8+1, 1.000 /*pause*/ );
}


/*	write one block of data to .tap file
*/
void RlesData::write_tap_file ( int fd )	throw(file_error,bad_alloc)
{
	ClearError(); 
	if( TotalSamples()==0 ) return;		// no data

	uint samples_per_pilot	= int( cc_per_pilot / cc_per_second * samples_per_second +0.5 );
	uint samples_per_sync1	= int( cc_per_sync1 / cc_per_second * samples_per_second +0.5 );
	uint samples_per_sync2	= int( cc_per_sync2 / cc_per_second * samples_per_second +0.5 );
	uint samples_per_bit_0	= int( cc_per_bit_0 / cc_per_second * samples_per_second +0.5 );
	uint samples_per_bit_1	= int( cc_per_bit_1 / cc_per_second * samples_per_second +0.5 );

	uint phases_per_pilot_sequence_min	= 256;			// mind. zu lesende Pilothalbphasen
	uint samples_per_pilot_min	= samples_per_pilot - (samples_per_pilot-samples_per_sync1)/2;	// Absolut limit
	uint samples_per_sync1_max	= samples_per_sync1 + (samples_per_pilot-samples_per_sync1)/4;	// Absolut limit
	uint samples_per_sync12_max	= (samples_per_sync1+samples_per_sync2) *3/2;		// Absolut limit for sync1+sync2
	uint samples_per_bit11_max	= samples_per_bit_1*2 *10/8;						// Absolut limit for 2 data bit phases

/*	phase 1:	
	erwarte mind. 1000 pilothalbphasen		~ 0.6 sec
	es wird die gesamtzeit für eine hi-low-phase gemessen, um auch stark asynchrone signale zu erkennen.
	es wird jedoch ein 'enges' Limit gesetzt: SOLLLÄNGE*9/10 ... SOLLLÄNGE*10/9
	danach wird der pilotton als gültig anerkannt.
*/		
	cuptr q = buffer.Data();
	cuptr e = q + buffer.Size();
	uint a = 0;					// last runlength
	uint b = 0;					// current runlength
	#define next_nibble()		*q++					// note: kein test auf 'verteilte' samples
	#define get_next_nibble()	a=b; b=next_nibble(); 
	
p1:	for( uint count=0; q<e && count<phases_per_pilot_sequence_min; count++ )
	{
		get_next_nibble();
		if( a+b < 2*(samples_per_pilot*9/10) ) goto p1;	// zu kurz
		if( a+b > 2*(samples_per_pilot*10/9) ) goto p1;	// zu lang
		if( b < samples_per_pilot_min ) goto p1;		// zu unsymmetrisch => erkennen des syncs ungewiss
	}

/*	phase 2:	
	warte auf 1. sync-flanke
	Der Pilot wird als gültig anerkannt, solange zwei aufeinander folgende Phasen im Limit liegen.
	es wird ein 'weites' Limit akzeptiert: SOLLLÄNGE*8/10 ... SOLLLÄNGE*10/8.
	Sobald eine Halbphase anliegt, die kurz genug ist, wird der Sync-Impuls erwartet.
*/		
	while( q<e )
	{
		get_next_nibble();
		if (b <= samples_per_sync1_max) break;			// sync!
		if( a+b < 2*(samples_per_pilot*8/10) ) goto p1;	// zu kurz
		if( a+b > 2*(samples_per_pilot*10/8) ) goto p1;	// zu lang
	}
		
/*	phase 3:	
	erwarte 2. sync-flanke
	Die erste Halbphase war <= SYNC1_MAX.
	Beide Halbphasen müssen <= 2*SYNC1_MAX sein. Zu kurz gibt es nicht.
*/
	get_next_nibble();
	if( a+b > samples_per_sync12_max ) goto p1;	// zu lang
	
/*	phase 4:	
	decodiere Daten
*/
	if (e>q+0x100022) e = q+0x100022;
	char bu[0x10002];
	ptr z = bu;
	uchar byte;
	int bits=0;

	while( q<e )
	{
		get_next_nibble();
		get_next_nibble();
		if ( a+b > samples_per_bit11_max ) break;		// zu lang => end recording
		byte += byte + (a+b>=samples_per_bit_0+samples_per_bit_1);
		if (++bits==8) { bits=0; *z++ = byte; }
	}

// write data to file:

	if (z-bu>=2)			// mind. type + data[none] + crc.
	{
		write_ushort_z( fd, z-bu );
		write_bytes( fd, bu, z-bu );
	}
	else
	{
		ClearError();		// nothing written but ok
	}
}



/*	get minor block info 
	returned cstr is const or temp
*/
cstr RlesData::MinorBlockInfo() const
{
	int blen = Max( 0, int(data_end-data_start)/16     -2/*type,crc*/ );
	int bpos = Max( 0, int(buffer.Pos()-data_start)/16 -1/*type*/     );

	if(recording)
	{
		if( buffer.Pos()==0 ) return "waiting for pilot";
		if( data_start==0 && buffer.Pos()<=512 ) return "synchronizing on pilot";
		if( data_start==0 ) return "synchronized. waiting for data";
		if( data_end==0 ) return Using( "recording byte %i", bpos );
		return "no signal";									// silence
	}
	else	// playback:	
	{
		if( data_end && buffer.Pos() >= data_end ) return "";	// post-data silence  ((gap))
		if( minor_block_info ) return minor_block_info;			// additional header info
		if( bpos>=blen ) return "";								// end of block
		return Using( "%i/%i", bpos, blen );					// other data => print position/blocklength
	}
}


/*	get major block info
	returned cstr is const or temp
*/
cstr RlesData::MajorBlockInfo() const
{
	XXTRAP(recording);	//	if( recording ) return "Recording";		// security only; should not happen
	if( data_end && buffer.Pos() >= data_end ) return ""; // "No signal";	// post-data silence  ((gap))
	return TapeData::MajorBlockInfo();
}


void RlesData::CalcMajorBlockInfo ( )
{

	uint samples_per_bit_0	= int( cc_per_bit_0 / cc_per_second * samples_per_second +0.5 );
	uint samples_per_bit_1	= int( cc_per_bit_1 / cc_per_second * samples_per_second +0.5 );
	uint limit01 = samples_per_bit_0+samples_per_bit_1;

	uchar bu[20]; 
	uchar const* q = buffer.Data()+data_start;

	int data_bytes = (data_end-data_start)/16;
	int imax = Min(data_bytes,20);

	for( int i=0; i<imax; i++ ) 
	{
		uchar byte = 0;
		for( int j=0;j<8;j++ ) { byte += byte + ((uint)q[0]+(uint)q[1]>=limit01); q+=2; }
		bu[i] = byte;
	}

	SetBlockInfoFromData( data_bytes, bu );
}













