/*	Copyright  (c)	Günter Woigk 2000 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/


#ifndef Tape_h
#define	Tape_h

#import  "types.h"
#import  "DspTime.h"

										// from "DspTime.h" / "Dsp.mm":
extern	Time	ticks_per_sec;			// DSP-Konstante & Zeitbasis des Systems: samples/second
										// typically 44100.0



enum	File_Type	{ unknown_file, tap_file, tzx_file, aiff_file, rles_file };
enum	Sample_Size	{ ss_bit=0, ss_byte=1, ss_short=2, ss_sample=4, ss_stereo=8 };
typedef ulong Sample_Freq;



// ////////////////////// generic block data container /////////////////////////////


class TapeData
{
	File_Type		block_type;						// should be same as TapeFile's file_type for all blocks
	bool			is_writable;					// depends on file type and write flag for file
	bool			is_modified;
	bool			is_recording;					// set after Write(…), cleared with Read(…) and Stop()
	cstr			block_info;						// info text (NULL if not yet determined)
	Time			total_play_time;				// total play time of block
	Time			current_position;				// current record/playback position inside block

// standard handling implemented by TapeData::read(…) and write(…):
	Sample_Size		sample_size;					// depends on file type and quality setting
	Sample_Freq		sample_freq;					// depends on file type and quality setting
	ptr				buffer;							// buffer for Read(…) and Write(…)
	ulong			buffer_size;					// allocated samples
	ulong			buffer_used;					// used samples
	ulong			buffer_ptr;						// read/write sample position


private:
					TapeData		( const TapeData& );	// prohibit
	TapeData&		operator=		( const TapeData& );	// prohibit
	
protected:
					TapeData		( File_Type, Sample_Size, Sample_Freq );

public:
	virtual			~TapeData		();
	
	File_Type		BlockType		()				{ return block_type; }
	bool			IsWritable		()				{ return is_writable; }
	bool			IsModified		()				{ return is_modified; }
	Time			TotalPlayTime	()				{ return total_play_time; }
	Time			CurrentPosition	()				{ return current_position; }
	virtual cstr	BlockInfo		();
	virtual cstr	MinorBlockInfo	();
	void			SetBlockInfo	( cstr s )		{ if(block_info!=s) { delete[] block_info; block_info = NewCopy(s); } }

// writing/reading with system ticks_per_sec:						// ((44100 Hz))
	long			Write			( Sample* buffer, long n );		// returns samples written: remainder are gap/next block
	long			Read			( Sample const* buffer, long n );// returns samples read: remainder are beyond end of block
	long			Stop			()				{ is_recording = no; }

// reading from / writing to file:
	virtual OSErr	WriteToFile		( int fd );
	virtual OSErr	ReadFromFile	( int fd );
};


// ////////////////////// .tap file /////////////////////////////

class TapData : TapeData
{
public:
					TapData			()				: TapeData(tap_file,ss_bit,ticks_per_sec/2){}
	virtual OSErr	WriteToFile		( int fd );
	virtual OSErr	ReadFromFile	( int fd );
};


// ////////////////////// .tzx file /////////////////////////////

class TzxData : TapeData
{
};


// ////////////////////// .aiff file /////////////////////////////

class AiffData : TapeData
{
};


// ////////////////////// .rles file /////////////////////////////

class RlesTapeData : TapeData
{
};



// //////////////////////////////////////////////////////////////
// ////////////////////// tape file  ////////////////////////////
// //////////////////////////////////////////////////////////////

class TapeFile
{
	cstr		file_path;
	File_Type	file_type;
	bool		is_writable;
	bool		is_write_protected;
	bool		is_modified;
	OSErr		file_error;

	TapeData**	block;
	int			max_blocks;
	int			used_blocks;
	int			current_block;
	
	
public:
				TapeFile			();
				~TapeFile			();
				TapeFile			( const TapeFile& );	// prohibit
	TapeFile&	operator=			( const TapeFile& );	// prohibit

	OSErr		Open				( cstr file_path );
	OSErr		Sync				();
	OSErr		Close				();
	OSErr		Revert				();				// only possible for file types which work on cached buffer

	cstr		FilePath			()				{ return file_path; }
	File_Type	FileType			()				{ return file_type; }
	bool		IsWritable			()				{ return is_writable; }
	bool		IsWriteProtected	()				{ return is_write_protected; }
	bool		IsModified			()				{ return is_modified; }
	OSErr		FileError			()				{ return file_error; }

//	Sample_Size	SampleSize			();
//	void		SetSampleSize		( Sample_Size );
//	Sample_Freq	SampleFrequency		();
//	void		SetSampleFrequency	( Sample_Freq );

	void		Play				( Sample const* buffer, int count );
	void		Record				( Sample* buffer, int count );
	void		Stop				();

	Time		TotalPlayTime		();
	Time		CurrentPosition		();
	void		SeekPosition		( Time );
	bool		IsAtEndOfFile		();
	bool		IsAtStartOfFile		();

	int			TotalBlocks			()				{ return used_blocks; }
	int			CurrentBlock		()				{ return current_block; }
	Time		StartOfBlock		();
	Time		EndOfBlock			();

	bool		IsAtStartOfBlock	();
	bool		IsAtEndOfBlock		();
	bool		IsNearStartOfBlock	( Time proximity = 1.0 );
	bool		IsNearEndOfBlock	( Time proximity = 1.0 );

	void		SeekBlock			( int n );
	void		SeekNextBlock		();				// goto start of next block (if not already eof)
	void		SeekPrevBlock		();				// goto start of this block or prev block
	void		SeekStartOfBlock	();
	void		SeekEndOfBlock		();
	void		SeekStartOfNextBlock();
	void		SeekEndOfPrevBlock	();
	void		SeekNextGap			();
	void		SeekPrevGap			();

	void		TruncateBlock		();
	void		JoinWithNextBlock	();
	void		SplitBlock			();
	void		DeleteBlock			( int n );
	void		InsertBlock			( int n );

	cstr		BlockInfo			();				// e.g. "Program: <name>"
	cstr		MinorBlockInfo		();				// e.g. "246/30000 bytes"
	void		SetBlockInfo		( cstr );
};






#endif

















