/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	class TapeFile: representation for program tapes
*/


#define SAFE	3
#define	LOG 	2

#include <fcntl.h>
#include <unistd.h>
#import  "unix/file_utilities.h"
#import  "cstrings/cstrings.h"
#import  "TapeFile.h"
#import  "RlesData.h"
#import  "Dsp.h"
#import  "AlertPanel.h"





/*	create TapeFile instance
	creates file or loads data
	using ReadFile() 
*/
TapeFile::TapeFile ( cstr path )
{
	XLogIn("new TapeFile(\"%s\")",path);
	
	state			= stopped;
	writable		= no;

	file_path		= NULL;
	fd				= -1;

	file_type		= (FileType)Peek4X(CatStr(LowerStr(ExtensionFromPath(path)),"  "));
	data_type		= unknown_data;
	
	if(RlesData::CanRead(file_type))
	{
		data_type	= rles;
		writable	= RlesData::CanWrite(file_type);
	}
	else
	{
		file_type = unknown_file;
		ForceError(filetypeerror); 
		return;
	}
	
	file_path		= NewCopy(path);
	fd				= open( path, writable?O_RDWR|O_CREAT:O_RDONLY, 0664 );
	writable		= fd>0 && writable;
	if( fd<0 ) fd	= open( path, O_RDONLY, 0664 );	
	if( fd<0 ) 
	{
		block.Purge(); 
		block.Append(new_block());
	}
	else
	{
		ReadFile();
	}
	if( Error() ) ShowAlert( "TapeFile:TapeFile(\"%s\"): %s",path,ErrorStr());
}


/*	destroy TapeFile instance
	save changes if is_modified is set.
*/
TapeFile::~TapeFile()
{		
	XLogIn("delete TapeFile(\"%s\")",file_path);

	Stop();		// -> prevent Play() and Record() on sound interrupt
	
	if(writable)
	{
		if(state>=recording) Stop();

		bool modified=no;
		for(uint i=0;i<block.Size()&&!modified;i++) { modified = block[i]->IsModified(); }

		if (modified) 
		{
			WriteFile();
			if(Error()) ShowAlert("Write file \"%s\" failed: %s\n",file_path,ErrorStr());
		}
	}

	if(fd>0) 
	{
		int n = close(fd);
		if (n==-1) ShowAlert("Close file \"%s\" failed: %s\n",file_path,ErrorStr());
	}

	delete[] file_path;
	for( uint i=0; i<block.Size(); i++ ) { delete block[i]; }
}

	

//	==============================
//	=========== QUERIES ==========
//	==============================
 

/*	Get position of start of current block
*/
Time TapeFile::StartOfBlock()
{
	Time t = 0.0;
	for( uint i=0; i<block.Pos(); i++ ) t += block[i]->TotalPlaytime();
	return t;
}


/*	Get position of end of current block
*/
Time TapeFile::EndOfBlock()
{
	return StartOfBlock() + block.Peek()->TotalPlaytime();
}


/*	Get current position on tape
*/
Time TapeFile::CurrentPosition()
{
	return StartOfBlock() + block.Peek()->CurrentPosition();
}


/*	Get total play time of tape
*/
Time TapeFile::TotalPlaytime()
{
	Time t = 0.0;
	for( uint i=0; i<block.Size(); i++ ) t += block[i]->TotalPlaytime();
	return t;
}


/*	Get major block info.
	This is either "Recording.." while recording,
	or decoded from the block data, 
	or supplied from info data within the file,
	or a user-set value
*/
cstr TapeFile::MajorBlockInfo()
{
	if(state>=recording) return "Recording";
	if(IsAtEndOfTape()) return "End of tape";
	TapeData* current_block = block.Peek();
	cstr info = current_block->MajorBlockInfo(); if(info) return info;
	current_block->CalcMajorBlockInfo();
	info = current_block->MajorBlockInfo(); if(info) return info;
	
	return "No major block info (TODO)"; // "MajorBlockInfo: TODO";
}


/*	Get minor block info.
	This is either the current time offset within the block,
	or the current byte read/written.
	For header blocks this should be some decoded info.
*/
cstr TapeFile::MinorBlockInfo()
{
	return block.Peek()->MinorBlockInfo();
}





//	==============================
//	=========== ACTIONS ==========
//	==============================


TapeData* TapeFile::new_block()
{
	switch(data_type)
	{
	case rles:	return new RlesData;
	default:	TODO(); return 0;		// TapeFile::new_block():  other data types
	}
}


/*	read data from file
		read from file fd
		create data blocks acc. to data_type
		decode acc. to file_type
		split into blocks (done by TapeData)
		assert one final, empty block
		rewind tape
	ok:		errno cleared
	error:	errno set, some blocks may be loaded
*/
void TapeFile::ReadFile()
{
	XLogIn("TapeFile:ReadFile");
	Stop();

	if(0) switch(file_type)
	{
	case tap_file:		break;		// no header to skip
	case tzx_file:		break;		// header is skipped by TapeData::ReadFromFile()
	case rles_file:		break;		// header is skipped by TapeData::ReadFromFile()	
	default:			break;
	}
	
	if( block.Size() )				// just in case..
	{
		for( uint i=0; i<block.Size(); i++ ) delete block[i]; 
		block.Purge();				// clear block[]; reset Pos()
	}

r:	TapeData* p = new_block();
	block.Append( p );					
	long n = p->ReadFromFile(fd,file_type);		// -1=error, 0=eof, >0=block read
	XXXASSERT( (Error()!=0) == (n==-1) );
	//XXXASSERT( (n<=0) == (p->TotalPlaytime()<=0.0) );		
	if( n==-1 ) 
	{
		cstr s = Using( "zxsp:TapeFile:ReadFile: %s\n",ErrorStr() );
		LogLine(s); 
		ShowAlert(s);
		p->Purge();
	}
	else if( p->TotalPlaytime()>0.0 ) { XXLogLine("play time = %1.3f",(float)p->TotalPlaytime()); goto r; }
	else { XXLogLine("ok."); };		// finished loading ok

	if( block.Size()>1 && n==0 ) { delete p; block.Resize(block.Size()-1); }

	XXXASSERT(block.Pos()==0);		// rewind..
	block[0]->SeekStart();			// ..tape
}


/*	write data to file
		write to file fd
		encode acc. to file_type
	ok:		errno cleared
	error:	errno set
*/
void TapeFile::WriteFile()
{	
	XLogIn("TapeFile:WriteFile");
	Stop();

	if(!writable) { ForceError(filenotwritable); return; }

	ClearError();
	if(lseek(fd,0,SEEK_SET)==-1) return;

	try
	{
		switch(file_type)
		{
		case tap_file:		break;				// no header
		case tzx_file:		write_bytes(fd,"ZXTape!\x1a\x01\x0c",10); break;	// version 1.12 header
		case rles_file:		write_bytes(fd,"RlesTape1.0",12); break;			// version 1.0  header
		default:			TODO(); break;		// WriteFile() with other file types
		}
	}
	catch(file_error){return;}

	for( ulong n=0; NoError() && n<block.Size(); n++ ) { block[n]->WriteToFile(fd,file_type); }
	if( NoError() ) { off_t fpos = lseek(fd,0,SEEK_CUR); if(fpos!=-1) ftruncate(fd,fpos); }
}


/*	play tape:
	read samples from tape file buffers
	*** executed on the sound interrupt ***
	does not overwrite but merge the sound into the destination buffer (( buffer[i++] += tape_sample ))
	when not playing, then just no sound is added
*/
void TapeFile::Play ( Sample* buffer, int count )
{
	if( state!=playing ) return;
	XXXASSERT(block.Avail());
	
	for(;;)
	{
		int n = block.Peek()->Play(buffer,count);
		if( n>=count ) return;
		if(block.Avail()<=1) return;		// don't skip beyond last block
		block.Pos()++;
		buffer+=n; count-=n;
	}
}

 
/*	record tape:
	write samples into the tape file buffers
	*** executed on the sound interrupt ***
*/
void TapeFile::Record ( Sample const* buffer, int count )
{
	XXXASSERT(block.Avail());

#define	MIN_DATA_SIGNAL_LEVEL	0.05
#define	MAX_SAMPLES_PER_PHASE	( samples_per_second / 1000.0 )		/* ~ 500 Hz */
#define MAX_POST_GAP_TIME		2.50

	if(state<recording) return;		// not recording

// High-pass filter is now part of DSP audio output proc:
//	Sample z[count];
//	center = DSP::HighPassFilter ( z, buffer, count, center, 0.001 /*damping TODO*/ );
//	buffer = z;

	switch(state)
	{
	case stopped:			// not recording.
	case playing: return;	// not recording.

	case recording_audio:	// record audio: high or low level, long half phases
		TODO();				// recording audio is an option for AIFF or WAV files. Currently not implemented.

	case recording_pre_gap:	
	case recording_post_gap:

	// pre-gap silence:
		// low signal level OR		((silence))
		// phase too long			((noise))
		//
		// The pre-gap silence is always skipped and never recorded.
		// To include the preceeding silence in a tape file recording, 
		// start recording the first block with state = recording_post_gap
		// or setup the first block manually.

	// post-gap silence: 
		// low signal level OR		((silence))	
		// phase too long			((noise))
		//
		// The post-gap silence signal starts with a high signal level after the last data signal half phase,
		// but this phase will be too long.
		// Typically only max. 2 sec trailing silence should be recorded.

		// Die folgende Suchfunktion sucht nur _innerhalb_ eines Buffers und dort 
		// nach zwei aufeinander folgende Halbphasen, 
		//	  -	die von der Länge her passen (nicht zu lang) und
		//	  -	die eine ausreichende Amplitude haben.
		//		statt jeweils abs(sample) >= MIN_DATA_SIGNAL_LEVEL wird auf
		//		s_hi-s_lo >= 2*MIN_DATA_SIGNAL_LEVEL getestet, so dass auch
		//		das einschwingende Signal nach DSP::HighPassFilter() bereits erkannt wird.
		// Prinzipiell wird also der Signalstart um bis zu 2 Halbphasen zu spät erkannt,
		// wenn er kurz vor Ende des Buffers einsetzt.

rpg:	{
			Sample const* p = buffer;			// p = running pointer
			Sample const* e = p + count;		// e = end of buffer

		// skip initial partial phase
			if(hi_phase) while( p<e && *p>=0.0 ) p++;	
			else		 while( p<e && *p< 0.0 ) p++;

		// search for 2 valid half phases:
			int n = 0;							// counter for valid half phases
			Sample const* a = p;				// start of 1st half phase 
			Sample const* b = p;				// start of last half phase
			Sample s_lo=0.0,s_hi=0.0;			// max. elongation
			
			while( p<e && n<=2 )
			{
				b = p;							// start of current half phase
				
				if( *p<0.0 ) { s_lo=0.0; while( p<e && *p< 0.0 ) { s_lo = Min(*p++,s_lo); } }
				else		 { s_hi=0.0; while( p<e && *p>=0.0 ) { s_hi = Max(*p++,s_hi); } }

				if( p-b > MAX_SAMPLES_PER_PHASE ) { n=0; a=p; continue; }		// phase too long => noise
				
				if( n==1 && s_hi-s_lo < 2*MIN_DATA_SIGNAL_LEVEL ) { a=p; continue; }	// not loud enough => silence
				
				n++;							// potential valid half phase for data signal
			}

		// if no start of data signal found:
			if( p>=e )				
			{
				if( state==recording_post_gap ) 
				{
					if( post_gap_length < MAX_POST_GAP_TIME ) 
					{ block.Peek()->Record(buffer,count,silence); }		// record post data silence only
					post_gap_length += count / samples_per_second;
				}
				hi_phase = *b >= 0.0;
				phase_samples = p-b;
				break;				// switch(state)
			}

		// start of data signal found at a:

			if( state==recording_post_gap ) 
			{
				if( post_gap_length < MAX_POST_GAP_TIME ) 
				{ block.Peek()->Record( buffer, a-buffer, silence ); }	// record post data silence only
				post_gap_length += (a-buffer) / samples_per_second;
			}
		
			hi_phase = *a >= 0.0;
			phase_samples = 0;
			count -= a-buffer;					// step back to a
			buffer = a;

			if( state==recording_post_gap )		// trailing silence => record data into next block
			{
				block.Read()->Stop();			// stop & skip current block
				if( block.AtEnd() ) block.Append(new_block());
			}
			
			block.Peek()->SeekStart();
			block.Peek()->StartRecording();
			state  = recording_data;
			goto rd;							// record data
		}

	case recording_data:	
rd:	// data signal:
		// high signal level						not tested ((could be tested to detect drop outs))
		// fast zero crossing						not tested
		// duration of half phases within limit		tested: samples <= MAX_SAMPLES_PER_PHASE
		{
		// find end of data, if within this block:
			Sample const* p = buffer;				// p = running pointer
			Sample const* e = p + count;			// e = end of buffer
			Sample const* a = p - phase_samples;	// a = start of current phase
			if(hi_phase) goto d1; else goto d0;
			
			while( p<e )
			{
		d1:		while( p<e && *p>=0.0 ) { p++; } 
				if( (p-a)>MAX_SAMPLES_PER_PHASE ) goto dx; else a=p;
		d0:		while( p<e && *p< 0.0 ) { p++; }
				if( (p-a)>MAX_SAMPLES_PER_PHASE ) goto dx; else a=p;
			}

		// no end of data signal found within buffer.
		// whole buffer is data signal:
			block.Peek()->Record(buffer,count,data_signal);
			hi_phase = *a>=0.0;
			phase_samples = p-a;
			break;	// switch(state)

		// end of data signal detected within buffer.
		// silence (non-data) starts at a:
		dx:	if(a>buffer) block.Peek()->Record(buffer,a-buffer,data_signal);	// record remaining data signal
			;;;																// TODO: set marker for data end in current block
			if(a<buffer) a=buffer;

			state = recording_post_gap;
			post_gap_length = 0.0;
			hi_phase = *a>=0.0;
			phase_samples = 0;
			count -= (a-buffer); 
			buffer = a; 
			goto rpg;				// record post-gap silence
		}
	}
}	
	

/*	record tape:
	write samples into the tape file buffers
	*** executed on the sound interrupt ***
*/
void TapeFile::Record ( StereoSample const* buffer, int count )
{
	XXXASSERT(block.Avail());

#define	MIN_DATA_SIGNAL_LEVEL	0.05
#define	MAX_SAMPLES_PER_PHASE	( samples_per_second / 1000.0 )		/* ~ 500 Hz */
#define MAX_POST_GAP_TIME		2.50

	if(state<recording) return;		// not recording

// High-pass filter is now part of DSP audio output proc:
//	Sample z[count];
//	center = DSP::HighPassFilter ( z, buffer, count, center, 0.001 /*damping TODO*/ );
//	buffer = z;

	switch(state)
	{
	case stopped:			// not recording.
	case playing: return;	// not recording.

	case recording_audio:	// record audio: high or low level, long half phases
		TODO();				// recording audio is an option for AIFF or WAV files. Currently not implemented.

	case recording_pre_gap:	
	case recording_post_gap:

	// pre-gap silence:
		// low signal level OR		((silence))
		// phase too long			((noise))
		//
		// The pre-gap silence is always skipped and never recorded.
		// To include the preceeding silence in a tape file recording, 
		// start recording the first block with state = recording_post_gap
		// or setup the first block manually.

	// post-gap silence: 
		// low signal level OR		((silence))	
		// phase too long			((noise))
		//
		// The post-gap silence signal starts with a high signal level after the last data signal half phase,
		// but this phase will be too long.
		// Typically only max. 2 sec trailing silence should be recorded.

		// Die folgende Suchfunktion sucht nur _innerhalb_ eines Buffers und dort 
		// nach zwei aufeinander folgende Halbphasen, 
		//	  -	die von der Länge her passen (nicht zu lang) und
		//	  -	die eine ausreichende Amplitude haben.
		//		statt jeweils abs(sample) >= MIN_DATA_SIGNAL_LEVEL wird auf
		//		s_hi-s_lo >= 2*MIN_DATA_SIGNAL_LEVEL getestet, so dass auch
		//		das einschwingende Signal nach DSP::HighPassFilter() bereits erkannt wird.
		// Prinzipiell wird also der Signalstart um bis zu 2 Halbphasen zu spät erkannt,
		// wenn er kurz vor Ende des Buffers einsetzt.

rpg:	{
			StereoSample const* p = buffer;			// p = running pointer
			StereoSample const* e = p + count;		// e = end of buffer

		// skip initial partial phase
			if(hi_phase) while( p<e && *p>=0.0 ) p++;	
			else		 while( p<e && *p< 0.0 ) p++;

		// search for 2 valid half phases:
			int n = 0;							// counter for valid half phases
			StereoSample const* a = p;				// start of 1st half phase 
			StereoSample const* b = p;				// start of last half phase
			Sample s_lo=0.0,s_hi=0.0;			// max. elongation
			
			while( p<e && n<=2 )
			{
				b = p;							// start of current half phase
				
				if( *p<0.0 ) { s_lo=0.0; while( p<e && *p< 0.0 ) { s_lo = Min((p++)->Mono(),s_lo); } }
				else		 { s_hi=0.0; while( p<e && *p>=0.0 ) { s_hi = Max((p++)->Mono(),s_hi); } }

				if( p-b > MAX_SAMPLES_PER_PHASE ) { n=0; a=p; continue; }		// phase too long => noise
				
				if( n==1 && s_hi-s_lo < 2*MIN_DATA_SIGNAL_LEVEL ) { a=p; continue; }	// not loud enough => silence
				
				n++;							// potential valid half phase for data signal
			}

		// if no start of data signal found:
			if( p>=e )				
			{
				if( state==recording_post_gap ) 
				{
					if( post_gap_length < MAX_POST_GAP_TIME ) 
					{ block.Peek()->Record(buffer,count,silence); }		// record post data silence only
					post_gap_length += count / samples_per_second;
				}
				hi_phase = *b >= 0.0;
				phase_samples = p-b;
				break;				// switch(state)
			}

		// start of data signal found at a:

			if( state==recording_post_gap ) 
			{
				if( post_gap_length < MAX_POST_GAP_TIME ) 
				{ block.Peek()->Record( buffer, a-buffer, silence ); }	// record post data silence only
				post_gap_length += (a-buffer) / samples_per_second;
			}
		
			hi_phase = *a >= 0.0;
			phase_samples = 0;
			count -= a-buffer;					// step back to a
			buffer = a;

			if( state==recording_post_gap )		// trailing silence => record data into next block
			{
				block.Read()->Stop();			// stop & skip current block
				if( block.AtEnd() ) block.Append(new_block());
			}
			
			block.Peek()->SeekStart();
			block.Peek()->StartRecording();
			state  = recording_data;
			goto rd;							// record data
		}

	case recording_data:	
rd:	// data signal:
		// high signal level						not tested ((could be tested to detect drop outs))
		// fast zero crossing						not tested
		// duration of half phases within limit		tested: samples <= MAX_SAMPLES_PER_PHASE
		{
		// find end of data, if within this block:
			StereoSample const* p = buffer;				// p = running pointer
			StereoSample const* e = p + count;			// e = end of buffer
			StereoSample const* a = p - phase_samples;	// a = start of current phase
			if(hi_phase) goto d1; else goto d0;
			
			while( p<e )
			{
		d1:		while( p<e && *p>=0.0 ) { p++; } 
				if( (p-a)>MAX_SAMPLES_PER_PHASE ) goto dx; else a=p;
		d0:		while( p<e && *p< 0.0 ) { p++; }
				if( (p-a)>MAX_SAMPLES_PER_PHASE ) goto dx; else a=p;
			}

		// no end of data signal found within buffer.
		// whole buffer is data signal:
			block.Peek()->Record(buffer,count,data_signal);
			hi_phase = *a>=0.0;
			phase_samples = p-a;
			break;	// switch(state)

		// end of data signal detected within buffer.
		// silence (non-data) starts at a:
		dx:	if(a>buffer) block.Peek()->Record(buffer,a-buffer,data_signal);	// record remaining data signal
			;;;																// TODO: set marker for data end in current block
			if(a<buffer) a=buffer;

			state = recording_post_gap;
			post_gap_length = 0.0;
			hi_phase = *a>=0.0;
			phase_samples = 0;
			count -= (a-buffer); 
			buffer = a; 
			goto rpg;				// record post-gap silence
		}
	}
}	
	



/*	stop tape:
	stop recording.
*/
void TapeFile::Stop()
{
 	if( state==stopped ) return;
	XXXASSERT(block.Avail());

	block.Peek()->Stop();
	state = stopped;
}


/*	start playing:
*/
void TapeFile::StartPlaying()
{
	if( state==playing ) return;
	XXXASSERT(block.Avail());

	block.Peek()->StartPlaying();
	state = playing;
}


/*	start recording:
	rewind block and record into this block 
	except if near end of block then record into the next block.
*/
void TapeFile::StartRecording()
{
	if( state>=recording ) return;
	XXXASSERT(block.Avail());
	
	if( IsNearEndOfBlock(1.5) && block.Peek()->TotalPlaytime()>=3.0 ) 
	{
		block.Pos()++;
		if(block.AtEnd()) block.Append(new_block());
	}

	//block.Peek()->SeekStart();			// done when switching to state = recording_data
	//block.Peek()->StartRecording();		// ""
	state = recording_pre_gap;
	
	hi_phase = no;			// current phase := low phase		((note: rles_data.buffer starts with low phase))
	phase_samples = 0;
	post_gap_length = 0.0;
	//center = 0.0;
}



/*	Move tape to selected position
*/
void TapeFile::SeekPosition ( Time t )
{
	Stop();

	Time a = StartOfBlock();		 while( t<a  && block.Pos()    ) { block.Pos()--; a -= PlaytimeOfBlock(); }
	Time e = a+PlaytimeOfBlock(); while( t>=e && block.Avail(2) ) { e += PlaytimeOfBlock(); block.Pos()++; }
	block.Peek()->SeekPosition( t-a );
}


/*	Move tape position to the start of the selected block
*/
void TapeFile::SeekBlock ( int n )
{
	Stop();
	
	Limit(0,n,(int)block.Size()-1);
	block.Pos() = n;
	block[n]->SeekStart();
}


/*	Move tape position to the start of the current block
*/
void TapeFile::SeekStartOfBlock()
{
	Stop();

	block.Peek()->SeekStart(); 
}


/*	Move tape position to the end of the current block
*/
void TapeFile::SeekEndOfBlock()
{
	Stop();

	block.Peek()->SeekEnd();
}


/*	Move tape position to the start of the next block.
	except if already in last block, then append empty block
	except if last block is already empty, then do nothing
*/
void TapeFile::SeekStartOfNextBlock()
{
	Stop();

	if( block.Avail(2) )
	{
		block.Pos()++;
		block.Peek()->SeekStart(); 
	}
	else if( block.Peek()->TotalPlaytime()>0.0 )
	{
		block.Pos()++;
		block.Append( new_block() );
	}
}


/*	Move tape position to the end of the previous block.
	Must not be called if already in first block.
*/
void TapeFile::SeekEndOfPrevBlock()
{
	Stop();

	XXXASSERT( block.Pos()>0 );

	block.Pos()--;
	block.Peek()->SeekEnd();
}


/*	Truncate current block at current tape position.
	Keep the tape position inside the current block.
*/
void TapeFile::TruncateBlock() { Stop(); TODO(); }


/*	Join this and the next block.
	Keep the tape position inside the current block.
*/
void TapeFile::JoinWithNextBlock() { Stop(); TODO(); }


/*	Split block at current tape position.
	Set the tape position to the start of the second block.
*/
void TapeFile::SplitBlock() { Stop(); TODO(); }


/*	Delete the current block.
	Set the tape position to the start of the next block.
	Except if this is the last block, then just purge it.
*/
void TapeFile::DeleteBlock()
{
	Stop();

	if( block.Avail(2) ) 
	{
		delete block.Peek();				// delete the block
		block.Delete(block.Pos(),1);		// close gap in block[]
		block.Peek()->SeekStart();
	}
	else
	{
		block.Purge();
	}
}


/*	Insert an empty block at the position of the current block.
	Set the tape position to the start of this block. (implicitely, new block is empty)
*/
void TapeFile::InsertBlock()
{
	Stop();

	block.Insert(block.Pos(),1);			// create gap in block[]
	block.Peek() = new_block();				// create new block in gap
}


/*	Store user-defined block info
*/
void TapeFile::SetMajorBlockInfo ( cstr info )
{
	block.Peek()->SetMajorBlockInfo(info);
}




















