/*	Copyright  (c)	Günter Woigk 2000 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.


	class TapeFile
	--------------
	
	Representant for tape files inserted into the tape recorder.
	
	The TapeFile class is not reentrant.
	The user must take care for proper blocking,
	because (at least) Record() and Play() are called from 
	a different thread (the core audio io thread).
*/


#ifndef TapeFile_h
#define	TapeFile_h


#import  "templates/Buffer.h"
#import  "TapeData.h"


class TapeFile
{
	int			fd;
	cstr		file_path;
	FileType	file_type;
	Data_Type	data_type;
	bool		writable;			// depends on file type and OS write protection

	// current tape state:
	// basically stopped, playback or recording
	// but recording is further refined into recording preceeding silence, audio, data and trailing silence
	// 
	enum State	{	stopped,						// neither playing back nor recording
					playing,						// playing back
					recording,						// recording: if state ≥ recording					
					recording_pre_gap = recording,	// recording (or ignoring) silence before signal
					recording_audio,				// recording (or ignoring) audio signal
					recording_data,					// recording data signal
					recording_post_gap };			// recording (or ignoring) silence after signal
 
	State		state;
	bool		hi_phase;			// for state analysis while recording
	int			phase_samples;		// ""
	Time		post_gap_length;	// ""
//	Sample		center;				// ""

	Buffer<TapeData*> block;

	TapeData*	new_block			();

				TapeFile			( const TapeFile& );		// prohibit
	TapeFile&	operator=			( const TapeFile& );		// prohibit
	
public:
				TapeFile			( cstr file_path );
				~TapeFile			();

	void		WriteFile			();
	void		ReadFile			();

	cstr		FilePath			()							{ return file_path; }
	FileType	GetFileType			()							{ return file_type; }
	bool		IsWritable			()							{ return writable; }
	bool		IsRecording			()							{ return state>=recording; }
 
// record/playback with system samples_per_second
	void		Play				( Sample* buffer, int count );
	void		Record				( Sample const* buffer, int count );
	void		Record				( StereoSample const* buffer, int count );
	void		Stop				();
	void		StartRecording		();
	void		StartPlaying		();

	Time		TotalPlaytime		();
	Time		CurrentPosition		();
	void		SeekPosition		( Time );
	bool		IsAtEndOfTape		()							{ return IsLastBlock()  && IsAtEndOfBlock(); }
	bool		IsAtStartOfTape		()							{ return IsFirstBlock() && IsAtStartOfBlock(); }
	bool		IsNearEndOfTape		( Time proximity = 1.0 )	{ return IsLastBlock()  && IsNearEndOfBlock(proximity); }
	bool		IsNearStartOfTape	( Time proximity = 1.0 )	{ return IsFirstBlock() && IsNearStartOfBlock(proximity); }

// block handling:
	int			TotalBlocks			()							{ return block.Size(); }
	int			CurrentBlock		()							{ return block.Pos(); }
	Time		PlaytimeOfBlock		()							{ return block.Peek()->TotalPlaytime(); }
	Time		StartOfBlock		();
	Time		EndOfBlock			();
	bool		IsLastBlock			()							{ return block.Avail()==1; }
	bool		IsFirstBlock		()							{ return block.Pos()==0; }

	bool		IsAtStartOfBlock	()							{ return block.Peek()->IsAtStart();	}
	bool		IsAtEndOfBlock		()							{ return block.Peek()->IsAtEnd();	}
	bool		IsNearStartOfBlock	( Time proximity = 1.0 )	{ return block.Peek()->IsNearStart(proximity);	}
	bool		IsNearEndOfBlock	( Time proximity = 1.0 )	{ return block.Peek()->IsNearEnd(proximity);	}

	void		SeekBlock			( int n );					// goto start of block n
	void		SeekStartOfBlock	();
	void		SeekEndOfBlock		();
	void		SeekStartOfNextBlock();							// goto start of next block
	void		SeekEndOfPrevBlock	();							// goto end of prev block

	void		TruncateBlock		();
	void		JoinWithNextBlock	();
	void		SplitBlock			();
	void		DeleteBlock			();
	void		InsertBlock			();

	cstr		MajorBlockInfo		();							// e.g. "Program: <name>"
	cstr		MinorBlockInfo		();							// e.g. "246/30000 bytes"
	void		SetMajorBlockInfo	( cstr );
};






#endif

















