/*	Copyright  (c)	Günter Woigk 1995 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#ifndef ay_h
#define	ay_h

#import "ZxItem.h"


/* ==================================================
		ay-3-8912 emulation
		hardware independent part
=================================================== */
	

/* ----	the sound channels of the ay chip ---------------------------------------------
*/
class Channel
{
// ----	the amplifier stage of a channel ----------------------------------------------
private:
	Sample	soundvolume;	// volume of this channel if not use envelope
	Sample	envelopevolume;	// volume output of envelope generator
	Sample	volume;			// sound or envelope
	Sample	output;			// volume or 0
	bool	use_envelope;
	bool	amp_in;

	void	AmpInput		 ( bool );

public:
	void	SetEnvelopeVolume( uchar );
	void	SetEnvelopeSample( Sample );
	void	UseEnvelope		 ( bool );
	void	SetVolume		 ( uchar );
	Sample	GetOutput		 ( );
	bool	UsesEnvelope	 ( )		{ return use_envelope; }
	Sample	Volume			 ( )		{ return volume; }

// ----	the mixer stage of a channel --------------------------------------------------
private:
	bool	sound_enable;	// from mixer control register
	bool	sound_in;		// from sound generator
	bool	sound_out;		// output of gate: (sound_in||!sound_enable)
	bool	noise_enable;	// from mixer control register
	bool	noise_in;		// from noise generator
	bool	noise_out;		// output of gate: (noise_in||!noise_enable)

	void	MixSoundInput	( bool );

public:
	void	MixNoiseInput	( bool );
	void	EnableSound	 	( bool );
	void	EnableNoise	 	( bool );

// ----	the sound generator stage of a channel ---------------------------------------
private: 
	Time	time_for_cycle;
	Time	reload;			// reload value
	Time	when;			// next tick to toggle state
	uchar	fine,coarse;	// real calues for reload register

public:	
			Channel			( Time time_for_cycle );
	void	Reset			( Time now );

	void	SetFinePeriod	( uchar );
	void	SetCoarsePeriod	( uchar );
	void	Trigger			( );

	friend class Ay;
};


/* ----	the noise generator of the ay chip -------------------------------------------
*/
class Noise
{	Time	time_for_cycle;
	Time	reload;			// reload value
	Time	when;			// next tick to toggle output
	long	shiftreg;		// random number shift register; bit 0 is output
	Channel	*A,*B,*C;		// propagate output state
public:
			Noise			( Time time_for_cycle, Channel*,Channel*,Channel* );
	void	Reset			( Time now );
	void	SetPeriod		( uchar );
	void	Trigger			( );

	friend class Ay;
};


/* ----	the envelope generator of the ay chip ----------------------------------------
*/
class Envelope
{	Time	time_for_cycle;
	Time	reload;			// reload value
	Time	when;			// next tick to inc/dec volume
	uchar	fine,coarse;	// real values for reload register
	int  	index;			// current output state: volume 0 ... 15
	bool	repeat;			// repeat ramping
	bool	invert;			// toggle ramping direction
	int  	direction;		// ramping up/down/off
	Channel	*A,*B,*C;		// propagate output state

public:
			Envelope		( Time time_for_cycle, Channel*,Channel*,Channel* );
	void	Reset			( Time now );
	void	SetShape		( Time now, uchar );
	void	SetFinePeriod	( Time now, uchar );
	void	SetCoarsePeriod	( Time now, uchar );
	void	Trigger			( );

	friend class Ay;
};



/* ----	class for ay sound chip ---------------
*/
class Ay : public ZxItem
{
protected:
	ushort		select_mask;
	ushort		select_bits;
	ushort		write_mask;
	ushort		write_bits;
	Time		time_for_cycle;
	int			stereo_mix;
	
// AY sound chip registers:
	uint		AYregnr;		// selected register
	uchar		AYreg[16];		// 16 registers/ports of the AY chip

// internal state:
	Channel		A,B,C;
	Noise		N;
	Envelope	E;

// sample output:
// progress:
	Sample		volume;								// per channel: 0.0 .. 0.33
	Sample		current_value_l, current_value_r;	// current template sample ~ elongation
	Time		time_of_last_sample;				// bis zu diesem Zeitpunkt wurde schon Sound erzeugt
public:
	Ay		( Machine* machine, cstr name, cstr sel, cstr wr, cstr rd, Frequency psg_clocks_per_second );
	//~Ay		( );		


// set/get register
	// set register
	void		SetRegister		( Time, uint r, uchar n );
	void		SetRegister		( uint r, uchar n )		{ SetRegister( time_of_last_sample, r, n ); }

	// get register
	uchar const*Registers		( )						{ return AYreg; }
	uchar		GetRegister		( uint nr )				{ return AYreg[nr&0x0f]; }
	uint		RegNr			( )						{ return AYregnr; }

	// select register
	void		SelectRegister	( int nr )				{ AYregnr = nr&0x0f; }

	// read selected register
	uchar		ReadRegister	( )						{ return AYreg[AYregnr]; }

	// write to selected register 
	void		WriteRegister	( Time t, uchar n )	{ SetRegister(t,AYregnr,n); }

	void		SetVolume		( Sample v )			{ volume = v*0.33; }		
	void		SetClock		( Frequency psg_cycles_per_second );
	Frequency	GetClock		( )						{ return 1.0 / time_for_cycle; }

	// set stereo/mono:
	enum { mono,abc_stereo,acb_stereo };
	void		SetStereoMix	( int n )				{ stereo_mix = n; }
	int			GetStereoMix	( )						{ return stereo_mix; }

// IsA interface:
virtual bool	IsA				( int id ) const	{ return id==isa_Ay || ZxItem::IsA(id); }

// Item interface:
virtual	void	PowerOn			( Time, long cc );
virtual	void	Reset			( Time, long cc );
virtual void	DspBufferEnd	( Time, long cc );		
virtual	void	Output			( Time, long cc, ushort addr, uchar byte );
virtual	void	Input			( Time, long cc, ushort address, uchar& byte, uchar& mask );
};


class InternalAy : public Ay
{
public:
	InternalAy	( Machine* m, cstr sel, cstr wr, cstr rd, Frequency f )
								: Ay(m,"AY-3-8912",sel,wr,rd,f){}; 

virtual bool	IsA				( int id ) const	{ return id==isa_InternalAy || Ay::IsA(id); }
};


#endif










