/*	Copyright  (c)	Günter Woigk 1995 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.


	AY-3-8912 sound chip emulation
	------------------------------


$FFFD	%1111.----.----.--0-	ZX128 AY reg. select/read i/o
$BFFD	%1011.----.----.--0-	ZX128 AY reg. write -/o


	24.Mar.1995 kio	First revision on MacOS
	05.Jun.1995 kio	Combined ear and mic sound output
	29.Nov.1995 kio	AY-3-8912 emulation
	05.Jan.1999 kio	port to Linux
	09.jan.1999 kio	ay emu fixed, perfect sound sampling
	31.dec.1999 kio	started port back to MacOS
	13.jul.2000 kio implemented signals
*/


/* ==================================================================
		Hardware independent part
================================================================== */

#define SAFE	3
#define LOG		1

#import "Ay.h"
#import "DspTime.h"
#import "Dsp.h"
#import "Machine.h"
#import "cstrings/cstrings.h"
#import "utilities.h"

#define SendSignal(a,b,c,d) if(0){}else{}


// logarithmic volume table:
Sample logvol[16];


// valid bits masks:
const uchar AYregMask[16] =			// exising bits mask
{
 	0xff,0x0f,0xff,0x0f,0xff,0x0f, 	// channel A,B,C period
	0x1f, 							// noise period
	0xff, 							// mixer control
	0x1f,0x1f,0x1f, 				// channel A,B,C volume
	0xff,0xff,0x0f, 				// envelope period and shape
	0xff,0xff 						// i/o ports	
};


// registers after reset:
const uchar AYregReset[] = 
{ 	0xff,0x0f,0xff,0x0f,0xff,0x0f, 	// channel A,B,C period
	0x1f, 							// noise period
	0x3f, 							// mixer control
	0x00,0x00,0x00, 				// channel A,B,C volume
	0/*0xff*/,0/*0xff*/,0x00,		// envelope period and shape	note: some sq-tracker demos assume registers=0 after reset
	0xff,0xff 						// i/o ports
};



/* ===============================================================
		amplifier stage of a channel
=============================================================== */


/* ----	de/select envelope for this channel --------------------
*/
inline void Channel::UseEnvelope ( bool env )
{
	use_envelope = env;
	volume = env ? envelopevolume : soundvolume;
	if (amp_in) output = volume;
}


/* ----	set volume and select envelope --------------------------
		in: value written to the channel's volume register
		each channel has it's own amplifier.
		changing the volume takes effect immediately
		bits 0-3 define the channel's volume, which is stepped logarithmically
		and bit 4 overrides bits 0-3 selecting the envelope generator's volume
*/
void Channel::SetVolume ( uchar n )
{
	soundvolume = logvol[n&0x0f];
	UseEnvelope(n&0x10);
}


/* ----	set envelope volume -------------------------------------
		in: new sample value of envelope output
		immediately sets new volume if envelope is enabled
*/
inline void Channel::SetEnvelopeSample ( Sample n )
{
	envelopevolume = n;
	if (use_envelope) { volume = n; if (amp_in) output = n; }
}


/* ----	set envelope volume -------------------------------------
		in: new volume of envelope output
		immediately sets new volume if envelope is enabled
*/
inline void Channel::SetEnvelopeVolume ( uchar n )
{
	SetEnvelopeSample(logvol[n]);
}


/* ----	change state of binary sound input ----------------------
*/
inline void Channel::AmpInput ( bool in )
{
	amp_in = in;
	output = in ? volume : 0;
}


/* ----	get output of amplifier stage ---------------------------
*/
inline Sample Channel::GetOutput ( )
{
	return output;
}


/* ===============================================================
		mixer stage of a channel
=============================================================== */


/* ----	enable sound generation ----------------------------------
		in: !(bit_from_mixer_control_register)

	  	note:	in a real ay chip, sound output to the amplifier is 
				calculated from the mixer inputs like this:
			=>	output = (disable_sound||sound_input) && (disable_noise||noise_input)
				in this emulation we use the equivalent calculation:
	  		=>	output = (!enable_sound||sound_input) && (!enable_noise||noise_input)
*/
void Channel::EnableSound ( bool f )
{
	sound_enable = f;
	sound_out = sound_in || !f;
	AmpInput ( sound_out && noise_out );
}


/* ----	enable noise generation ----------------------------------
		in: !(bit_from_mixer_control_register)
*/
void Channel::EnableNoise ( bool f )
{
	noise_enable = f;
	noise_out = noise_in || !f;
	AmpInput ( sound_out && noise_out );
}


/* ----	change input from noise generator -------------------------
*/
inline void Channel::MixNoiseInput ( bool in )
{
	noise_in = in;

	if (noise_enable)
	{
		noise_out = in;
		AmpInput ( sound_out && noise_out );
	}
}


/* ----	change input from sound generator -------------------------
*/
inline void Channel::MixSoundInput ( bool in )
{
	sound_in = in;

	if (sound_enable)
	{
		sound_out = in;
		AmpInput ( sound_out && noise_out );
	}
}


/* ===============================================================
		sound generator stage of a channel 
=============================================================== */


/* ----	reset the whole channel ----------------------------------
		reset sound generator, mixer and amplifier
*/
void Channel::Reset ( Time now )
{
// reset counter stage
	sound_in = low;
	fine     = 0xff;
	coarse   = 0x0f;
	reload   = 0x0fff * time_for_cycle;
	when     = now + reload;

// reset mixer stage
	EnableSound(false);
	EnableNoise(false);

// reset amplifier stage
	SetVolume(0);		// volume=0 & envelope off
	SetEnvelopeSample(0);
}


/* ----	creator ---------------------------------------------------
		disable channel
		reset all stages
*/
Channel::Channel ( Time time_for_cycle )
{
	this->time_for_cycle = time_for_cycle * 8;		// incl. predivider 1/16, but flip every 1/2 cycle on/off !!
	Reset(Time(0.0));
}


/* ----	change a channel's pitch ----------------------------------
		the ay input clock is first divided by 16. then this clock is fed into a counter
		which's output is the mixer's "Soundin" state for this channel.
		programming a pitch of 0 seems to have the same effect as a pitch of 1
		each counter has 12 significant bits.

		it is not known when a new value takes effect:
		repeatedly programming the same value has no effect on the counter
		it may be a count down which restores the reload value when it reaches/underflows 0
		it may be a count up which is reset to 0 when it reaches/overlows the reload value

		the current implementation assumes that it is a count down and that the
		new value takes effect when the counter is reload
*/
void Channel::SetFinePeriod ( uchar n )
{
	fine = n;
	reload = (coarse*256+n)*time_for_cycle; 
	if (!reload) reload = time_for_cycle;
}


void Channel::SetCoarsePeriod ( uchar n )
{
	n &= 0x0f;
	coarse = n;
	reload = (n*256+fine)*time_for_cycle;
	if (!reload) reload = time_for_cycle;
}


void Channel::Trigger ( )
{
	MixSoundInput ( !sound_in );
	when += reload;
}


/* ===============================================================
			noise generator
=============================================================== */


/* ----	reset the noise generator --------------------------------
*/
void Noise::Reset ( Time now )
{	const bool s = high;

	reload   = 0x1f * time_for_cycle;
	when     = now + reload;
	shiftreg = 0x0001FFFF;
	A->MixNoiseInput(s);
	B->MixNoiseInput(s);
	B->MixNoiseInput(s);
}


/* ----	creator ---------------------------------------------------
		since the noise generator actively propagates it's output state 
		to the sound channels, it needs to know where they are
*/
Noise::Noise ( Time time_for_cycle, Channel* a, Channel* b, Channel* c )
{
	this->time_for_cycle = time_for_cycle * 16;		// incl. predivider 1/16
	A=a;
	B=b;
	C=c;
	Reset(Time(0.0));
}


/* ----	set noise counter period -----------------------------------
		in: Value written to noise period register

		there is 1 noise generator. it's output may be mixed to any channel.
		the ay input clock is first divided by 16. 
		then this clock is fed into a programmable counter.
		this counter has 5 significant bits.
		programming a pitch of 0 seems to have the same effect as a pitch of 1.
		the counter's output is fed into the clock input of a random generator, 
		which's output is the mixer's "Noisein" state for every channel.
	
		The random generator of the ay chip is a 17-bit shift register. 
		The input to the shift register is bit0 xor bit2 (where bit0 is the output).

		the implementation of counters seems to be very similar for channels and noise.
		so the following is the same as for channels:

		it is not known when a new value takes effect:
		repeatedly programming the same value has probably no effect on the counter
		it may be a count down which restores the reload value when it reaches/underflows 0
		it may be a count up which is reset to 0 when it reaches/overlows the reload value

		the current implementation assumes that it is a count down and that the
		new value takes effect when the counter is reload

		note: the MAME sources say that it is a count up
*/
inline void Noise::SetPeriod ( uchar n )
{
	n&=0x1f; n+=!n;
	reload = n * time_for_cycle;
}


/* ----	Trigger noise counter --------------------------------------	
		Trigger() must be called when according to When() the counter expired
*/
void Noise::Trigger ( )
{
// flip the gates
	long oldshiftreg = shiftreg;
	shiftreg = (oldshiftreg>>1) + ((oldshiftreg<<16)^(oldshiftreg<<14));
	when += reload;
	
// change in output state ?
	if ( (shiftreg^oldshiftreg)&1 )
	{
		bool s = shiftreg&1;
		A->MixNoiseInput(s);
		B->MixNoiseInput(s);
		C->MixNoiseInput(s);
	}
}


/* ===============================================================
			envelope generator
=============================================================== */


/* ----	internal representation of the sound chip's state -------------------------

		envelope generator:
		there is 1 envelope generator. 
		envelope shape is controlled by attack, invert and hold.
		the 'repeat' bit (D3 of reg. 13) is not used and always treated as set:
		therefore %00xx is converted to %1001 and %01xx is converted to %1111 
		which give the same results.

		D3 = repeat
		D2 = attack
		D1 = toggle
		D0 = hold

		1111, 01xx:	/_____	 repeat &  attack &  toggle &  hold	||  !repeat & attack
		1110:		/\/\/\	 repeat &  attack &  toggle & !hold
		1101:		/~~~~~	 repeat &  attack & !toggle &  hold
		1100:		//////	 repeat &  attack & !toggle & !hold
		1011:		\~~~~~	 repeat & !attack &  toggle &  hold
		1010:		\/\/\/	 repeat & !attack &  toggle & !hold
		1001, 00xx:	\_____	 repeat & !attack & !toggle &  hold	||  !repeat & !attack
		1000:		\\\\\\	 repeat & !attack & !toggle & !hold

		the envelope generator is running free and only retriggered 
		when a new envelope is selected.
*/


/* ----	reset the envelope generator -------------------------------------------
*/
void Envelope::Reset ( Time now )
{
	reload  = 0x0000ffff * time_for_cycle;
	when    = now + reload;
	fine    = coarse = 0xff;
	index   = 0;
	repeat  = invert = off;
	direction = 0;

	A->SetEnvelopeSample(0);
	B->SetEnvelopeSample(0);
	C->SetEnvelopeSample(0);
}


/* ----	creator ---------------------------------------------------------------
		since the envelope generator actively propagates it's output state 
		to the sound channels, it needs to know where they are
*/
Envelope::Envelope ( Time time_for_cycle, Channel* a, Channel* b, Channel* c )
{
	this->time_for_cycle = time_for_cycle * 16;	// incl. predivider 1/16
	A=a;
	B=b;
	C=c;
	Reset(Time(0.0));
}


/* ----	set counter low byte -------------------------------------------------
		in: byte written to envelope fine period register
*/
inline void Envelope::SetFinePeriod ( Time now, uchar n )
{
	fine = n;
	if(direction)
	{
		reload = ((long)coarse*256+n) * time_for_cycle;
		if( !reload ) reload = time_for_cycle;
		if( when>now+reload ) when = now + reload;
	}
}


/* ----	set counter high byte -------------------------------------------------
		in: byte written to envelope coarse period register
*/
inline void Envelope::SetCoarsePeriod ( Time now, uchar n )
{
	coarse = n;
	if(direction)
	{
		reload = ((long)n*256+fine) * time_for_cycle;
		if( !reload ) reload = time_for_cycle;
		if( when > now+reload ) when = now + reload;
	}
}


/* ----	select new envelope shape -------------------------------------------
		in: byte written to the envelope control register
		this automatically restarts the envelope generator
*/
void Envelope::SetShape ( Time now, uchar c )
{
	if (!(c&8)) c = c&4 ? 0x0f : 0x09;

	index	  = BIT(2,c) ? 0 : 15;
	direction = BIT(2,c) ? 1 : -1;
	invert    = BIT(1,c); 
	repeat    = !BIT(0,c);

	reload	  = ((long)coarse*256+fine) * time_for_cycle;
	if( !reload ) reload = time_for_cycle;
	when	  = now + reload;
	
	Sample v = logvol[index];
	A->SetEnvelopeSample(v);
	B->SetEnvelopeSample(v);
	C->SetEnvelopeSample(v);
}


/* ---- trigger the envelope counter ---------------------------------------
		Trigger() must be called when according to When() the counter expired
*/
void Envelope::Trigger ( )
{
	index += direction;

	if (index & 0xF0)		// ramp finished
	{
		if (repeat)
		{
			if (invert) { index = ~index; direction = -direction; }
		}
		else
		{
			direction=0; if (!invert) index = ~index;
			reload = 0xffff * time_for_cycle;
		}
		index  &= 0x0F;
	}

	when += reload;

	Sample v = logvol[index];
	A->SetEnvelopeSample(v);
	B->SetEnvelopeSample(v);
	C->SetEnvelopeSample(v);
}



/* ===============================================================
			ay sound chip 
=============================================================== */

/* ----	standard creator -----------------------------------
		all channels are disabled and must be enabled with EnableXXX()
		sound output is disabled and must be enabled with Resume()
*/
Ay::Ay ( Machine* m, cstr name, cstr s, cstr w, cstr r, Frequency freq )
:	ZxItem(m, name?name:"AY-3-8912", And(w,s), r ),
	select_mask(MaskForSpec(s)),
	select_bits(BitsForSpec(s)),
	write_mask(MaskForSpec(w)),
	write_bits(BitsForSpec(w)),
	time_for_cycle(1/freq),
	A(1/freq),
	B(1/freq),
	C(1/freq),
	N(1/freq,&A,&B,&C),
	E(1/freq,&A,&B,&C)
{
	//if(m->ay==NULL) m->ay = this;
	memcpy(AYreg,AYregReset,16);
	AYregnr				= 0;
	volume				= 0;
	current_value_l		= 0;
	current_value_r		= 0;
	time_of_last_sample	= 0;
	stereo_mix			= acb_stereo;
	
	logvol[0]	= 0.0;
	logvol[15]	= 1.0;
	for (uint i=14;i>=1;i--) logvol[i] = logvol[i+1] * 0.8;		// org: ~ 3.5 dB
	if(XXLOG) { LogIn("AY volume table:"); for (uint i=0;i<15;i++) LogLine("%f",logvol[i]);	}
	SetVolume(1.0);
}



/* ----	destructor ----------------------------------------
*/
//#import "ZonxBox.h"		// so that gcc knows that Ay is a parent class of ZonxBox
//Ay::~Ay ( )
//{
	//if( machine->ay==this) machine->ay = machine->FindItem(isa_Ay);	// other Ay chip or NULL
//}


/* ----	attach sound chip to machine ------------------------------
*/
void Ay::PowerOn ( Time now, long cc )
{
	current_value_l		= 0;
	current_value_r		= 0;
	time_of_last_sample	= now;			// bis zu diesem Cycle wurde schon Sound erzeugt
	A.when = now;
	B.when = now;
	C.when = now;
	N.when = now;
	E.when = now;

	Reset( now, cc );
}

void Ay::Reset ( Time t, long /*cc*/ )
{
	AYregnr = 7;
	for( uint i=0; i<16; i++ ) { SetRegister(t,i,AYregReset[i]); }
}


/* ----	write data to register ----------------------------------------------------
		generate sound up to now
		then modify register 
*/
void Ay::SetRegister ( Time when, uint regnr, uchar newvalue )
{
	Time now;
	int who;

	XXTRAP(regnr>=NELEM(AYreg));
	
	do 
	{
	// who is next ?
		{ who=6; now=when; }		// new command is next

		if ( N.when-now<0 ) 
			if ( ~AYreg[7]&0x38 )		// for speed...
			{ who=1; now=N.when; }	// noise is next
			else N.when=now+1000*time_for_cycle;

		if ( A.when-now<0 )
			if ( AYreg[8]&&~AYreg[7]&0x09 )
			{ who=3; now=A.when; }	// channel A is next
			else A.when=now+1000*time_for_cycle;

		if ( B.when-now<0 )
			if ( AYreg[9]&&~AYreg[7]&0x12 )
			{ who=4; now=B.when; }	// channel B is next
			else B.when=now+1000*time_for_cycle;

		if ( C.when-now<0 )
			if ( AYreg[10]&&~AYreg[7]&0x24 )
			{ who=5; now=C.when; }	// channel C is next
			else C.when=now+1000*time_for_cycle;

		if ( E.when-now<0 )	
			if ( A.UsesEnvelope()||B.UsesEnvelope()||C.UsesEnvelope() )	// for speed...
			{ who=2; now=E.when; }	// envelope is next
			else E.when=now+1000*time_for_cycle;

	// handle next:
		switch (who)
		{
		case 1: N.Trigger(); break;
		case 2: E.Trigger(); break;
		case 3: A.Trigger(); break;
		case 4: B.Trigger(); break;
		case 5: C.Trigger(); break;

		default:
		// set register:
			newvalue &= AYregMask[regnr];
		//	if (AYreg[regnr]==newvalue && regnr!=13) break;		---> eel_demo !!!
			
			switch (regnr)
			{
			case 0:  A.SetFinePeriod(newvalue);		break;
			case 1:  A.SetCoarsePeriod(newvalue);	break;
			case 2:  B.SetFinePeriod(newvalue);		break;
			case 3:  B.SetCoarsePeriod(newvalue);	break;
			case 4:  C.SetFinePeriod(newvalue);		break;
			case 5:  C.SetCoarsePeriod(newvalue);	break;
			case 6:  N.SetPeriod(newvalue);			break;
			case 7:
				{	uchar c = ~newvalue;
					uchar t = newvalue ^ AYreg[7];
					if(t&1) A.EnableSound(c&1);	if(t&8)  A.EnableNoise(c&8);
					if(t&2) B.EnableSound(c&2);	if(t&16) B.EnableNoise(c&16);
					if(t&4) C.EnableSound(c&4);	if(t&32) C.EnableNoise(c&32); 
					if(t&0xc0) {
						if( t&0x40 && AYreg[14]!=0xff ) SendSignal( port_A, this, when, newvalue&0x40 ? AYreg[14] : 0xff );
						if( t&0x80 && AYreg[15]!=0xff ) SendSignal( port_B, this, when, newvalue&0x80 ? AYreg[15] : 0xff );
					}
				}	break;
			case 8:  A.SetVolume(newvalue);			break;
			case 9:  B.SetVolume(newvalue);			break;
			case 10: C.SetVolume(newvalue);			break;
			case 11: E.SetFinePeriod(now,newvalue);		break;
			case 12: E.SetCoarsePeriod(now,newvalue);	break;
			case 13: E.SetShape(now,newvalue);		break;
			
			case 14: if (AYreg[7]&(1<<6)) { if(AYreg[14]!=newvalue) { SendSignal( port_A, this, when, newvalue ); } }
					 else continue;			// don't store value!
					 break;
			
			case 15: if (AYreg[7]&(1<<7)) { if(AYreg[15]!=newvalue) { SendSignal( port_B, this, when, newvalue ); } }
					 else continue;			// don't store value!
					 break;
			}

			AYreg[regnr] = newvalue;
		}

	// write samples to dsp
		if( now>time_of_last_sample )
		{
			DSP::OutputSamples(current_value_l,current_value_r,time_of_last_sample,now );
			time_of_last_sample = now;
		}
		switch(stereo_mix)
		{
			default:	// mono
				current_value_l = 
				current_value_r = volume * ( A.GetOutput()+B.GetOutput()+C.GetOutput() );
				break;
			case abc_stereo:
				current_value_l = volume * ( 2.0*A.GetOutput() + B.GetOutput() );
				current_value_r = volume * ( 2.0*C.GetOutput() + B.GetOutput() );
				break;
			case acb_stereo:
				current_value_l = volume * ( 2.0*A.GetOutput() + C.GetOutput() );
				current_value_r = volume * ( 2.0*B.GetOutput() + C.GetOutput() );
				break;
		}
	}	
	while(who!=6);
}


void Ay::Output ( Time t, long cc, ushort addr, uchar byte )
{
	if( (addr&select_mask)==select_bits ) SelectRegister(byte); else
	if( (addr&write_mask) == write_bits ) 
	{
		XXLogLine("Ay:Output: register %i = $%02x",(int)AYregnr,(int)byte);
		WriteRegister(t,byte);
	}
}


void Ay::Input ( Time, long cc, ushort addr, uchar& byte, uchar& mask )
{
		XXLogLine("Ay:Input: register %i = $%02x",(int)AYregnr,(int)ReadRegister());

		byte &= ReadRegister(); 
		mask= 0xff;
}


void Ay::DspBufferEnd ( Time t/*seconds*/, long cc )
{
	SetRegister( t, 0, GetRegister(0) );
	time_of_last_sample -= seconds_per_dsp_buffer;
	A.when -= seconds_per_dsp_buffer;
	B.when -= seconds_per_dsp_buffer;
	C.when -= seconds_per_dsp_buffer;
	N.when -= seconds_per_dsp_buffer;
	E.when -= seconds_per_dsp_buffer;
}


/*	Eingangsfrequenz ändern.
*/
void Ay::SetClock ( Frequency psg_cycles_per_second )
{
	double freq_factor = psg_cycles_per_second * time_for_cycle;
	time_for_cycle      = 1/psg_cycles_per_second;
	Time now         = time_of_last_sample;
	
	A.when = now + (A.when-now) / freq_factor;	A.time_for_cycle /= freq_factor; A.reload /= freq_factor;
	B.when = now + (B.when-now) / freq_factor;	B.time_for_cycle /= freq_factor; B.reload /= freq_factor;
	C.when = now + (C.when-now) / freq_factor;	C.time_for_cycle /= freq_factor; C.reload /= freq_factor;
	N.when = now + (N.when-now) / freq_factor;	N.time_for_cycle /= freq_factor; N.reload /= freq_factor;
	E.when = now + (E.when-now) / freq_factor;	E.time_for_cycle /= freq_factor; E.reload /= freq_factor;
}








