/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		1

#import "KempstonMouseInterfacePanel.h"
#import "KempstonMouseInterface.h"
#import "kio/kio.h"
#import "DisplayList.h"
#import "Machine.h"
#import "NXTextField.h"
#import "Mouse.h"


/*	Kempston Mouse Interface emulation

Notes on mouse grabbing:
	The mouse is made invisible and locked to the "Grab mouse" button.
	In OS X 10.3.9 the mouse may not correctly be moves over the button.
	I could not find a official conversion method for window to local screen to global screen coordinates.
	Getting the display bounds works differently in OS X 10.3.9 and OS X 10.4:
		10.4 returns the bounds as set for the desktop
		10.3.9 returns the current bounds, as currently set for fullscreen mode.
	Clicking the mouse is catched and forwarded to the topmost window in 10.4
	but it goes simply through the fullscreen and hits whatever there is, frequently moving the app into the background.
	Klicks seem never be forwarded to any window in 10.3.9.
	many views order mouseMoved events and don't cancel the request later, resulting into huge cpu usage. 
	(frame rate down to 3% on my iBook!)
	And lots more garbage which i have forgotten right now.
	
	To circumvent all this, patches throughout the app are applied:
	
	MainApp:sendEvent:
	This is the first entry point for events and they are dispatched here to my windows and views:
		if the mouse is grabbed, 
		then:
			forward mouseMoved events only to the mouse panel. (they aren't really needed by anyone else!)
			forward other mouse related events only to the mouse panel, which maintains dx, dy and the buttons.
			key events are currently not forwarded directly to the keyboard panel, 
				because keys for modal dialogs and when editing views must be handled too.
		mouse clicks through to other apps are not dispatched here so can't be handled here.

	Screen:startFullscreen and stopFullscreen:
		to handle the OS X 10.3.9 "clicking through to other apps" problem:
		sets the screen window to full size of the grabbed & resized fullscreen
		=> clicks cannot go through to other apps => i will get all clicks in MainApp:sendEvent.
		window frame is saved and restored later.
	
	KempstonMouseInterfacePanel:grabMouse:
		Mouse pointer is located over my panel (over the button exactly) and nailed there.
		This does not reliably work in 10.3.9, because CGDisplayBounds() returns 
		bounds in different state (10.4: as for desktop, 10.3.9: as currently set for offscreen)
		so i will change it to somewhere outside (invisible) menu bar and outside (invisible) window title bar,
		since i now get all clicks anyway.
*/


@implementation KempstonMouseInterfacePanel

-(id)initWithContentRect:(NSRect)contentRect 
			   styleMask:(uint)styleMask 
			     backing:(NSBackingStoreType)backingType 
				   defer:(BOOL)flag
{
	XLogIn("KempstonMouseInterfacePanel:initWithContentRect");
	
	return kempston_mouse_if_panel = 
	[super initWithContentRect:contentRect styleMask:styleMask backing:backingType defer:flag];
}


/*
-(void)awakeFromNib { XLogIn("KempstonMouseInterfacePanel:awakeFromNib"); [self setDelegate:self]; }
-(void)windowDidResignKey:(NSNotification*)n { XLogIn("KempstonMouseInterfacePanel:windowDidResignKey"); }
-(void)windowDidBecomeKey:(NSNotification*)n { XLogIn("KempstonMouseInterfacePanel:windowDidBecomeKey"); }
*/


- (IBAction)grabMouse:(id)sender		
{
	Mouse::GrabMouse(); 
}

- (void)releaseMouse:(id)sender			
{
	Mouse::ReleaseMouse(); 
}


- (IBAction)setMouseScaling:(id)sender	
{
	XXASSERT( sender == popup_scale_mouse );
	Mouse::SetScaling( [sender indexOfSelectedItem] );
}


-(void)updatePanel:(Machine*)m
{
	KempstonMouseInterface* mif = m->kempston_mouse_interface; 

	static bool old_mif = 1;
	if( old_mif != (mif!=0) )
	{
		old_mif = mif!=0;
		[self enableSubviews:old_mif];
	}

	if(!mif) return;	// frontmost machine has no mouse interface => exit
	
	static bool old_grabbed = no;
	if( old_grabbed!=Mouse::IsGrabbed() )
	{
		old_grabbed = Mouse::IsGrabbed();
		[button_grab_mouse setTitle:old_grabbed?@"CMD to exit":@"Grab mouse"];
	}

	static uint old_buttons = 0xffff;
	if( old_buttons!=mif->GetButtons() )
	{
		old_buttons = mif->GetButtons();
		[view_buttons setBinaryValue:old_buttons withZeros:"%------LR" andOnes:"%------11"];
	}

	static ushort old_x = 0xffff;
	if( old_x!=mif->GetXAxis() )
	{
		old_x=mif->GetXAxis();
		[view_x_axis setIntValue:old_x];
	}

	static ushort old_y = 0xffff;
	if( old_y!=mif->GetYAxis() )
	{
		old_y = mif->GetYAxis();
		[view_y_axis setIntValue:old_y];
	}
}

@end















