/*	Copyright  (c)	Günter Woigk 1995 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	Manage the paging control register of a ZX Spectrum+ 128K
*/

#define	SAFE	3
#define	LOG		1

#import "Mmu128k.h"
#import "Machine.h"
#import "Z80.h"
#import "Ula.h"



#define	o_addr_128	"0---.----.----.--0-"	// üblicher Port: 0x7ffd


/*	creator for use by derived classes:
	• MmuPlus3
*/
Mmu128k::Mmu128k ( Machine* m, cstr name, cstr o_addr, cstr i_addr )
:	Mmu( m, name, o_addr, i_addr )
{
	XXLogIn("new Mmu128k"); 
}


Mmu128k::Mmu128k ( Machine* m )
:	Mmu( m, "Mmu 128k", o_addr_128, NULL )
{
	XLogIn("new Mmu128k"); 
}


void Mmu128k::Reset ( Time, long cc )	
{
	XLogIn("Mmu128k:Reset"); 

	if( cc > ula->GetCpuCycleOfNextCrtRead() ) ula->UpdateScreenUpToCycle(cc);
	SetPort7ffd(0x00);
}


void Mmu128k::PowerOn ( Time t, long cc )
{
	XLogIn("Mmu128k:PowerOn");
	
	Mmu::PowerOn(t,cc);
	XXASSERT(ram.Count()==0x20000);
	XXASSERT(rom.Count()==0x08000);
	//if( cc > ula->GetCpuCycleOfNextCrtRead() ) ula->UpdateScreenUpToCycle(cc);
	SetPort7ffd(0x00);
}


/*	page in ram and rom pages for 48k mode
	also suitable for +2A/+3
*/
void Mmu128k::page_mem_48k()
{
	XXXASSERT(IsLockedTo48k());		// assume: locked to 48k 

	cpu->MapRam( 3*0x4000, 0x4000, &ram[0*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	cpu->MapRam( 2*0x4000, 0x4000, &ram[2*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	cpu->MapRam( 1*0x4000, 0x4000, &ram[5*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	if(GetMachine()->InternalRomDisabled()) return;
	cpu->MapMem( 0*0x4000, 0x4000, &rom[rom.Count()-0x4000/*for +128k/+2A/+3*/], nowritepage, NULL, 0 );
}


/*	page in selected rom 
	assumed: not (yet) locked to 48k
*/
inline void Mmu128k::page_rom_128k()
{
//	XXXTRAP(IsLockedTo48k());	// kann in Output() gerade gelockt worden sein

	if( GetMachine()->InternalRomDisabled() ) return;
	int i = port_7ffd & 0x10;
	cpu->MapMem( 0x0000, 0x4000, &rom[i*0x400], nowritepage, NULL, 0 /*and no waitstates*/ );
}


/*	page in selected ram at $c000
	assumed: not (yet) locked to 48k
*/
inline void Mmu128k::page_ram_128k()
{
//	XXXTRAP(IsLockedTo48k());	// kann in Output() gerade gelockt worden sein

	int i = port_7ffd & 0x07;
	cpu->MapRam( 3*0x4000, 0x4000, &ram[i*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
}


/*	page in rom and ram pages
	assumed: not locked to 48k
*/
inline void Mmu128k::page_mem_128k()
{
	XXXTRAP(IsLockedTo48k());	// assumed: not locked to 48k 

	page_ram_128k();
	cpu->MapRam( 2*0x4000, 0x4000, &ram[2*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	cpu->MapRam( 1*0x4000, 0x4000, &ram[5*0x4000], ula->GetWaitmap(), ula->GetWaitmapSize() );
	page_rom_128k();
}



/*	set port $7ffd 
	override 48k lock
*/
void Mmu128k::SetPort7ffd ( uchar byte )
{
	XXLogIn( "Mmu128k:SetPort7ffd:$%2x",(int)byte );

	port_7ffd = byte;

	if( byte&0x20 )		// locked to 48k ?
	{
		page_mem_48k();
		ula->SetVideoRam( &ram[5*0x4000] );
	}
	else
	{
		page_mem_128k();
		ula->SetVideoRam( &ram[5*0x4000 + (byte&0x08)*0x1000] );
	}
}


void Mmu128k::Output ( Time, long cc, ushort addr, uchar byte )
{
	XXLogIn("Mmu128k:Output($%2x)",(int)byte);

	if( IsLockedTo48k() ) return;			// mmu disabled. 
											// Note: some demos seem to assume that the rom/ram state is freezed
											// and not set to 48k compatible mode!
	uchar toggled = port_7ffd ^ byte;
	port_7ffd = byte;

	if( toggled&0x08 )						// screen changed?
	{										// (byte&0x08) ? ram+0x4000*7 : ram+0x4000*5;
		if( cc>=ula->GetCpuCycleOfNextCrtRead() ) ula->UpdateScreenUpToCycle(cc);
		ula->SetVideoRam( &ram[5*0x4000 + (byte&0x08)*0x1000] );
	}
		
	if( toggled&0x07 ) page_ram_128k();		// ram page at $c000 changed?
	if( toggled&0x10 ) page_rom_128k();		// rom page changed?
}












