/*	Copyright  (c)	Günter Woigk 2007 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.
*/

#define	SAFE	3
#define	LOG		1

#import "ZonxBox.h"
#import "ZonxBoxPanel.h"
#import "Machine.h"
#import "utilities.h"



// ------------------------------------------------------------
//			Bi-Pak ZON X Sound Box
// ------------------------------------------------------------

/*	AY-3-8192 sound chip

	xtal:			derived from /Clk pin at bus connector
		162500		ZX81
	
	connected lines:

		/clk
		/iorq
		/wr
		A0-4,A7			same for ZX81 and ZX Spectrum!

	Address decoding (issue 5):

		74LS11: triple 3-AND:

			and1 = A3+A2+A4
			and2 = A1+A0
			and3 = and1+nor1+and2

		74LS02: quad 2-NOR:

			nor1 = !( /wr | /iorq )
			nor2 = !A7
			nor3 = !( nor2 | nor4 )		= A7 + and1+nor1+and2
			nor4 = !and3				= !(and1+nor1+and2)

		AY-3-8912:
					   ______ ______
	ANALOG CHANNEL C =|1     U    28|= DA0
			  TEST 1 =|             |= DA1
			Vcc (+5v)=|             |= DA2
	ANALOG CHANNEL B =|             |= DA3
	ANALOG CHANNEL A =|             |= DA4
			Vss (Gnd)=|  AY-3-8912  |= DA5
				IOA7 =|             |= DA6
				IOA6 =|             |= DA7
				IOA5 =|             |= BC1			nor3 = A3+A2+A1+A0+A4+!(wr|iorq) + A7
				IOA4 =|             |= BC2			+5V
				IOA3 =|             |= BDIR			and3 = A3+A2+A1+A0+A4+!(wr|iorq)
				IOA2 =|             |= A8			+5V
				IOA1 =|             |= /RESET		power-on
				IOA0 =|14_________15|= CLOCK

	BC1     BDIR    Function

	0       0       Inactive
	0       1       Write to register		= A3+A2+A1+A0+A4+!(wr|iorq) + A7=0
	1       0       Read from register		= this is not possible
	1       1       Select register			= A3+A2+A1+A0+A4+!(wr|iorq) + A7=1



	used io addresses in ZX81 sample code:
		out	0xDF			select register
		out	0x0F			write to register

	suggested io addresses for ZX Spectrum:
		out 0xFF			select register
		out 0x7F			write to register

	ZONX: TODO:	The sample code from the ZON X manual for the ZX81 should not work, 
				because they set A4=0 when writing to the register. ?!??
				The ZON X-81 box used port addresses $CF and $0F, where A0=0 in both cases. 
				So this may be an error in the documentation, 
				and the ZON X sample code for the ZX81 should use $DF and $1F.
*/

static cstr z_name[2] = { "Bi-Pak ZON X-81",     "Bi-Pak ZON X"        };
static cstr s_addr[2] = { "----.----.1--0.1111", "----.----.1--1.1111" };
static cstr	w_addr[2] = { "----.----.0--0.1111", "----.----.0--1.1111" };
#define	i_addr	NULL


ZonxBox::ZonxBox ( Machine* m, uint modl )
:	Ay( m, z_name[modl], s_addr[modl], w_addr[modl], i_addr, m->GetInfo()->cpu_cycles_per_second/2.0 )
{
	XXASSERT(modl<2); 
	model = modl;	
	m->zonx_box = this;
	[zonx_box_panel showPanel];
}

ZonxBox::~ZonxBox ( )
{
	GetMachine()->zonx_box = NULL;
}

void ZonxBox::Reset ( Time, long )
{
	// does only a power-on reset
}

void ZonxBox::Input ( Time, long, ushort, uchar&, uchar& )
{
	// can't be read
}


void ZonxBox::SetModel ( uint newmodel )
{
	if( newmodel==model ) return;
	XXASSERT(newmodel<2);
	
	GetMachine()->Lock();		// block run_for_sound

		model = newmodel;
		SetName( z_name[newmodel] );

		select_mask = MaskForSpec(s_addr[newmodel]);
		select_bits = BitsForSpec(s_addr[newmodel]);
		write_mask  = MaskForSpec(w_addr[newmodel]);
		write_bits  = BitsForSpec(w_addr[newmodel]);

		cstr o_spec = And(s_addr[newmodel],w_addr[newmodel]);
		out_mask = MaskForSpec(o_spec);
		out_bits = BitsForSpec(o_spec);
	
	GetMachine()->Unlock();
}
































	