/*	Copyright  (c)	Günter Woigk 1996 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	Z80 cpu emulation
*/

#ifndef	Z80_h
#define	Z80_h

#import "ZxItem.h"
#import "Machine.h"
#import "ZxInfo.h"
#import "Ula.h"
#import "Z80 options.h"


// ----	memory pages ----

#define			CPU_PAGESIZE			(1<<CPU_PAGEBITS)
#define			CPU_PAGEMASK			(CPU_PAGESIZE-1)
#define			CPU_PAGES				(0x10000>>CPU_PAGEBITS)

#if defined(_BIG_ENDIAN)	// m68k, ppc
	struct twobytes { uchar hi,lo; };
#else
	struct twobytes { uchar lo,hi; };
#endif	

struct PgInfo
{
    ushort*		core_r;				// flags in the high byte, data in the low byte
    ushort*		core_w;				// flags in the high byte, data in the low byte
	uchar*		waitmap_r; 
	long		waitmap_r_size;	
	uchar*		waitmap_w; 
	long		waitmap_w_size;	
  #ifdef CPU_STATISTICS
	ulong*		count_r;
	ulong*		count_w;
	ulong*		count_x;
  #endif

	ushort&		both_r				( ushort const addr )		{ return core_r[addr]; }
	ushort&		both_w				( ushort const addr )		{ return core_w[addr]; }
	uchar&		data_r				( ushort const addr )		{ return ((twobytes*)(core_r+addr))->lo; }
	uchar&		data_w				( ushort const addr )		{ return ((twobytes*)(core_w+addr))->lo; }
	uchar&		flags_r				( ushort const addr )		{ return ((twobytes*)(core_r+addr))->hi; }
	uchar&		flags_w				( ushort const addr )		{ return ((twobytes*)(core_w+addr))->hi; }
};



// ----	registers ----

union Z80Regs
{
	ushort	nn[16];
	struct	 { ushort af,bc,de,hl, af2,bc2,de2,hl2, ix,iy,pc,sp, iff; } rr;	
   #if defined(_BIG_ENDIAN)			// m68k, ppc
	struct	 { uchar a,f,b,c,d,e,h,l, a2,f2,b2,c2,d2,e2,h2,l2, xh,xl,yh,yl,pch,pcl,sph,spl, iff1,iff2, r,i,im,nmi; } r;
   #else	// _LITTLE_ENDIAN or _PDP_ENDIAN
	struct	 { uchar f,a,c,b,e,d,l,h, f2,a2,c2,b2,e2,d2,l2,h2, xl,xh,yl,yh,pcl,pch,spl,sph, iff1,iff2, r,i,im,nmi; } r; 
   #endif   
};
	


// ----	dummy read and write pages ----

extern ushort	noreadpage[CPU_PAGESIZE];		// dummy page for unmapped memory
extern ushort	nowritepage[CPU_PAGESIZE];		// dito



class Z80 : public ZxItem
{
protected:
	PgInfo		page[CPU_PAGES];	// attached memory 

	Ula*		crtc;
//	long		crt_cycle;			// crt execution counter (based on cpu T cycle counter)
	long		cpu_cycle;			// cpu T cycle counter

	Z80Regs		registers;			// z80 registers
	long		interrupt_signal;	// interrupt signal active (until cycle)

				Z80					( const Z80& );		// prohibit
    Z80&		operator=			( const Z80& );		// prohibit


public:

// standard stuff:
				Z80			 		( Machine* m );
				~Z80				( );

// Item interface:
virtual bool	IsA					( int id ) const		{ return id==isa_Z80 || ZxItem::IsA(id); }
virtual void	PowerOn				( Time, long cc );		// reset cpu and T cycle counters
virtual void	Reset				( Time, long cc );		// reset cpu

// Run the Cpu:
	bool		Run					( long end_cc, int options );	

// access T cycles counter:
    long&		CpuCycle			( )					{ return cpu_cycle; }
//	long&		CrtCycle			( )					{ return crt_cycle; }

// access registers:
	Z80Regs&	GetRegisters		( )						{ return registers; }

// attach memory:
    void		MapRam				( ushort addr, ushort size, ushort* data, uptr waitmap, long wm_size );
    void		MapRom				( ushort addr, ushort size, ushort* data, uptr waitmap, long wm_size );
    void		MapWom				( ushort addr, ushort size, ushort* data, uptr waitmap, long wm_size );
    void		MapMem				( ushort addr, ushort size, ushort* r_data, ushort* w_data, uptr waitmap, long wm_size );

// access memory:
	PgInfo&		Page				( ushort const addr )					{ return page[addr>>CPU_PAGEBITS]; }

    ushort*		RdPtr				( ushort const addr )					{ return Page(addr).core_r + addr; }
    ushort*		WrPtr				( ushort const addr )					{ return Page(addr).core_w + addr; }

	uchar		Peek				( ushort const addr )					{ return Page(addr).both_r(addr); }
	void		Poke				( ushort const addr, uchar const c )	{		 Page(addr).data_w(addr) = c; }
	ushort		Peek2				( ushort addr );
	void		Poke2				( ushort addr, ushort n );
	ushort		Pop2				( );
	void		Push2				( ushort n );
   
// interrupt:
	long&		InterruptSignal		( )						{ return interrupt_signal; }
	void		RaiseInterrupt		( long end_cc=1<<30 )	{ interrupt_signal = end_cc; }
	void		ClearInterrupt		( )						{ interrupt_signal = 0; }
	void		TriggerNmi			( )						{ registers.r.nmi  = 1; } // nmi is edge-triggered and sets a flipflop in the cpu

// Disassembling
	char*		_xword				( uchar n, ushort &ip );
	int			OpcodeLegalState	( ushort ip );			enum { legal, illegal, weird };
	int			OpcodeLength		( ushort ip );			
	char*		OpcodeMnemo			( ushort ip );
	char*		Disassemble			( ushort &ip );

// Debugger:
	ushort		stack_breakpoint;
	
// Profiler:
  #ifdef CPU_STATISTICS
	void		MapXCounts			( ushort addr, ushort sz, ulong* map );
	void		MapRCounts			( ushort addr, ushort sz, ulong* map );
	void		MapWCounts			( ushort addr, ushort sz, ulong* map );
	void		MapCounts			( ushort addr, ushort sz, ulong* rmap, ulong* wmap, ulong* xmap );

    ulong&		cnt_x				( ushort const addr )					{ return Page(addr).xcount[addr];	  }
    ulong&		cnt_r				( ushort const addr )					{ return Page(addr).rcount[addr];	  }
    ulong&		cnt_w				( ushort const addr )					{ return Page(addr).wcount[addr];	  }
	
	long		cnt_op[256];
	long		cnt_cb[256];
	long		cnt_ed[256];
	long		cnt_xy[256];
	long		cnt_xycb[256];
  #endif
};



#endif















