/*	Copyright  (c)	Günter Woigk   1995-2005
  					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	1995-06-29 kio	started work on pascal string lib
	2000-01-28 kio	started work on c string lib
	2002-01-19 kio	started unix port
	2002-01-26 kio	simplified Using(): now using vsnprintf()
	2002-01-26 kio	added xpool[] for unlimited string length support
	2003-05-27 kio	StrToDouble()
	2003-06-07 kio	fixed problem with VaUsing(va_list); fixed bugs in StrToDouble()
	2003-11-28 kio	fixed problem with VaUsing(va_list); again.
*/

#define	SAFE	1
#define	LOG		1

#include	"config.h"
#include	<math.h>

#ifdef HAVE_NAN_H
#include	<nan.h>
#endif

#include	"cstrings.h"

#ifdef USE_KIOS_THREADS
#include	"disp/disp.h"
#endif

#ifndef NAN
const double NAN = 0.0/0.0;
#endif


DEBUG_INIT_MSG("CStrings <init>");		// not sure whether the pool has init code or not




/* ----------------------------------------------------------
			POOL HANDLING
---------------------------------------------------------- */

#ifdef _USE_KIOS_THREADS

	#define	QSPOOLSIZE	5000		// size of pool for semi-durable
	#define	QSMAXLEN	255

	static char pool[QSPOOLSIZE];
	static idx = 0;


	/* ----	allocate and return a string from the pool ---------------------------------
			quick and dirty hack to allocate semi-durable strings
			use them quickly, for the buffer will be re-used after some more calls ...
			intended to prepare output of numbers etc.
			the string's last byte is preset with 0
			
			2006-10-24 kio: renamed QString() to TempString(), because QString collides with Qt's class QString
	*/
	static char* QuickStr ( int len )
	{	int len1=len+1;

		TRAP(len>QSMAXLEN);
		XTRAP(idx<0||idx>QSPOOLSIZE);

		Tid tid = MyTid();
		if (tid)
		{
			if (!tid->qspool)
			{
				tid->qspool = new char[QSPOOLSIZE];
				XXLog( "allocating qspool for \"%s\"%s", tid->name, tid->qspool ? "" : " failed!" );
				if (!tid->qspool) goto c;
				tid->qsidx = 0;
			}

			if (tid->qsidx+len1>QSPOOLSIZE) tid->qsidx=0;
			tid->qsidx += len1;
			tid->qspool[tid->qsidx-1] = 0;
			XXLogOut();
			return tid->qspool+tid->qsidx-len1;
		}

	c:	if (idx+len1>QSPOOLSIZE) idx=0;
		idx += len1;
		pool[idx-1] = 0;

		XXLogOut();
		return pool+idx-len1;
	}



#else	/* don't _USE_KIOS_THREADS */
	// ••• TO DO: separate pool per thread! •••
	// ••• TO DO: manual flushing and/or temporary pool! •••

	#define QSMAXLEN		100							// max. size to go into the pool
	#define	QSDURABILITY	32							// guaranteed durability (number of TempString() calls)
	#define	QSPOOLSIZE		(QSMAXLEN*QSDURABILITY)		// size of pool for semi-durable

	static char	 pool[QSPOOLSIZE];
	static char* poolptr = pool;
	const  char* poolend = pool+QSPOOLSIZE;

	static str	 xpool[QSDURABILITY];
	static str*	 xptr = xpool;
	const  str*  xend = xpool+QSDURABILITY;
	static int	 xcnt = 0;

	#if XXSAFE
		struct INIT_cstr { INIT_cstr(); };
		static INIT_cstr anyNameIsGood;
		INIT_cstr::INIT_cstr()
		{
			TRAP(poolptr!=pool || poolend!=pool+QSPOOLSIZE || xptr!=xpool || xend!=xpool+QSDURABILITY);
		}
	#endif

	void FlushQSPool ( )
	{
		xcnt=0; for(xptr=xpool;xptr<xend;xptr++) if (*xptr) { delete[] *xptr; *xptr=NULL; }
	}


	/* ----	allocate and return a string from the pool ---------------------------------
			quick and dirty hack to allocate semi-durable strings
			use them quickly, for the buffer will be re-used after some more calls ...
			intended to prepare output of numbers etc.
			the string's last byte is preset with 0
	*/
	str TempString ( int len )
	{
		if (xcnt)
		{
			xcnt--;
			if (++xptr>=xend) xptr=xpool;
			if (*xptr) { delete[] *xptr; *xptr=NULL; }
		}

		if ((uint)len<QSMAXLEN)
		{
			XXXTRAP(poolptr<pool||poolptr>poolend);

			len++;
			if (poolptr+len>poolend) poolptr=pool;
			poolptr += len;
			poolptr[-1] = 0;
			return poolptr-len;
		}
		else
		{
			XTRAP(len<0);
			xcnt = QSDURABILITY;
			return *xptr = NewStr(len);
		}
	}
#endif	/* _USE_KIOS_THREADS */





#include "cstrings_functions.h"






