/*	Copyright  (c)	Günter Woigk   1995-2006
  					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	1995-06-29 kio	started work on pascal string lib
	2000-01-28 kio	started work on c string lib
	2002-01-19 kio	started unix port
	2002-01-26 kio	simplified Using(): now using vsnprintf()
	2002-01-26 kio	added xpool[] for unlimited string length support
	2003-05-27 kio	StrToDouble()
	2003-06-07 kio	fixed problem with VaUsing(va_list); fixed bugs in StrToDouble()
	2003-11-28 kio	fixed problem with VaUsing(va_list); again.
	2004-12-16 kio	split into functions and qstring part for multi threading safe version for Cocoa
*/

#ifndef cstrings_h
#error "must be included from cstrings.cp or similar"
#endif


#ifndef NAN
const double NAN = 0.0/0.0;
#endif


/* ----	allocate char[] -----------------------------------------
		deallocate with delete[]
		includes OMEM check
		presets terminating 0
*/
str NewStr ( int n )
{
	str c = new char[n+1];
	OMEM(c);
	c[n] = 0;
	return c;
}


/* ----	allocate char[] -----------------------------------------
		deallocate with delete[]
		includes OMEM check
		returns NULL if source string was NULL			29.jan.2001
*/
str NewCopy ( cstr s )
{
	str c = NULL;
	if (s)
	{
		c = NewStr(strlen(s));
		strcpy ( c, s );
	}
	return c;
}


#ifdef _MPW
/* ----	vsnprintf emulation for MPW ----------- TODO ----
		kio 2002-09-15 quick hack
*/
inline int vsnprintf ( str dest, size_t size, cstr format, va_list va )
{
	int bytes_written = vsprintf(dest,format,va);
	TRAP(bytes_written>size);
	return bytes_written;
}
#endif





/* ----------------------------------------------------------
			METHODS
---------------------------------------------------------- */


/* ----	Create duplicate of string ---------------------------------
		Put a string in the pool
*/
str DupStr ( cstr s )
{
	if (!s||!*s) return "";			// 28.dec.2000
	int n = strlen(s);
	str c = QuickStr(n);
	memcpy ( c,s,n );
	return c;
}



/* ---- compare strings -------------------------------------
		NULL pointers are assumed ""
*/
bool SameStr ( cstr s, cstr t )
{
	if (s&&t)										// kio 28.dec.2000
	{
		while (*s) if (*s++!=*t++) return false;
		return *t==0;								// kio 1.apr.2000
	}
	else
	{
		return ((!s)||(!*s))&&((!t)||(!*t));				// korr: kio 2002-01-20
	}
}



/* ----	search sub string ------------------------------------ kio 11.aug.2001 ----
*/
char* FindStr ( cstr target, cstr search )
{
	if (!search||!*search) return (char*)target;

	int t_len = StrLen(target);
	int s_len = StrLen(search);

	for ( int i=0; i<=t_len-s_len; i++ )
	{
		const char* s = search;
		const char* t = target+i;
		do { if (*s==0) return (char*)target+i; } while (*s++==*t++);
	}
	return NULL;	// not found
}



/* ----	search sub string ------------------------------------ kio 22.aug.2001 ----
*/
char* RFindStr ( cstr target, cstr search )
{
	int t_len = StrLen(target);
	int s_len = StrLen(search);

	if (!search||!*search) return (char*)target+t_len;
	const char* se = search+s_len;

	for ( int i=t_len-s_len; i>=0; i-- )
	{
		const char* s = search;
		const char* t = target+i;
		do { if (s==se) return (char*)target+i; } while (*s++==*t++);
	}
	return NULL;	// not found
}



/* ----	Create and clear a string ---------------------------
*/
str SpaceStr ( int n, char c )
{
	if (n<=0) return "";			// kio 26.nov.2000
	str s = QuickStr(n);
	memset ( s, c, n );
	return s;
}



/* ----	Convert pascal-style string to c-style string ------------------------------
*/
str CStr ( const uchar *p )
{
	XTRAP(p==0);

	str s = QuickStr(*p);
	memcpy(s,p+1,*p);
	return s;
}



// ----	Convert c-style string to pascal-style string ------------------------------
//		string length is limited to 255 (or QSMAXLEN)
uchar* PStr ( cstr c )
{
	XTRAP(c==0);

	int n = StrLen(c);
	if (n>MIN(QSMAXLEN,255)) n = MIN(QSMAXLEN,255);
	uchar* p = (uchar*)QuickStr(n);
	*p=n;
	const uchar* e = p+*p;
	while (p<=e) *++p=*c++;
	return p-n-1;
}



/* ----	create blanked-out copy of string ------------------------------------------
		new: 2002-01-27 kio
*/
str WhiteStr ( cstr q, char c )
{
	str s = DupStr(q);
	for( uchar*p=(uchar*)s; *p; p++ ) if(*p>' ') *p=c;
	return s;
}



/* ----	create string from char ----------------------------------------------------
*/
str CharStr ( char c )
{
	str s = QuickStr(1);
	*s = c;
	return s;
}



// ----	Replace any occurance of a character by another ----------------------------
str ReplaceStr ( cstr s, char old, char nju )
{
	str t = DupStr(s);
	for ( char* p=t; *p; p++ ) if (*p==old) *p=nju;
	return t;
}



// ----	Convert a string to all lower case -----------------------------------------
str LowerStr ( cstr s )
{
	str t = DupStr(s);
	ToLower(t);
	return t;
}



// ----	Convert a string to all upper case -----------------------------------------
str UpperStr ( cstr s )
{
	str t = DupStr(s);
	ToUpper(t);
	return t;
}



/* ----	Repeate string n times ------------------------------------ 22.aug.2001 ----
*/
char* MulStr ( cstr q, int n )
{
	if (q==NULL || *q==0 || n<=0) return "";

	int len = strlen(q);
	str s   = QuickStr(len*n);
	char* z = s;

//	while (n--) z = (char*)memcpy(z,q,len);			should work acc. to man page but doesn't
	while (n--) { memcpy(z,q,len); z+=len; }

	return s;
}



/* ----	Concatenate 2 strings ------------------------------------------------------
*/
char* CatStr ( cstr s1, cstr s2 )
{
	str s = QuickStr(StrLen(s1)+StrLen(s2));
	s[0]=0;
	if (s1) strcpy ( s, s1 );
	if (s2) strcat ( s, s2 );
	return s;
}



/* ----	Concatenate up to 6 strings -------------------------------------------------
*/
char* CatStr ( cstr s1, cstr s2, cstr s3, cstr s4, cstr s5, cstr s6 )
{
	str s = QuickStr( StrLen(s1)+StrLen(s2)+StrLen(s3)+StrLen(s4)+StrLen(s5)+StrLen(s6) );
	s[0]=0;
	if (s1) strcpy ( s, s1 );
	if (s2) strcat ( s, s2 );
	if (s3) strcat ( s, s3 );
	if (s4) strcat ( s, s4 );
	if (s5) strcat ( s, s5 );
	if (s6) strcat ( s, s6 );
	return s;
}



/* ----	Convert number to decimal string ----------
		the string is sized to fit (max. 12 char)
*/
char* NumStr ( long n )
{
	char* s = QuickStr(12);
	uint len = sprintf ( s, "%li", n );
	XXTRAP(len>12);
	return s;
}



/* ----	Convert floating point number to decimal string ---------- kio 15.aug.2001 ----
		the string is sized to fit
*/
char* NumStr ( double d )
{
	char* s = QuickStr(19);
	sprintf ( s, "%.14g", d );
	return s;
}



/* ----	Convert number to hexadecimal string ----------
*/
char* HexStr ( long n, int digits )
{
	const char hex[] = "0123456789ABCDEF";
	char* c = QuickStr(digits);

	while(digits)
	{
		c[--digits] = hex[n&0x0f];
		n >>= 4;
	}
	return c;
}



/* ----	create string [a ... [e ----------------------  1.apr.00 kio
*/
char* SubStr ( cstr a, cstr e )
{
	if (e<=a) return "";
	char* c = QuickStr(e-a);
	memcpy ( c, a, e-a );
	return c;
}



/* ----	create left substring ------------------------
		splits n bytes from the left
		if source string is shorter, the result is shorter too
*/
char* LeftStr ( cstr s, int n )
{
//	long m = StrLen(s); if (m<n) n=m;		<-- dangerous!
	const char*a=s; 		// kio 2.aug.2001
	const char*e=s+n;		// ""
	while (a<e&&*a) a++;	// ""
	n = a-s;				// ""

	if (n<=0) return "";
	char* c = QuickStr(n);
	memcpy ( c, s, n );
	return c;
}



/* ----	create right substring ------------------------
		splits n bytes from the right
		if source string is shorter, the result is shorter too
*/
char* RightStr ( cstr s, int n )
{
	long m = StrLen(s); if (m<n) n=m;
	if (n<=0) return "";
	char* c = QuickStr(n);
	memcpy ( c, s+m-n, n );
	return c;
}



/* ----	create mid substring ------------------------
		skip a bytes from the left then split n bytes
		if source string is shorter, the result is shorter too
*/
char* MidStr ( cstr s, int a, int n )
{
	if (a<0) { n+=a; a=0; }			// korr. 6.aug.2001 kio
	n = Min(n,StrLen(s)-a);
	if (n<=0) return "";
//	return LeftStr(s+a,n);
	return SubStr(s+a,s+a+n);		// kio 2002-03-29
}



/* ----	create mid-to-end substring -------------------
		kio 2002-03-29
*/
char* MidStr ( cstr s, int a )
{
	if(!s) return "";
	if (a<0) a=0;
	return SubStr(s+a,strchr(s,0));
}



/* ----	create formatted string ----------------------------------
		2002-01-26 kio	now using vsnprintf()
		2003-11-28 kio	now calling VaUsing() which calls vsnprintf()
*/
str Using ( cstr format, ... )
{
	va_list va;
	va_start(va,format);
	str s = VaUsing(format,va);
	va_end(va);
	return s;

/*	old:
	OSErr err = Error();			// vsnprintf sets spurious errors => fix

	char* s = QuickStr(QSMAXLEN-1);	// allocate in pool, not xpool

	va_list va;
	va_start(va,format);
	uint n = vsnprintf( s, QSMAXLEN, format, va );
	va_end(va);

	if (n>=QSMAXLEN)				// didn't fit into QSMAXLEN ?
	{
		if (int(n)<0) return ""; 	// error  or  pre glibc 2.1 and output truncated

		s = QuickStr(n);
		va_start(va,format);
		uint m = vsnprintf( s, n+1, format, va );
		va_end(va);
		if (m>=n) return "";		// error
	}

	Error() = err;					// all ok => restore errno
	return s;
*/
}



/* ----	VaUsing() -----------------------
		2003-06-26 kio: hopefully fixed "va_list z = va;" problem
		2003-11-28 kio: hopefully fixed "va_copy" problem
		note: caller must call va_start() prior and va_end() afterwards.
*/
str VaUsing ( cstr format, va_list va )
{
	XXLogIn("VaUsing()");
	#define MAXLEN	QSMAXLEN			// recommended: QSMAXLEN  ((256))
	str s = QuickStr(MAXLEN-1);			// MAXLEN<=QSMAXLEN => allocate in pool, not xpool
	OSErr err = Error();				// vsnprintf sets spurious errors => fix

/*	from the GNU autoconf manual:
	va_copy: The ISO C99 standard provides va_copy for copying va_list variables.
	It may be available in older environments too, though possibly as __va_copy
	(e.g., gcc in strict C89 mode). These can be tested with #ifdef.
	A fallback to memcpy (&dst, &src, sizeof(va_list)) will give maximum portability.

	va_list is not necessarily just a pointer.
	It can be a struct (e.g., gcc on Alpha), which means NULL is not portable.
	Or it can be an array (e.g., gcc in some PowerPC configurations),
	which means as a function parameter it can be effectively call-by-reference
	and library routines might modify the value back in the caller
	(e.g., vsnprintf in the GNU C Library 2.1).
*/
	va_list z;
	memcpy(&z,&va,sizeof(va_list));

	int n = vsnprintf( s, MAXLEN, format, z );
//	va_end(z);		<-- probably an error to call va_end() on the copy
//						va_end() probably fixes the return stack

	if (n>=MAXLEN)						// didn't fit into buffer ?
	{
		XXLog("resized: %i characters",n);
		s = QuickStr(n);
		vsnprintf( s, n+1, format, va );
	}

	Error() = err;						// restore errno
	XXLogOut("");
	return s;
}



/* ----	create beautified text for number ---------------------------
		from project vipsi
*/
str SpeakingNumber ( double n )
{
#if defined(_MACINTOSH)
	if (n>=0x7ff80000) return DupStr("2++ GB");
#endif

	if (n>=1000000000) return Using( "%7.3lf GB", n/1000000000 );
	if (n>=1000000   ) return Using( "%7.3lf MB", n/1000000 );
	if (n>=1000      ) return Using( "%7.3lf kB", n/1000 );
	                   return Using( "%4.0lf bytes", n );
}



/* ----	convert string ------------------------------
*/
static const char cc[] = "\\\"\e\a\b\f\n\r\t\v";	// control codes
static const char ec[] = "\\\"eabfnrtv";			// escape characters

static const char* StrComp(cstr a, cstr b)
{	char c,d;
	while((c=*a)&&(d=*b)&&c==d) {a++;b++; }
	return a;
}

char NextChar ( char const *& p )
{	char c;
	c = *p++; if (c!='\\') return c;	// plain char
	c = *p++;
	if( (c-'0')&~7 )					// no octal digit
	{
		cptr q = strchr(ec,c);
	 	if (c==0) p--;					// dont advance text pointer if "\" at end of text
		if (q) return cc[q-ec]; 		// control code
		return c;						// unkn. escape char / char escaped for unknown reason
	}
	else								// octal digit
	{
		uchar n = c-'0';
		if (((c=*p++-'0')&~7)==0) { n=n*8+c; if (((c=*p++-'0')&~7)==0) return n*8+c; }
		p--; return n;
	}
}


str ConvertedTo ( cstr s0, strconvtype ctype )
{	str  s;
	char c;
	int  i;
	char *q,*z;
	uchar uc;

	if (ctype==str_quoted)
	{
		s = ConvertedTo(s0,str_printable);
		return CatStr("\"",s,"\"");
	}

	if (!s0||!*s0) return "";
	z = s = (str)s0;

	if (ctype==str_html)
	{
		for ( i=0; (c=s[i]); i++ )
		{
			switch(c)
			{
			default: 	continue;
			case '<':	z = "&lt;";   break;
			case '>':	z = "&gt;";   break;
			case '&':	z = "&amp;";  break;
			case '"':	z = "&quot;"; break;		// 2002-07-30 kio: added
			case '\n':	z = "<br>";   break;		// 2004-07-08 kio: added
			}
			s = CatStr( LeftStr(s,i), z, s+i+1 ); i+=3;
		}
	}

// str_printable:
	else
	{
		while ( (uc=*z) )
		{
			if( uc>=' ' && uc!=0x7F && uc!='\\' && uc!='\"' ) { z++; continue; }
	//		if( uc>=' ' && uc<0x7F && uc!='\\' && uc!='\"' ) { z++; continue; }
	//		if( uc>=0x80&&UTF8CharValid(z)) { z=UTF8StrNextChar(z); continue; }

			if ((q=strchr(cc,uc)))
			{
				*z = ec[q-cc];
				q=s; s=CatStr(SubStr(s,z),"\\",z); z+=2+(s-q); continue;
			}
			char oc[]="\\000";
			oc[1]='0'+(uc>>6); uc&=(1<<6)-1;
			oc[2]='0'+(uc>>3); uc&=(1<<3)-1;
			oc[3]='0'+(uc>>0);
			q=s; s=CatStr(SubStr(s,z),oc,z+1); z+=4+(s-q); continue;
		}
	}

	return s!=s0 ? s : DupStr(s0);
}

str ConvertedFrom ( cstr s0, strconvtype ctype )
{	char c;
	char *q,*z;
	str  s;
	int  n;

	if(!s0||!*s0) return "";

	s = DupStr(s0);

	if (ctype==str_html)
	{
		q = z = strchr(s,'&');

		if(q) for(;;)
		{
			while((c=*q++)!='&') { *z++=c; if(!c) return s; }
			if (*StrComp("amp;", q)==0) { *z++ = '&'; q+=4; continue; }
			if (*StrComp("quot;",q)==0) { *z++ = '"'; q+=4; continue; }		// 2002-07-30 kio: added
			if (*StrComp("gt;",  q)==0) { *z++ = '>'; q+=3; continue; }
			if (*StrComp("lt;",  q)==0) { *z++ = '<'; q+=3; continue; }
			*z++='&';	// unquoted '&'
		}
	}

	else
	{
		if (ctype==str_quoted)
		{
			n=strlen(s);
			c=s[0];
			if( n>=2 && (c=='"'||c=='\'') && s[n-1]==c )
			{	//  backslashes zählen:  ist "abc\\\\\\\\" gequoted ?
				for (q=s+n-2;*q=='\\';q--) {}
				if( (n-(q-s))&1 ) { s++; q[1]=0; }
			}
		}

	// str_printable:
		{
			q = z = strchr(s,'\\');

			if(q) for(;;)
			{
				while((c=*q++)!='\\') { *z++=c; if(!c) return s; }

				c=*q++;
				if( (c-'0')&~7 )	// no octal digit
				{
					const char* p;
					if ((p=strchr(ec,c))) *z++ = cc[p-ec]; 	// control code
					else { *z++ = '\\'; q--; }				// unkn. escape char
				}
				else				// octal digit
				{
					n=c-'0';
					if (((c=*q++-'0')&~7)==0) { n=n*8+c; if (((c=*q++-'0')&~7)==0) n=n*8+c; else q--; } else q--;
					*z++ = n;
				}
			}
		}
	}

	return s;
}



/* ----	convert ascii string to numeric value --------------------------------
		=>	±123.456e±78
			±'zb.ascii'
			±$abcd.ef
			±%101010.101010

		Works according to 'strtod'.

		Wenn endptr nicht NULL ist, wird ein Zeiger auf das letzte konvertierte
		Zeichen in endptr gespeichert.

		Wenn  keine Konvertierung stattgefunden hat, wird 0 zurückgeliefert und
		der Wert von nptr an der Stelle endptr gespeichert.

		Wenn der korrekte Wert einen Überlauf verursacht, wird HUGE_VAL
		zurückgeliefert (in Abhängigkeit des Vorzeichens des Wertes) und
		ERANGE wird in errno gespeichert. Verursacht der korrekte Werte einen
		negativen Überlauf, wird 0 zurückgegeben und ERANGE wird in errno
		gespeichert.
*/
double StrToDouble ( cptr q0, cptr* endptr )
{
	cptr   q   = q0;
	char   c   = *q; while(c>0&&c<=' ')   { c = *++q; }
	bool   neg = c=='-'; if (neg||c=='+') { c = *++q; }
	cptr   e   = q0;
	cptr   a;
	double f   = 0.0;
	int	   n;

	if (_is_dec_digit(c))		// 123.456e78
	{
		f = strtod(q,(ptr*)&e);
		if(e==q) e=q0;
		if (Error()==HUGE_VAL) ForceError(numbertoobig);
		//XXLog("strtod(%s) = %.20f",SubStr(q0,e),f);
	}
	else if (c=='$')			// $abc.def
	{
		a=q+1;
		while ( (n=_digit_value(*++q))<16 ) { f = ldexp(f,4) + n; }
		if (*q=='.')
		{
			a=q+1;
			while ( (n=_digit_value(*++q))<16 ) { f = ldexp(f,4) + n; }
			f = ldexp(f,4*(a-q));
		}
		if(q>a) e = q;
	}
	else if (c=='%')			// %101010.101010
	{
		a=q+1;
		while ( (n=_digit_val(*++q))<2 ) { f = ldexp(f,1) + n; }
		if (*q=='.')
		{
			a=q+1;
			while ( (n=_digit_val(*++q))<2 ) { f = ldexp(f,1) + n; }
			f = ldexp(f,a-q);
		}
		if(q>a) e = q;
	}
	else if (c=='\'') 			// 'abcd'
	{
		while((c=*++q))
		{
			if(c=='\'') { e=++q; break; }
			if(c=='\\') { c = NextChar(q); q--; }
			f = ldexp(f,8) + (uchar)c;
		}
	}

	if(e==q0) { f=NAN; SetError(notanumber); }
	if (endptr)	*endptr = e;
	return neg ? -f : f;
}


str ParseStr ( cptr& text )
{	char c;
	cptr q = text;

	if (*q!='"') return NULL; else q++;

	while ((c=*q++))
	{
		if (c=='"') break;
		if (c!='\\') continue;
		q--; NextChar(q);
	}

	if (c) { str s = Unescaped(SubStr(text+1,q-1)); text = q; return s; }
	else return NULL;
}


/* ----	read line from file ----------------------------------------- [kio 2004-04-30]
		read text up to \0, \n, \r, \n\r or \r\n
		returns string allocated with new[]  (( up to 100 char / 50% over-allocated! ***TODO*** ))
		returns NULL at EOF/Error => check ferror(f)
*/
str NewReadStr ( FILE* f )
{
	TRAP(EOF>=0);			// should be ok  =>  nop

	int sz = 100;			// initial max. size
	str s  = NewStr(sz);
	int n  = 0;

// read character-by-character and break on linebreak
	for(;;)
	{
		int c = getc(f);

		if(c<=13)
		{
			if (c==EOF) { if(n) break; delete[] s; return NULL; }	// EOF/Error
			if (c==0) break;
			if (c==10||c==13)
			{
				int c2 = getc(f);
				if (c2!=EOF && c+c2!=23) ungetc(c2,f);
				break;
			}
		}

		if(n==sz)		// s[] full => need more!
		{
			char* z = NewStr( sz+=sz/2 );
			memcpy(z,s,n);
			delete[] s;
			s = z;
		}

		s[n++] = c;
	}

	s[n] = 0;		// string delimiter
	return s;		// allocated with new[]
}






/* ----	skip white spaces in string -------------------------------
		indexes starting with 1
		idx points to the next word or to the end of the string
void QSSkipSpace ( uchar *s, short&idx )
{	int i;
	for (i=idx; i<=*s && s[i]<=' '; i++){};
	idx=i;
}
*/

/* ----	find white space in string --------------------------
		indexes starting with 1
		idx points behind the actual word
void QSSkipWord ( uchar *s, short&idx )
{	int i;
	for (i=idx; i<=*s && s[i]>' '; i++){};
	idx=i;
}
*/

/* ----	extract next word from string -------------------------------------
		indexes starting with 1
		kio 22.5.98
uchar* QSNextWord ( uchar*s, short&idx )
{
	QSSkipSpace(s,idx);
	short a=idx;
	QSSkipWord(s,idx);
	return QSMidStr(s,a,idx-1);
}
*/

/* ----	find next occurance of a special character in string ---------------
		indexes starting with 1
		idx points to the found char or behind string
void QSFindChar ( uchar *s, uchar c, short&idx )
{	int i;

	for (i=idx; i<=*s && s[i]!=c; i++){};
	idx=i;
}
*/


/* ----- write line to file ---------------------------------------
OSErr QSWriteStr ( uchar *s, short fileID )
{	long tfu;

	tfu = *s;
	return FSWrite ( fileID, &tfu, s+1 );
}
*/



/* reformat date & time
uchar* QSDateTimeStr ( uchar* date, uchar* time )
{
	date = QSReplaceStr(date,' ','.'); if (date[5]=='.') date=QSCatStr(QSLeftStr(date,4),QSRightStr(date,6));
	time = QSLeftStr(time,5);
	return QSCatStr(QSCatStr(date,"\p "),time);
}
*/



/* ---- Uhrzeit ------------------------------------------------------

	format = "0h0m0s0"
		wobei Nullen jeweils die Mindestzahl von Stellen angeben und
		h, m und s Stunden-, Minuten-, Sekunden- und Ticks-Felder trennen.

		Für h wird :, für m wird : und für s wird , ausgegeben.



uchar* ClockStr ( unsigned int t, const uchar* f )
{
}
*/

/* ---- Zeitdauer ------------------------------[kio 5.1.98]---------

	format = "\p0h0m0s0"
		wobei Nullen jeweils die Mindestzahl von Stellen angeben und
		h, m und s Stunden-, Minuten-, Sekunden- und Ticks-Felder trennen.

		Für m kann auch " und für s kann auch ' angegeben werden.
		Das jeweils benutzte Trennzeichen wird im Ausgabestring verwendet.

		Alle Zeichen, die nicht h, m, s " oder ' sind, werden wie Nullen behandelt
		und an der entsprechenden Stelle im String eingesetzt, wenn der Platz
		dort sonst leer wäre.

		"0"00'00"	 ergibt z.B. "13"30'00"
		"--h00m00s0" ergibt z.B. "-4h13m04s17"
		"000h0"	 	 ergibt z.B. "001h1746"

		Wird NULL als Formatstring angegeben, wird ein dem Wert angepasstes Format
		variabler Länge benutzt.

uchar* QSDurationStr ( unsigned int d, const uchar* f )
{	int 	n;
	uchar	*p;
	uchar	*s1,*s2;

// default format
	if ( !f )
	{	if ( d>=60*60*60 ) 	f = "\p0h00m00s"; else
		if ( d>=60*60 )		f = "\p0m00s"; else
							f = "\p0s00";
	}

// runden
	switch ( f[*f] )
	{
	case 's':
	case '\"':	d += 60/2;		break;
	case 'm':
	case '\'':	d += 60*60/2; 	break;
	case 'h':	d += 60*60*60/2;break;
	}

// textsammler
	s1 = "\p";

// stunden
	p = (uchar*)memchr(f+1,'h',*f);
	if (p)
	{	n=p-(f+1);
		s2 = QSNumStr ( d/(60*60*60) );
		if ( *s2<n ) s2 = QSCatStr ( QSLeftStr(f,n-*s2), s2 );
		s1 = QSCatStr ( s2, "\ph" );
		d %= (60*60*60);
		f = QSRightStr ( f, n+2 );
	}

// minuten
	p = (uchar*)memchr(f+1,'m',*f); if (!p) p = (uchar*)memchr(f+1,'\'',*f);
	if (p)
	{	n=p-(f+1);
		s2 = QSNumStr ( d/(60*60) );
		if ( *s2<n ) s2 = QSCatStr ( QSLeftStr(f,n-*s2), s2 );
		s2 = QSCatStr ( s2, QSCharStr(*p) );
		s1 = QSCatStr ( s1,s2 );
		d %= (60*60);
		f = QSRightStr ( f, n+2 );
	}

// sekunden
	p = (uchar*)memchr(f+1,'s',*f); if (!p) p = (uchar*)memchr(f+1,'\"',*f);
	if (p)
	{	n=p-(f+1);
		s2 = QSNumStr ( d/(60) );
		if ( *s2<n ) s2 = QSCatStr ( QSLeftStr(f,n-*s2), s2 );
		s2 = QSCatStr ( s2, QSCharStr(*p) );
		s1 = QSCatStr ( s1,s2 );
		d %= (60);
		f = QSRightStr ( f, n+2 );
	}

// ticks
	n=*f;
	if (n)
	{	s2 = QSNumStr ( d );
		if ( *s2<n ) s2 = QSCatStr ( QSLeftStr(f,n-*s2), s2 );
		s1 = QSCatStr ( s1,s2 );
	}

	return s1;
}
*/





