/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	
	Maintain a color map for indexed color images and Gif files
	-----------------------------------------------------------
	
	started: 		2007-06-01 kio
	last modified:	2007-07-09 kio

Overview:

	the Colormap contains an array of colors, each made out of 3 bytes for R, G and B.
	
	the Colormap can be sized to contain no colors  or  up to 2^n colors with n = 1 to 8  <=>  2 to 256 colors.
	the size of the Colormap is always extended to full 2^n.
	
	a Colormap size of 2^0 colors <=> 1 color is not allowed and set to 2^1 <=> 2 colors instead.
	this is basically to support gif color maps, which must be at least 2^1 <=> 2 colors in size, if present.

	the Colormap maintains a transparent color index.
	
	the Colormap maintains the number of actually used colors.
	
	when adding colors with AddColor() or CondAddColor() or AddTranspColor() to a fully occupied Colormap, 
	then the colormap is extended to the next allowed size, that is, to 2^(n+1). (but stepping from 0 to 2 directly.)
	
	if the Colormap is shrinked, then it is again only shrinked to the next allowed size, 
	e.g. shrinking to 5 colors will actually shrink to 2^3 = 8 colors,
	and the used color count and the transparent color index are validated.

	used colors can be accessed with operator[], which returns a pointer (!) to the first color component.	
*/


#ifndef Colormap_h
#define Colormap_h

#include "kio/kio.h"
#include "templates/Array.h"


typedef uchar Comp;
typedef Comp const cComp;
typedef class Colormap const cColormap;
typedef Array<Comp> Cmap;


class Colormap : protected Cmap
{
protected:
//	Comp*	Cmap::array		// allocated memory	
//	int		Cmap::count		// allocated memory size
	int		used_colors;	// number of actually used colors:   [0 .. [count/3
	int		transp;			// transparent color index: unset or [0 .. [used_colors

static int	bits			( ushort n )	{ return n==0 ? 0 : Log2(uchar(n-1))+1; }	// f(0)=0, f(1)=1, f(2^n)=n
static int	valid_count		( ushort n )	{ XXLIMIT(n,256); return n==0 ? 0 : 1<<bits(n); }
																						// 0 or 2,4,8,16,32,64,128,256

private:
// init, copy, kill:
	void	init			()							{ used_colors=0; transp=unset; }
	void	init			( cComp* q, int n, int i )	{ memcpy(array,q,n*3); used_colors=n; transp=i; } 


public:		

// constants for transparent color index and find_color(..)
	enum	{ unset = -1, not_found = -1 };				


// create, copy, destroy:
			Colormap		()							throw()						  { init(); }
			Colormap		( int n )					throw(bad_alloc,limit_error) : Cmap(valid_count(n)*3) { init(); }
			Colormap		( cComp* q, int n, int i=unset )
														throw(bad_alloc,limit_error) : Cmap(valid_count(n)*3) { init(q,n,i); }
//			~Colormap		()							{ kill(); }
//			Colormap		( cColormap& q )			{ init(q); }
//  Colormap& operator=  	( cColormap& q )			{ copy(q); }


// color map:

	Comp*	GetCmap			() throw()					{ return array; }
	cComp*	GetCmap			() const throw()			{ return array; }
	int		CmapByteSize	() const throw()			{ return count; }

	void	PurgeCmap		() throw()					{ Cmap::Purge(); init(); }
	int		CmapSize		() const throw()			{ return count/3; }
	int		CmapBits		() const throw()			{ return bits(count/3); }
	void	GrowCmap		( int n ) throw(bad_alloc,limit_error)	{ n=valid_count(n); Cmap::Grow(n*3); }
	void	ShrinkCmap		( int n ) throw(limit_error);


// transparent color:

	bool	HasTranspColor	() const throw()			{ return transp!=unset; }
	int		TranspColor		() const throw()			{ return transp; }
	void	SetTranspColor	( int i ) throw(index_error){ XXLIMIT(i+1,used_colors+1); transp = i; }
	void	ClearTranspColor() throw()					{ transp = unset; }
	int		AddTranspColor	() throw(limit_error)		{ if(transp==unset) { transp = AddColor(0,0,0); } return transp; }


// actually used colors:

	int		UsedColors		() const throw()			{ return used_colors; }
	int		UsedBits		() const throw()			{ return bits(used_colors); }
	void	GrowColors		( int n ) throw(bad_alloc,limit_error){ if(n>used_colors) { GrowCmap(n); used_colors=n; } }
	void	ShrinkColors	( int n ) throw(limit_error);

	Comp*	operator[]		( int i ) throw(index_error)		{ XXINDEX(i,used_colors); return array+i*3; }
	cComp*	operator[]		( int i ) const throw(index_error)	{ XXINDEX(i,used_colors); return array+i*3; }

	int		AddColor		( Comp r, Comp g, Comp b )	throw(bad_alloc,limit_error) { Comp q[]={r,g,b}; return AddColor(q); }
	int		AddColor		( cComp q[3] )				throw(bad_alloc,limit_error);

	int		CondAddColor	( cComp q[3] )				throw(bad_alloc,limit_error);
	int		CondAddColor	( Comp r, Comp g, Comp b )	throw(bad_alloc,limit_error);


// find color by it's rgb components: 
// skip transparent color (if any); return index or not_found

	int		FindColor		( Comp r, Comp g, Comp b )	const;	
	int		FindColor		( cComp p[3] ) const		{ return FindColor(p[0],p[1],p[2]); }


// debug:

	cstr	CmapStr			() const;
};


#endif


















