/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	The LZW encoder and decoder were written 
	by Gershon Elber and Eric S. Raymond 
	as part of the GifLib package 
	and released under the MIT license.

	To my best knowing the copyright on the algorithm has expired.
*/


#define SAFE	3
#define LOG		1

#import "GifEncoder.h"
#import "unix/file_utilities.h"
#import "fcntl.h"
#import "unistd.h"
#import "globals.h"



// ######################################################################################
//							hash table for lzw encoder
// ###################################################################################### 


/*	The 32 bits of a hash table entry contain two parts: the key and the code.
	The code is 12 bits since the algorithm is limited to 12 bits.
	The key is a 12 bit prefix code + 8 bit new char = 20 bits.
*/
#define HT_GET_KEY(x)	((x) >> 12)
#define HT_GET_CODE(x)	((x) & 0x0FFF)
#define HT_PUT_KEY(x)	((x) << 12)
#define HT_PUT_CODE(x)	((x) & 0x0FFF)


/*	Generate a hash key from the given unique key.
	The given key is assumed to be 20 bits as follows:
		lower 8 bits are the new postfix character,
		upper 12 bits are the prefix code.
*/
inline int gif_hash_key( ulong key )
{
	return ((key >> 12) ^ key) & HT_KEY_MASK;
}


/*	Clear the hash_table to an empty state.
*/
inline void GifEncoder::clear_hash_table( )
{
	for( int i=0; i<HT_SIZE; i++ ) { hash_table[i] = 0xFFFFFFFFL; }
}


/*	Insert a new item into the hash_table.
	The data is assumed to be new.
*/
inline void GifEncoder::add_hash_key( ulong key, int code )
{
	int hkey = gif_hash_key(key);
	while (HT_GET_KEY(hash_table[hkey]) != 0xFFFFFL) { hkey = (hkey + 1) & HT_KEY_MASK; }
	hash_table[hkey] = HT_PUT_KEY(key) | HT_PUT_CODE(code);
}


/*	Test whether a key exists in the hash_table:
	returns its code or -1.
 */
inline int GifEncoder::lookup_hash_key( ulong key )
{
	int hkey = gif_hash_key(key);
	ulong htkey;

	while ((htkey = HT_GET_KEY(hash_table[hkey])) != 0xFFFFFL) 
	{
		if( key == htkey ) return HT_GET_CODE(hash_table[hkey]);	// exists
		hkey = (hkey + 1) & HT_KEY_MASK;
	}
	return -1;	// does not exist
}




// ######################################################################################
//								lzw encoder
// ###################################################################################### 


/*	Write a byte to a Gif file.

	This function is aware of Gif block structure and buffers
	chars until 255 can be written, writing the size byte first.
	If FLUSH_OUTPUT is the char to be written, the buffer is
	written and an empty block appended.
*/
void GifEncoder::write_gif_byte( int ch )
{
	if( file_state == IMAGE_COMPLETE ) return;

	if( ch == FLUSH_OUTPUT )
	{
		if(bufsize) 
		{
			write_uchar(fd, bufsize);
			write_bytes(fd, buf, bufsize);
			bufsize = 0;
		}
		/* write an empty block to mark end of data */
		write_uchar(fd, 0);
		file_state = IMAGE_COMPLETE;
	}
	else 
	{
		if( bufsize==255 ) 
		{
			/* write this buffer to the file */
			write_uchar(fd, bufsize);
			write_bytes(fd, buf, bufsize);
			bufsize = 0;
		}
		buf[bufsize++] = ch;
	}
}


/*	Write a Gif code word to the output file.

	This function packages code words up into whole bytes
	before writing them. It uses the encoder to store
	codes until enough can be packaged into a whole byte.
*/
void GifEncoder::write_gif_code( int code )
{
	if( code == FLUSH_OUTPUT ) 
	{
		// write all remaining data:
		while (shift_state > 0)
		{
			write_gif_byte(shift_data & 0xff);
			shift_data >>= 8;
			shift_state -= 8;
		}
		shift_state = 0;
		write_gif_byte(FLUSH_OUTPUT);
	}
	else 
	{
		shift_data  |= ((long) code) << shift_state;
		shift_state += running_bits;

		while( shift_state >= 8 )
		{
			// write full bytes:
			write_gif_byte(shift_data & 0xff);
			shift_data >>= 8;
			shift_state -= 8;
		}
	}

	// If code can't fit into running_bits bits, raise its size.
	// Note: codes above 4095 are for signalling.
	if( running_code >= max_code_plus_one && code <= 4095 )
	{
    		max_code_plus_one = 1 << ++running_bits;
	}
}


/*	Initialise the lzw encoder
	in: cmap depth (bits).
*/
void GifEncoder::StartImage( int cmap_bits )
{
	XLogIn("GifEncoder:StartImage(bits)");

	XXASSERT( fd>=0 );				// file must already be opened with CreateFile()
	if(cmap_bits<2) cmap_bits = 2;	// min. table size is 2 bit for algorithm reasons

	depth				= cmap_bits;
	file_state			= IMAGE_SAVING;
	position			= 0;
	bufsize				= 0;
	buf[0]				= 0;
	clear_code			= 1 << cmap_bits;
	eof_code			= clear_code + 1;
	running_code		= eof_code + 1;
	running_bits		= cmap_bits + 1;
	max_code_plus_one	= 1 << running_bits;
	current_code		= FIRST_CODE;
	shift_state			= 0;
	shift_data			= 0;

	write_uchar( fd, cmap_bits );	// Write the lzw minimum code size
	clear_hash_table();				// Clear hash table
	write_gif_code( clear_code );	// output Clear code
}


/*	Write one scanline of pixels out to the gif file,
	compressing that line using lzw into a series of codes.
*/
void GifEncoder::WritePixelRow( cuptr pixel, uint length )
{
	XXXLogIn("GifEncoder:WritePixelRow(px,n)");

	int   new_code;
    ulong new_key;
    uchar pixval;
	uint  i = 0;

	int current_code = this->current_code == FIRST_CODE ? pixel[i++] : this->current_code;

    while( i < length )
    {
		pixval = pixel[i++]; // Fetch next pixel from stream 

		// Form a new unique key to search hash table for the code
		// Combines current_code as prefix string with pixval as postfix char 
		new_key = (((ulong) current_code) << 8) + pixval;
		if( (new_code = lookup_hash_key(new_key)) >= 0 ) 
		{
			// This key is already there, or the string is old,
			// so simply take new code as current_code 
			current_code = new_code;
		}
		else 
		{
			// Put it in hash table, output the prefix code,
			// and make current_code equal to pixval 
			this->write_gif_code( current_code );
			current_code = pixval;

			// If the hash_table if full, send a clear first
			// then clear the hash table: 
			if (this->running_code >= LZ_MAX_CODE) 
			{
				this->write_gif_code(this->clear_code);
				this->running_code = this->eof_code + 1;
				this->running_bits = this->depth + 1;
				this->max_code_plus_one = 1 << this->running_bits;
				clear_hash_table();
			}
			else 
			{
				// Put this unique key with its relative code in hash table 
				add_hash_key( new_key, this->running_code++ );
			}
		}
    }

    // Preserve the current state of the compressor: 
    this->current_code = current_code;
}


/* Write a block of pixels to the gif file
*/
void GifEncoder::WritePixelRect ( cuptr pixel, uint w, uint h, uint dy )
{
	XXLogIn("GifEncoder:WritePixelRect(px,w,h,dy)");

	XXASSERT( (short)dy>=(short)w );
	XXASSERT( w<0x8000 && h<0x8000 );
	
	while(h--)
	{
		WritePixelRow( pixel, w );
		pixel += dy;
	}
}


/* Flush last bits to file
*/
void GifEncoder::FinishImage( )
{
	XLogIn("GifEncoder:FinishImage()");

	write_gif_code(current_code);
	write_gif_code(eof_code);
	write_gif_code(FLUSH_OUTPUT);
}


/* Start image, write image, finish image
*/
void GifEncoder::WritePixelmap ( int bits_per_pixel, cPixelmap& pm )
{
	XXLogIn("GifEncoder:WritePixelmap(bits,pmap)");

	StartImage( bits_per_pixel );
	WritePixelRect( pm.Pixels(), pm.W(), pm.H(), pm.DY() );
	FinishImage();
}


/*	Open gif file for writing
	and write gif header
*/
void GifEncoder::OpenFile( cstr path )
{
	XLogIn("GifEncoder:OpenFile()");

	XXASSERT( fd==-1 );

	fd = open( path, O_WRONLY|O_CREAT, 0664 );
	if (fd<0) throw(file_error(Error()));

	WriteGif89aHeader();
}
								  

/*	Write Comment Block with creator message 
	Write Trailer byte
	Close file
*/
void GifEncoder::CloseFile()
{
	XLogIn("GifEncoder:CloseFile()");

	XXASSERT(fd!=-1);

	if(file_state!=IMAGE_COMPLETE) FinishImage();
	WriteCommentBlock( Using("Made with %s %s.%s",appl_name,appl_version_str,appl_pre_version_str) );
	WriteGifTrailer();
	close(fd); fd=-1; 
}


/*	Write sub image block to gif file
	use the global color table or include a local table
	The image block consists of:
	- Image Descriptor
	- optional local color table
	- lzw compressed pixel data
*/
void GifEncoder::WriteImage ( Pixelmap* pmap, Colormap* cmap )
{
	XXLogIn("GifEncoder:WriteImage(pmap,cmap)");

	XXASSERT( cmap->UsedColors()>0 );
	
	if( cmap==&global_cmap )	// TODO: or cmap==global_cmap[to sizeof(cmap)]
	{
		WriteImageDescriptor( pmap );
	}
	else
	{
		WriteImageDescriptor( pmap, cmap );
	}

// write pixels:
	WritePixelmap( cmap->UsedBits(), *pmap );
}


























