/*	Copyright  (c)	Günter Woigk 2007 - 2007
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	The LZW encoder and decoder were written 
	by Gershon Elber and Eric S. Raymond 
	as part of the GifLib package.

	To my best knowing the copyright on the algorithm has expired.
*/


#include "kio/kio.h"
#include "zxsp_types.h"
#include "unix/file_utilities.h"
#include "Pixelmap.h"
#include "Colormap.h"


#define HT_SIZE         8192    /* 13 bit hash table size */
#define HT_KEY_MASK     0x1FFF  /* 13 bit key mask */

#define LZ_MAX_CODE     4095    /* Largest 12 bit code */
#define LZ_BITS         12

#define FLUSH_OUTPUT    4096    /* Impossible code = flush */
#define FIRST_CODE      4097    /* Impossible code = first */
#define NO_SUCH_CODE    4098    /* Impossible code = empty */

#define IMAGE_LOADING   0       /* file_state = processing */
#define IMAGE_SAVING    0       /* file_state = processing */
#define IMAGE_COMPLETE  1       /* finished reading or writing */



/* misc. block flags:
*/
enum	// Screen Descriptor:
{
	has_global_cmap	 = 1<<7,
	glob_cmap_sorted = 1<<3,
	img_cmap_bits	 = 1<<4,	// bits 4-6
	glob_cmap_bits	 = 1<<0		// bits 0-2
};
enum	// Sub Image Descriptor:
{
	has_local_cmap	 = 1<<7,
	loc_cmap_sorted	 = 1<<5,
	rows_interlaced	 = 1<<6,
	loc_cmap_bits	 = 1<<0		// bits 0-2
};
enum	// Graphic Control Block:  ((Gif89a++))
{
	no_disposal		 = 0<<2,	// don't restore: do as you like..
	keep_new_pixels  = 1<<2,	// don't restore: keep new pixels
	restore_bgcolor  = 2<<2,	// restore to background color
	restore_pixels   = 3<<2,	// restore previous pixels
	no_transparency	 = 0x00,
	has_transparency = 0x01
};


class GifEncoder
{
	int			fd;
	Colormap	global_cmap;		// if not used, then global_cmap.used = 0

	int			depth, clear_code, eof_code, running_code, running_bits,
				max_code_plus_one, current_code, shift_state,
				file_state, position, bufsize;
    ulong		shift_data;
    uchar		buf[256];

    ulong		hash_table[HT_SIZE];
	void		clear_hash_table	();
	void		add_hash_key		( ulong key, int code );
	int			lookup_hash_key		( ulong key );

	void		write_gif_code		( int );
	void		write_gif_byte		( int );


public:
	GifEncoder()			{ fd=-1; }
	~GifEncoder()			{ if(fd!=-1) CloseFile(); }
			

/*	open file and write gif file header
*/
	void OpenFile			( cstr path );
	bool ImageInProgress	()		{ return fd!=-1; }

/*	Clear global cmap and write logical screen descriptor:
*/
	void WriteScreenDescriptor ( ushort w, ushort h, ushort colors, uchar aspect=0 )
	{
		XXASSERT( colors>0 );
		global_cmap.PurgeCmap();
		uchar flags = (Bits(colors)-1)<<4;
		uchar bu[] = { w,w>>8, h,h>>8, flags, 0/*bgcolor*/, aspect };
		write_bytes( fd, bu, sizeof(bu) );
	}

/*	Set global cmap, write logical screen descriptor and global cmap:
*/
	void WriteScreenDescriptor ( ushort w, ushort h, cColormap& cmap, uchar aspect=0 )
	{
		XXASSERT( cmap.UsedColors()>0 );
		global_cmap = cmap; 
		uchar bits  = cmap.UsedBits() -1;	
		uchar flags = has_global_cmap + (bits<<4) + (bits<<0); 
		uchar bu[] = { w,w>>8, h,h>>8, flags, 0/*bgcolor*/, aspect };
		write_bytes( fd, bu, sizeof(bu) );
		WriteColormap(cmap);
	}

/*	FinishImage()  (if not yet done),
	write comment block with creator message, 
	write trailer byte and close file
*/
	void CloseFile			();


/*	Compression of image pixel stream:
*/
	void StartImage			( int bits_per_pixel );
	void WritePixelRow		( cuptr pixel, uint w );
	void WritePixelRect		( cuptr pixel, uint w, uint h, uint dy=0 );
	void FinishImage		();
	void WritePixelmap		( int bits_per_pixel, cPixelmap& pixel );


/*	Write different kinds of blocks:
*/
	void WriteGif87aHeader	()		{ write_bytes( fd, "GIF87a", 6 ); };
	void WriteGif89aHeader	()		{ write_bytes( fd, "GIF89a", 6 ); };
	void WriteGifTrailer	()		{ write_uchar( fd, 0x3B ); };


/*	Write descriptor for following sub image:
*/
	void WriteImageDescriptor ( ushort x, ushort y, ushort w, ushort h, uchar flags=0 )
	{
		uchar bu[] = { 0x2c, x,x>>8, y,y>>8, w,w>>8, h,h>>8, flags };
		write_bytes( fd,bu,sizeof(bu) );
	}
	void WriteImageDescriptor ( cPixelmap& pm, uchar flags=0 )	{ WriteImageDescriptor( pm.X1(), pm.Y1(), pm.W(), pm.H(), flags ); }
	void WriteImageDescriptor ( cPixelmap* pm, uchar flags=0 )	{ WriteImageDescriptor( *pm, flags ); }
	void WriteImageDescriptor ( cPixelmap* pm, cColormap* cmap )	
	{
		XXASSERT( cmap->CmapSize()>=2 );
		WriteImageDescriptor( *pm, has_local_cmap + cmap->CmapBits()-1 ); 
		WriteColormap( cmap );
	}


/*	Write colormap
	Colormap size must be 2^n colors (1≤n≤8) 
	((not tested. will silently write nothing for cmap_size==0.))
*/
	void WriteColormap ( cColormap& cmap )	{ write_bytes( fd, cmap.GetCmap(), cmap.CmapByteSize() ); }
	void WriteColormap ( cColormap* cmap )	{ WriteColormap(*cmap); }


/*	Enable looping animations.								GIF89a++
*/
	void WriteLoopingAnimationExtension ( ushort max_loops = 0 )	
	{
		uchar bu[] = { 0x21, 0xff, 11, 'N','E','T','S','C','A','P','E','2','.','0', 
					   3/*sub block size*/, 1, max_loops,max_loops>>8, 0 };
		write_bytes( fd, bu, sizeof(bu) );
	}


/*	Write Comment Block										GIF98a++
*/	
	void WriteCommentBlock ( cstr comment )					
	{
		int n = strlen(comment); XXXLIMIT(n,255);
		uchar bu[] = {0x21,0xfe,n}; write_bytes(fd,bu,3);
		write_bytes(fd,comment,n+1);
	}


/*	Write Graphic Control Block:							GIF89++
	Animation delay and transparent color.		
*/
	void WriteGraphicControlBlock ( ushort delay/*1/100sec*/, int transp_index = Colormap::unset )
	{	
		uchar flags = keep_new_pixels + (transp_index!=Colormap::unset);
		uchar bu[] = { 0x21, 0xf9, 4, flags, delay,delay>>8, transp_index, 0 };
		write_bytes( fd, bu, sizeof(bu) );				
	}	


/*	Write sub image to gif file
	use global_cmap or include a local table
*/
	void WriteImage ( Pixelmap* pm, Colormap* cm );
	void WriteImage ( Pixelmap& pm, Colormap& cm )	{ WriteImage(&pm,&cm); }
	void WriteImage ( Pixelmap& pm )				{ WriteImage(&pm,&global_cmap); }
};
















