/*	Copyright  (c)	Günter Woigk 2007 - 2007
					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	
	template class Buffer 
	---------------------

	kio 2007-02-19
	kio	2007-06-06		now using Array<T> for storing the data
						=> copy creator and operator=() do no longer move() but create shared instances
						=> use Unshare() before writing to Buffer if Buffer may be shared

    data type must not have constructor/destructor!
	
Method overview:
	
	create, grow, shrink and append to buffer
	optimized for successive resizing
	access with operator[] syntax
	access with file syntax: read, write, pos
	operator= and copy creator create shared data instances => use Unshare() if required
	
	
Exceptions:
	
	none				if possible 
	std::bad_alloc		where resizing/growing occurs
	index_error			where indexes are validated: operator[] and read	((this is a subclass of internal_error)) 
	internal_error		constructor Buffer(used,max)	if used>max

Methods:

    Buffer()			create empty buffer
    Buffer(n)			create buffer with n elements (cleared to 0)
    Buffer(max,n)		allocate max elements, but initially use only n. (cleared to 0)
    ~Buffer()			destructor

	operator=			creates shared data[] 
	Buffer(Buffer&)     creates shared data[] 

	bool IsShared()		ask whether data is shared
	Unshare()			unshare data if shared

	T* Data()			access buffer
	n  Size()			get used size
	n  Allocated() 		get allocated size

	T& operator[n]		access element n.		note: n must point inside used buffer.
    
	Grow	(newsize)	grow size.				allocates newsize+padding. cleares buffer to 0.
	Shrink  (newsize)	shrink size.			reallocates buffer only if there is substantial gain.
	Resize	(newsize)	shrink or grow.			deprecated. use Grow() or Shrink(). cleares buffer to 0.
	Allocate(newmax)	grow allocated buffer.	reallocates buffer. buffer size and data are preserved.
	ShrinkToFit()
    Purge	()			purge data. truncate to 0.
    Clear	(T)			set all elements to T.
	Clear	()			set all elements to 0.

	Append	(T*,n)		append data from T*.
	Append	(Buffer&)	append data from other Buffer. (copied)
    Append	(T)			append single element.
    Append	(T,n)		append n elements.	

	Insert	(i,n)		insert n elements at i.  (cleared to 0)
	Delete	(i,n)		delete n elements at i

// using Read() and Write() like a file:
	n&	Pos  ()			access r/w ptr for reading or setting. may point beyond buffer end.
		Write(T)		store one element *ptr++
		Write(T*,n)		store block of data *ptr++
		Write(T,n)		store n elements
	T&	Read ()			read one element *ptr++.	note: ptr must point inside used buffer!
	T&	Peek ()			read one element *ptr.	    note: ptr must point inside used buffer!
	T&	rRead()			read one element *--ptr.	note: ptr-1 must point inside used buffer!
	n	Read (T*,n)		read block of data *ptr++.	returns amount actually read before buffer end. 
						returns 0 at or beyond end of buffer.
	n   Avail()			returns elements available for reading *ptr++
	if  Avail(n)		are n elements available for reading *ptr++ ?
	if	AtEnd()			is ptr at or beyond end of buffer?	((ptr>=size))
	if	BeyondEnd()		is ptr beyond end of buffer?		((ptr>size))
*/


#ifndef Buffer_h
#define Buffer_h

#include "kio/kio.h"
#include "SharedArray.h"



template<class T>
class Buffer
{
	SharedArray<T>	data;
	ulong		used;
	ulong		max()		{ return data.Count(); }
	ulong		ptr;

	void		shrink		( ulong newmax )		throw()				{ data.Shrink(newmax); if(used>newmax) used = newmax; }
	void		grow		( ulong newmax )		throw(bad_alloc)	{ data.Grow(newmax); }	// if(newmax>max)
	void		grow_pp		( ulong newmax )		throw(bad_alloc)	{ if(newmax>max())      grow((newmax+2)/3*4); }
	void		shrink_pp	( ulong newmax )		throw()				{ if(newmax<=max()/4*3) shrink(newmax); }
		
public:
				Buffer		( )						throw()				: data(), used(0), ptr(0) {}
				Buffer		( ulong n )				throw(bad_alloc);
				Buffer		( ulong m, ulong n )	throw(bad_alloc);
				~Buffer		( )						throw()				{}
    Buffer&		operator=	( Buffer const& q )		throw();
                Buffer		( Buffer const& q )		throw();

	T*			Data		( )						throw()				{ return data.Data(); }
	T const*	Data		( )						const throw()		{ return data.Data(); }
	ulong		Size		( )						const throw()		{ return used; }	// used size
	ulong		Allocated	( )						const throw()		{ return max();  }	// allocated size

	bool		IsShared	( )						const throw()		{ return data.IsShared(); }
	void		Unshare		( )						throw(bad_alloc)	{ data.Unshare(); }

	T&			operator[]	( ulong index )			throw(index_error)	{ return data[index]; }
	T const&	operator[]	( ulong index )			const throw(index_error){ return data[index]; }

	void		Grow		( ulong newused )		throw(bad_alloc)	{ if(newused>used) grow_pp((used=newused)); }
	void		Shrink		( ulong newused )		throw()				{ if(newused<used) shrink_pp((used=newused)); }
	void		Resize		( ulong newused )		throw(bad_alloc)	{ grow_pp(newused);shrink_pp(newused);used=newused; }
	void		Allocate	( ulong newmax )		throw(bad_alloc)	{ grow(newmax); }
	void		ShrinkToFit	( )						throw(bad_alloc)	{ shrink(used); }
    void		Purge		( )						throw()				{ data.Purge(); used = ptr = 0; }
    void		Clear		( T a )					throw()				{ data.Clear(a); }
    void		Clear		( )						throw()				{ data.Clear(/*with 0*/); }

	void		Append		( T const* q, ulong n )	throw(bad_alloc);
	void		Append		( Buffer const& q )		throw(bad_alloc)	{ Append(q.data, q.used); }
    void		Append		( T a )					throw(bad_alloc)	{ grow_pp(used+1); data[used++] = a; }
	void		Append		( T a, ulong n )		throw(bad_alloc);

	void		Insert		( ulong idx, ulong n )	throw(bad_alloc);
	void		Delete		( ulong idx, ulong n )	throw();

// using Read() and Write() like a file:
	void		Write		( T )					throw(bad_alloc);
	void		Write		( T const*, ulong n )	throw(bad_alloc);
	void		Write		( T, ulong n )			throw(bad_alloc);
	ulong		Read		( T*, ulong n )			throw();
	T&			Read		( )						throw(index_error)	{ return operator[](ptr++); }
	T&			rRead		( )						throw(index_error)	{ return operator[](--ptr); }
	T&			Peek		( )						throw(index_error)	{ return operator[]( ptr ); }
	ulong&		Pos			( )						throw()				{ return ptr; }
	ulong		Pos			( )						const throw()		{ return ptr; }
	ulong		Avail		( )						const throw()		{ return used>=ptr ? used-ptr : 0; }
	bool		Avail		( ulong n )				const throw()		{ return ptr+n <= used; }
	bool		AtEnd		( )						const throw()		{ return ptr >= used; }
	bool		BeyondEnd	( )						const throw()		{ return ptr >  used; }
};

 
template<class T>
inline Buffer<T>::Buffer ( ulong n ) throw(bad_alloc)	
: data(n), used(n), ptr(0) 
{ 
}


template<class T>
inline Buffer<T>::Buffer ( ulong m, ulong n ) throw(bad_alloc)
: data(m), used(n), ptr(0) 
{
	XXXTRAP(n>m); 
}

 
template<class T>
inline void Buffer<T>::Append ( T const* q, ulong n ) throw(bad_alloc)
{
	if(n)
	{
		grow_pp(used+n);
		memcpy( data.Data()+used, q, n*sizeof(T) );
		used += n;
	}
}


template<class T>
inline void Buffer<T>::Append ( T a, ulong n ) throw(bad_alloc)
{
	if(n)
	{
		grow_pp(used+n);
        T* p = data.Data()+used;
        T* e = p+n;
        while(p<e) { *p++ = a; }
		used += n;
	}
}


template<class T>
Buffer<T>&	Buffer<T>::operator= ( Buffer const& q ) throw()			
{
	data = q.data;
	used = q.used;	
	ptr  = q.ptr;	
    return *this;
}


template<class T>
inline Buffer<T>::Buffer ( Buffer const& q ) throw()			
:	data(q.data)
{
    used = q.used;
	ptr  = q.ptr;
}


template<class T>
void Buffer<T>::Write ( T a ) throw(bad_alloc)
{
	grow_pp(ptr+1);
	data[ptr++] = a;
	if(ptr>used) used=ptr;
}


template<class T>
void Buffer<T>::Write ( T const* p, ulong n ) throw(bad_alloc)
{
	if(n)
	{
		grow_pp(ptr+n);
		memcpy( data.Data()+ptr, p, sizeof(T)*n );
		ptr += n;
		if(ptr>used) used=ptr;
	}
}

template<class T>
void Buffer<T>::Write ( T a, ulong n ) throw(bad_alloc)
{
	if(n)
	{
		grow_pp(ptr+n);
        T* p = data.Data()+ptr;
        T* e = p+n;
        while(p<e) { *p++ = a; }
		ptr += n;
		if(ptr>used) used=ptr;
	}
}


// read via ptr:
// returns bytes_read which may be less than requested
template<class T>
ulong Buffer<T>::Read ( T* bu, ulong n ) throw()
{
	if(ptr<used)
	{
		if (used-ptr<n) n=used-ptr;
		T* z = bu;
		T* q = data.Data()+ptr;
		ptr += n;
		while(n--) { *z++ = *q++; }
		return z-bu;
	}
	
	return 0;	// beyond end
}


template<class T>
void Buffer<T>::Insert ( ulong idx, ulong cnt )	throw(bad_alloc)
{
	if( idx<used )
	{
		Grow(used+cnt);					// bad_alloc
		memmove( data.Data()+idx+cnt, data.Data()+idx, (used-idx-cnt)*sizeof(T) );
		memset( data.Data()+idx, 0, cnt*sizeof(T) );
	}
	else
	{
		Grow(idx+cnt);					// bad_alloc
	}
}


template<class T>
void Buffer<T>::Delete ( ulong idx, ulong cnt )	throw()
{
	if( idx+cnt < used )
	{
		memmove( data.Data()+idx, data.Data()+idx+cnt, (used-idx-cnt)*sizeof(T) );
		used -= cnt;		
	}
	else if( idx < used )
	{
		used = idx;
	}
}


#endif


