/*	Copyright  (c)	Günter Woigk 2005 - 2005
					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	
	template class Queue 
	--------------------
	
	This Class is appropriate for flat data elements without constructor or destructor.
	It can be set to auto-expand, auto-expand up to a given limit or to static queue size.
	It can be used for thread-thread or thread-signal communication on single-processor systems.

	- Elements T must not have a constructor, destructor and assignment operator.
		memcpy() is used in this implementation.

	- This class is thread-save but not multi-processor save.
		the processor caches lead to reading of outdated data: data members and buffer.
		->  threads on different processors must use mutex locks and
			signals on different processors must be blocked 
			on multi-processor machines both queue ends must be in the same thread.

	- A signal-to-thread queue (where signal is writing) must be of to static size.
		calling new[] and delete[] from a signal handler is probably bad.	  
*/

#ifndef Queue_h
#define Queue_h

#include "kio/kio.h"


template<class T>
class Queue
{
	T*		bu;
	ulong	sz;
	ulong	max_sz;
	ulong	rptr;
	ulong	wptr;

			Queue		( const Queue& );			// prohibit
	Queue&	operator=	( const Queue& );			// prohibit

	ulong	poke		( ulong dptr, T const * qptr, ulong qcnt );
	bool	grow		( ulong new_size );
	void	assert_2eN	( ulong n )					{ if(XXSAFE) while((long)n>0) n<<=1; TRAP(n!=1<<31); }

public:

			Queue		( )							: bu(NULL),		   sz(0),    max_sz(1<<30), rptr(0), wptr(0) {}
			Queue		( ulong size )				: bu(new T[size]), sz(size), max_sz(size),  rptr(0), wptr(0) { assert_2eN(size); }
			Queue		( ulong size, ulong max )	: bu(new T[size]), sz(size), max_sz(max),   rptr(0), wptr(0) { assert_2eN(size); assert_2eN(max); }
			~Queue		( )							{ delete[] bu; }
			
	ulong	Peek		( T* p, ulong n );
	ulong	Read		( T* p, ulong n )			{       n = Peek(p,n); rptr+=n; return n; }
	ulong	Write		( T const * p, ulong n )	{ ulong m = poke(wptr,p,n); if(m<n&&sz<max_sz&&grow(sz+n-m)) { m+=poke(wptr+m,p+m,n-m); }  
																		   wptr+=m; return m; }

	void	Drop		( ulong n )					{ rptr = Min(rptr+n,wptr); }
	void	Flush		( )							{ rptr = wptr; }
	
	ulong	Used		( )							{ return wptr-rptr; }
	ulong	Free		( )							{ return max_sz-(wptr-rptr); }
};


template<class T>
bool Queue<T>::grow ( ulong ncnt )
{
	XXXTRAP(sz==max_sz);
	ulong nsz; T* nbu;
	if(sz)
	{
		nsz = sz; do { nsz<<=1; } while( nsz<ncnt && nsz<max_sz ); 
a:		nbu = new T[nsz]; if(!nbu) { nsz>>=1; if(nsz>sz) goto a; else goto x; }
		for( ulong i=0; i<nsz; i+=sz ) memcpy( nbu+i, bu, sz*sizeof(T) );
		T* z = bu; bu = nbu; sz = nsz; delete[] z; 
	}
	else
	{
		nsz = 1<<4; while( nsz<ncnt ) { nsz<<=1; }
b:		nbu = new T[nsz]; if(!nbu) { nsz>>=1; if(nsz>>4) goto b; else goto x; }
		bu = nbu; sz = nsz; 
	}
x:	return nbu!=NULL;
}


template<class T>
ulong Queue<T>::Peek ( T* dptr, ulong dcnt )
{
	ulong n=0,m=0;							// elements read
	ulong rptr = this->rptr;				// local copy. we are peeking => don't touch real rptr!
	ulong mask = sz-1;

	if( (rptr&mask) >= (wptr&mask) )		// used range of data wraps around buffer end
	{										// => at least 1 byte between rptr and bu.end available for writing
		n = sz - (rptr&mask);
		n = Min(n,dcnt);					// n >= 1
		memcpy( dptr, bu+(rptr&mask), n*sizeof(T) );
		if(n==dcnt) return n; 
		rptr += n; dcnt -= n; dptr += n; 
	}

	if( (rptr&mask) < (wptr&mask) )			// used range of buffer does not wrap around buffer end
	{										// => at least 1 byte between rptr and wptr available for writing
		m = wptr-rptr;
		m = Min(m,dcnt);					// m >= 1
		memcpy( dptr, bu+(rptr&mask), m*sizeof(T) ); 
	}

	return n+m;								// return elements read
}


template<class T>
ulong Queue<T>::poke ( ulong wptr, T* qptr, ulong qcnt )
{
	ulong n=0,m=0;							// items written
	ulong mask = sz-1;

	if( (wptr&mask) >= (rptr&mask) )		// used range of data wraps around buffer end
	{										// => at least 1 byte between rptr and bu.end available for writing
		n = sz - (wptr&mask);
		n = Min(n,qcnt);					// n >= 1
		memcpy( bu+(wptr&mask), qptr, n*sizeof(T) );
		if(n==dcnt) return n; 
		wptr += n; qcnt -= n; qptr += n; 
	}

	if( (wptr&mask) < (rptr&mask) )			// used range of buffer does not wrap around buffer end
	{										// => at least 1 byte between rptr and wptr available for writing
		m = rptr-wptr;
		m = Min(m,qcnt);					// m >= 1
		memcpy( bu+(wptr&mask), qptr, m*sizeof(T) ); 
	}

	return n+m;								// return items read
}



#endif











