/*	Copyright  (c)	Günter Woigk 2007 - 2008
					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	
	template class SharedArray 
	--------------------------

	Completely reworked:	2007-06-07 kio
	Last change:			2007-07-09 kio
	

	Manage a dynamically allocated array of type T
	
	-	delete[] data when Array is destroyed
	-	sharing data
	-	index validation in operator[]
	-	subslicing, resizing and concatenation
	-	Array for caller-supplied memory
	

	Restrictions:

	-	max. number of elements in array = max(int)		
		byte-size of array is not restricted!

	-	data type T must have 'flat' default creator, destructor, and copy creator:

		-	default creator:	data is cleared to 0
		-	destructor:			no action
		-	copy creator:		flat copy


	Shared arrays:
	
	-	copy creator and operator= create shared copies:
	
		->	thereafter data of source and destination are shared
		->	neither source nor destination data are writable (without affecting other shared instances)
		->	use Unshare() to unshare data before overwriting array data
		->	use Unshare() always when in doubt		
		
		Unshare() is only required before overwriting items in the array[].
		Unshare() is not required before overwriting the Array instance itself, e.g. in assignment, resize, etc.
		
	-	when returning a result Array from a procedure:
		-	the result Array is copied
		-	then the source Array is destroyed
		=>	finally the resulting Array is unshared, if the source Array was unshared.
*/


#ifndef SharedArray_h
#define SharedArray_h

#include "kio/kio.h"


typedef class std::bad_alloc bad_alloc;


template<class T>
class SharedArray
{
/*	data members:
*/
	ptr		data;		// allocated memory. Memory starts with a lock (share counter). 
						// data==NULL for user-supplied memory. data *may* be NULL for empty array.
	T*		array;		// -> array[0]. Supports subslicing.    array *may* be NULL for empty array.
	int		count;		// item count.	Supports subslicing.    count==0 for empty array.

	enum {	LOCK_SIZE = sizeof(T) <= sizeof(int)   ? sizeof(int) :
						_ALIGNMENT_REQUIRED==0     ? sizeof(int) : 
						sizeof(T) < _MAX_ALIGNMENT ? sizeof(T)   : _MAX_ALIGNMENT };


/*	private methods:
	instance-independent:
				lock		cast malloc address into a reference to the shared_counter 
				retain		increment the shared_counter 
				release		decrement the shared_counter and deallocate data if required
				malloc		allocate memory for n items and init shared_counter
				sz			convert item count to memory size
				copy		copy n items
				clear		clear n items with 0
*/
static int&		lock		( ptr p )			throw()				{ return *(int*)p; }
static void		retain		( ptr p )			throw()				{ if( p ) lock(p)++; }
static void		release		( ptr p )			throw()				{ if( p && --lock(p)<0 ) delete[] p; } 
static ptr		malloc		( int n )			throw(bad_alloc)	{ ptr p = new char[sz(n)+LOCK_SIZE]; lock(p)=0; return p; }
static size_t	sz			( size_t n )		throw()				{ return n*sizeof(T); }
static void		copy		( T*z, T const*q, int n ) throw()		{ memcpy(z,q,sz(n)); }
static void		clear		( T*z, int n )		throw()				{ memset(z,0,sz(n)); }

/*	private methods:
	init and kill:
			init()				init empty array
			init(count)			malloc and clear array
			init(SharedArray&)		init shared copy
			init(T*,count)		init with caller-supplied array
			kill				release data and deallocate malloced data if required.
*/
	void	init		( )					throw()					{ data = NULL; array = NULL; count = 0; /* retain(data) */ }
	void	init		( int n )			throw(bad_alloc);
	void	init		( SharedArray const &q ) throw()			{ data = q.data; array = q.array; count = q.count; retain(data); }
	void	init		( T* q, int n )		throw()					{ data = NULL; array = q; count = n>0?n:0; /* retain(data) */ }

	void	kill		( )					throw()					{ release(data); }


public:

/*	creator, destructor:
			SharedArray()				default constructor: empty array
			SharedArray(SharedArray&)		copy constructor: shared data
			SharedArray(T*,count)		use caller-supplied memory: sharing cannot be tracked!
			SharedArray(T const*,cnt)	use caller-supplied memory: sharing cannot be tracked & does not reflect writability!
			SharedArray(count)		array with malloced memory, cleared to 0, may throw bad_alloc.
			SharedArray(cnt,nothrow)	array with malloced memory, cleared to 0, empty array and errno set if malloc fails.
			~SharedArray()			destructor
			operator=(SharedArray&)	share data
			operator=(T)		create array with one item. deprecated.
*/
			SharedArray		( )					throw()				{ init(); }		

			SharedArray		( SharedArray const &q ) throw()		{ init(q); }
			SharedArray		( SharedArray const *q ) throw()		{ init(*q); }
			SharedArray		( T* q, int n )		throw()				{ init(q,n); }
			SharedArray		( T const*q, int n )throw()				{ init(const_cast<T*>(q),n); }
			SharedArray		( int n )			throw(bad_alloc)	{ init(n); }	
			SharedArray		( int n, nothrow_t )throw()				{ try{init(n);}catch(bad_alloc){ForceError(outofmemory);} }

			~SharedArray		( )				throw()				{ kill();  }

	SharedArray&	operator=	( SharedArray const &q ) throw()	{ if( this!=&q ) { kill(); init(q); } return *this; }
	SharedArray&	operator=	( T q )			throw(bad_alloc)	{ kill(); init(&q,1); return *this;  }


/*	access data members:
			Count()		get item count
			Data()		get array base address
			operator[i]	access item i. checks index during development.
*/
	uint	Count		( )			  const throw()					{ return count; }
	T*		Data		( )					throw()					{ return array; }
	T const*Data		( )			  const throw()					{ return array; }
	T&		operator[]	( int i )			throw(index_error)		{ XXXINDEX(i,count); return array[i]; }
	T const&operator[]	( int i )	  const throw(index_error)		{ XXXINDEX(i,count); return array[i]; }


/*	sharing data:
			IsShared	returns YES if calling Unshare() makes sense:
							if the array is malloced AND shared:			generically shared
							if the array is malloced AND empty:				unsharing and purging is useful
							if the array is caller-supplied AND not empty:	sharing cannot be determined!!
			UnShare		unshare shared array
						reallocate user-supplied memory
						reallocate subsliced array
						unshare and purge empty array
			Reallocate	unconditionally reallocte array, even if empty or unshared.
*/	
	bool	IsShared	( ) const throw()		{ return data ? count==0 || lock(data)>0 : count; }
	void	Unshare		( )	throw(bad_alloc)	{ if( count ) Reallocate(); else Purge(); }
	void	Reallocate	( )	throw(bad_alloc);


/*	subslicing:
			SubSlice	set array to subarray [i .. [i+n			a=i and e=i+n are forced inside [0 .. [count
			SubArray	return subarray [t .. [i+n					a=i and e=i+n are forced inside [0 .. [count
			Range		return subarray [a .. [e					a and e are forced inside [0 .. [count
			Left		return subarray [0 .. [n					n is forced inside [0 .. [count
			Right		return subarray [count-n .. [count			n is forced inside [0 .. [count
*/	
	void		SubSlice	( int i, int n )	throw();				
	SharedArray	SubArray	( int i, int n )	throw();
	SharedArray	Range		( int a, int e )	throw()					{ return SubArray(a,e-a); }
	SharedArray	Left		( int n )			throw()					{ return SubArray(0,n); }
	SharedArray	Right		( int n )			throw()					{ return SubArray(count-n,n); }

/*	resizing and clearing:
			Shrink		shrink if new_size < count
			Grow		grow   if new_size > count
			Purge		purge data and resize to 0
			Clear		overwrite array with 0s. does not check for shared data! => can be used for: Range(a,e).Clear()
			Clear(T)	overwrite array with Ts. does not check for shared data! => ""
*/
	void	Shrink		( int new_size )	throw();				
	void	Grow		( int new_size )	throw(bad_alloc);	
	void	Purge		( )					throw()					{ kill(); init(); }
	void	Clear		( )					throw()					{ clear( array, count ); }
	void	Clear		( T q )				throw()					{ T*p=array; T*e=p+count; while(p<e) *p++=q; }

/*	concatenation:
			Append		append user-supplied data to my array. if bad_alloc is thrown, then Array remained unmodified.
			operator+=	append other array to my array. if bad_alloc is thrown, then Array remained unmodified.
			operator=	return concatenation of my array and other array. throws bad_alloc if malloc fails.
*/
	SharedArray&	Append		( T const* q, int n )	 throw(bad_alloc);
	SharedArray&	operator+=	( SharedArray const &q ) throw(bad_alloc)		{ return Append(q.array,q.count); }
	SharedArray		operator+	( SharedArray const &q ) const throw(bad_alloc)	{ return SharedArray(this).Append(q.array,q.count); }
};



template<class T>		/* verified 2006-06-07 kio */
SharedArray<T>&
SharedArray<T>::Append ( T const* q, int n ) throw(bad_alloc)
{
/*	Append caller-supplied items to my array.
	does nothing if item count == 0.
	can be called to append subarray of own array!
	if bad_alloc is thrown, then this Array remained unmodified.
*/
	if(n>0)
	{
		ptr old_data  = data;
		T*  old_array = array;
		int old_size  = count;
	
		data  = malloc(old_size+n);
		array = (T*)(data+LOCK_SIZE); 
		count  = old_size+n;
	
		copy( array, old_array, old_size );	// copy own data
		copy( array+old_size, q, n );		// copy other data. may be subarray of my own array!
		release(old_data);					// => release own data after copying other data!
	}
	return *this;
}


#if 0
template<class T>
SharedArray<T>&
SharedArray<T>::operator+= ( SharedArray<T> const &q ) const throw(bad_alloc)
{
	ptr old_data  = data;
	T*  old_array = array;
	int old_size  = count;
	
	data  = malloc(old_size+q.count);
	array = (T*)(data+LOCK_SIZE); 
	count  = old_size+q.count;
	
	copy( array, old_array, old_size );
	copy( array+old_size, q.array, q.count );
	release(old_data);
	
	return *this;
}
#endif


#if 0
template<class T>		/* verified 2006-06-07 kio */
SharedArray<T>
SharedArray<T>::operator+ ( SharedArray<T> const &q ) const throw(bad_alloc)
{
	SharedArray<T> z;
		z.data  = malloc(count+q.count);
		z.array = (T*)(z.data+LOCK_SIZE); 
		z.count  = count+q.count;
	
	copy( z.array, array, count );
	copy( z.array+count, q.array, q.count );

	return z;
}
#endif


template<class T>		/* verified 2006-06-07 kio */
void
SharedArray<T>::SubSlice ( int i, int n ) throw()
{
	if(i<0) { n+=i; i=0; } if(i+n>count) n=count-i; 
	if(n<=0) { Purge(); } else { array+=i; count=n; } 
}


template<class T>		/* verified 2006-06-07 kio */
SharedArray<T>
SharedArray<T>::SubArray ( int i, int n ) throw()
{
	SharedArray<T> z = *this;
	z.SubSlice(i,n);
	return z;
}


template<class T>		/* verified 2006-06-07 kio */
void 
SharedArray<T>::init ( int n ) throw(bad_alloc)	
{
	XXXASSERT( n>=0 );
	init();				// in case malloc() throws bad_alloc

/*	Allocate and clear Array:

	note:
		The caller should not allocate the array itself with "new SharedArray(n)"
		but prefer static/global/local instances. The Array body is only 12 bytes.

	bad_alloc:
		If bad_alloc is thrown, then the SharedArray is initialized to an empty array
			and can be destroyed by the caller, e.g. by proper function exit after a "catch",
			or can be left unhandled, e.g. if the exception is not catched in the calling function.

		If multiple SharedArrays are instantiated in a function, then it is unclear after a bad_alloc exception
		which constructor failed and thus which SharedArrays were left uninitialized.
		=>	create multiple SharedArrays with the nothrow option and check errno or array sizes thereafter.
		I'm not shure how it is handled if one of the local instantiation fails and the exception is not catched.
		I'm not shure how it is handled if one of the local instantiation fails and the exception _is_ catched...
*/
	data  = malloc(n);
	array = (T*)(data+LOCK_SIZE); 
	count  = n;
	clear( array, n );
}	


#if 0
template<class T>		/* verified 2006-06-07 kio */
void 
SharedArray<T>::init ( T const *q, int n ) throw(bad_alloc)
{
	XXXASSERT( n>=0 );
	init();				// in case malloc() throws bad_alloc

	data  = malloc(n);
	array = (T*)(data+LOCK_SIZE); 
	count  = n;
	copy( array, q, n );
}
#endif
	

template<class T>		/* verified 2006-06-07 kio */
void 
SharedArray<T>::Reallocate ( ) throw(bad_alloc)		
{
/*	Unshare a shared array
	or reallocate a subsliced array
	or reallocate a caller-supplied array
	
	does not test whether shuffling around memory makes sense
	e.g.	reallocating an unshared array or
			reallocating an empty (count==0) array

	if bad_alloc is thrown, then Array remained unmodified
*/
	ptr old_data  = data;
	T*  old_array = array;
	
	data  = malloc(count);
	array = (T*)(data+LOCK_SIZE); 
	count  = count;
	copy( array, old_array, count );
	
	release(old_data);
}	


template<class T>		/* verified 2006-06-07 kio */
void 
SharedArray<T>::Shrink ( int new_size ) throw()
{
/*	Shrink Array:
	does nothing if requested new_size >= current count
	purges ( deallocates ) data if new_size == 0.
	silently handles new_size < 0 as new_size == 0.
	does not reallocate data if array points to caller-supplied data.
	silently keeps the old data if data reallocation fails.
	note: this call may or may not unshare the Array.
*/
	if( new_size>=count ) return;			// won't shrink
	if( new_size<=0 ) { Purge(); return; }	// empty array

	count = new_size;
	if( data )						// data==0 && array!=0  =>  caller-supplied data
	{
		try { Reallocate(); }		// try to reallocate data
		catch(bad_alloc){}		// but just don't do it if reallocation fails
	}
}


template<class T>		/* verified 2006-06-07 kio */
void 
SharedArray<T>::Grow ( int new_size ) throw(bad_alloc)	
{
/*	Grow Array:
	does nothing if requested new_size <= current count.	
	if bad_alloc is thrown, then Array remained unmodified.
*/
	if( new_size<=count ) return;

	int old_size  = count;
	ptr old_data  = data;
	T*  old_array = array;

	data  = malloc(new_size);
	array = (T*)(data+LOCK_SIZE); 
	count  = new_size;

	copy( array, old_array, old_size );
	clear( array+old_size, new_size-old_size );
	
	release(old_data);
}	


#endif











