/*	Copyright  (c)	Günter Woigk 2007 - 2008
					mailto:kio@little-bat.de

	This file is free software

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

	Redistribution and use in source and binary forms, with or without 
	modification, are permitted provided that the following conditions are met:

	• Redistributions of source code must retain the above copyright notice, 
	  this list of conditions and the following disclaimer.
	• Redistributions in binary form must reproduce the above copyright notice, 
	  this list of conditions and the following disclaimer in the documentation 
	  and/or other materials provided with the distribution.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
	AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
	THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
	PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR 
	CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
	EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
	OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
	ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
	
	template class SharedArray 
	--------------------------

	Completely reworked:	2007-06-07 kio
	Last change:			2007-07-09 kio
	

	Manage a dynamically allocated array of type T
	
	-	delete[] data when Array is destroyed
	-	sharing data
	-	index validation in operator[]
	-	subslicing, resizing and concatenation
	-	Array for caller-supplied memory
	

	Restrictions:

	-	max. number of elements in array = max(int)		
		byte-size of array is not restricted!

	-	data type T must have 'flat' default creator, destructor, and copy creator:

		-	default creator:	data is cleared to 0
		-	destructor:			no action
		-	copy creator:		flat copy


	Shared arrays:
	
	-	copy creator and operator= create shared copies:
	
		->	thereafter data of source and destination are shared
		->	neither source nor destination data are writable (without affecting other shared instances)
		->	use Unshare() to unshare data before overwriting array data
		->	use Unshare() always when in doubt		
		
		Unshare() is only required before overwriting items in the array[].
		Unshare() is not required before overwriting the Array instance itself, e.g. in assignment, resize, etc.
		
	-	when returning a result Array from a procedure:
		-	the result Array is copied
		-	then the source Array is destroyed
		=>	finally the resulting Array is unshared, if the source Array was unshared.
*/


#ifndef SharedArray_h
#define SharedArray_h

#include "kio/kio.h"


typedef class std::bad_alloc bad_alloc;


template<class T>
class SharedArray
{
/*	data members:
*/
	ptr		data;		// allocated memory. data==NULL for user-supplied memory. data *may* be NULL for empty array.
	T*		array;		// -> array[0]. Supports subslicing.    array *may* be NULL for empty array.
	int		count;		// item count.	Supports subslicing.    count==0 for empty array.
	SharedArray* next;	// linked list of shared arrays.


/*	private methods:
	instance-independent:
			copy	copy n items
			clear	clear n items with 0
*/
static void	copy	( T* z, T const* q, int n ) throw()	{ memcpy(z,q,n*sizeof(T)); }
static void	clear	( T* z, int n )				throw()	{ memset(z,0,n*sizeof(T)); }

/*	private methods:
	init and kill:
			init()				init with empty array
			init(count)			allocate and clear new array
			init(SharedArray&)	init with shared copy
			init(T*,count)		init with and take possession of supplied array
			kill				unlink and delete data if unshared
*/
	void	init	( )					throw()			{ data = array = NULL; count = 0; next = this; }
	void	init	( T* q, int n )		throw()			{ data = array = q;    count = n; next = this; }
	void	init	( SharedArray& q )	throw()			{ data=q.data; array=q.array; count=q.count; next=q.next; q.next=this; }
	void	init	( int n )			throw(bad_alloc);
	void	kill	( )					throw();


public:

/*	creator, destructor:
			SharedArray()				default constructor: empty array
			SharedArray(SharedArray&)	copy constructor: shared data
			SharedArray(T*,count)		use caller-supplied memory
			SharedArray(T const*,cnt)	use caller-supplied memory
			SharedArray(count)			array with new memory, cleared to 0, may throw bad_alloc.
			SharedArray(cnt,nothrow)	array with new memory, cleared to 0, empty array and errno set if alloc fails.
			~SharedArray()				destructor
			operator=(SharedArray&)		shared data
*/
			SharedArray		( )					throw()				{ init(); }		

			SharedArray		( SharedArray const &q ) throw()		{ init( const_cast<SharedArray&>(q) ); }
			SharedArray		( SharedArray const *q ) throw()		{ init( const_cast<SharedArray&>(*q) ); }
			SharedArray		( T* q, int n )		throw()				{ init(q,n); }
			SharedArray		( T const*q, int n )throw()				{ init(const_cast<T*>(q),n); }
			SharedArray		( int n )			throw(bad_alloc)	{ init(n); }
			SharedArray		( int n, nothrow_t )throw()				{ try{ init(n); }catch(bad_alloc){ errno=outofmemory; } }

			~SharedArray	( )					throw()				{ kill(); }

	SharedArray& operator=	( SharedArray const& q ) throw()		{ if( this!=&q ) { kill(); init(const_cast<SharedArray&>(q)); } return *this; }
	SharedArray	Copy		( )					throw(bad_alloc)	{ return SharedArray(this).Reallocate(); }


/*	access data members:
			Count()		get item count
			Data()		get array base address
			operator[i]	access item i. checks index during development.
*/
	int		Count		( )			  const throw()					{ return count; }
	T*		Data		( )					throw()					{ return array; }
	T const*Data		( )			  const throw()					{ return array; }
	T&		operator[]	( int i )			throw(index_error)		{ XXXINDEX(i,count); return array[i]; }
	T const&operator[]	( int i )	  const throw(index_error)		{ XXXINDEX(i,count); return array[i]; }


/*	sharing data:
			IsShared	returns YES if calling Unshare() makes sense:
							if the array is malloced AND shared:			generically shared
							if the array is malloced AND empty:				unsharing and purging is useful
							if the array is caller-supplied AND not empty:	sharing cannot be determined!!
			UnShare		unshare shared array
						reallocate user-supplied memory
						reallocate subsliced array
						unshare and purge empty array
			Reallocate	unconditionally reallocte array, even if empty or unshared.
*/	
	bool	IsShared	( ) const throw()		{ return next!=this; }
	void	Unshare		( )	throw(bad_alloc)	{ if( IsShared() ) Reallocate(); }
	SharedArray<T>&	Reallocate	( )	throw(bad_alloc);


/*	subslicing:
			MidSet		set array to subarray [i .. [i+n			a=i and e=i+n are forced inside [0 .. [count
			MidArray	return subarray [i .. [i+n					a=i and e=i+n are forced inside [0 .. [count
			SubArray	return subarray [a .. [e					a and e are forced inside [0 .. [count
			LeftArray	return subarray [0 .. [n					n is forced inside [0 .. [count
			RightArray	return subarray [count-n .. [count			n is forced inside [0 .. [count
*/	
	SharedArray& MidSet		( int i, int n )	throw();
	SharedArray	MidArray	( int i, int n )	throw()				{ return SharedArray(this).MidSet(i,n); }
	SharedArray	SubArray	( int a, int e )	throw()				{ return MidArray(a,e-a); }
	SharedArray	LeftArray	( int n )			throw()				{ return MidArray(0,n); }
	SharedArray	RightArray	( int n )			throw()				{ return MidArray(count-n,n); }

/*	resizing and clearing:
			Shrink		shrink if new_count < count
			Grow		grow   if new_count > count
			Purge		purge data and resize to 0
			Clear		overwrite array with 0s. does not check for shared data! => can be used for: Range(a,e).Clear()
			Clear(T)	overwrite array with Ts. does not check for shared data! => ""
*/
	void	Shrink		( int new_count )	throw();				
	void	Grow		( int new_count )	throw(bad_alloc);	
	void	Purge		( )					throw()					{ kill(); init(); }
	void	ClearData	( )					throw()					{ clear( array, count ); }
	void	ClearData	( T q )				throw()					{ T*p=array; T*e=p+count; while(p<e) *p++=q; }

/*	concatenation:
			Append		append item to this array.			if bad_alloc is thrown, then this array remained unmodified.
			operator+=	append other array to this array.	if bad_alloc is thrown, then this array remained unmodified.
			operator+	return concatenation of this and other array. throws bad_alloc if alloc fails.
*/
	SharedArray&	Append		( T const* q, int n )	 throw(bad_alloc);
	SharedArray&	operator+=	( SharedArray const& q ) throw(bad_alloc)		{ return Append(q.array,q.count); }
	SharedArray		operator+	( SharedArray const& q ) const throw(bad_alloc)	{ return SharedArray(this).Append(q.array,q.count); }
};



template<class T>
void	
SharedArray<T>::kill ( ) throw()
{
	if( next==this ) { delete[] data; }
	else { T* p=this; while(p->next!=this) { p=p->next; } p->next=next; } 
}


template<class T>
void 
SharedArray<T>::init ( int n ) throw(bad_alloc)	
{
	XXXASSERT( n>=0 );

/*	Allocate and clear Array:

	note:
		The caller should not allocate the array itself with "new SharedArray(n)"
		but prefer static/global/local instances. The Array body is only 12 bytes.

	bad_alloc:
		If bad_alloc is thrown, then the SharedArray is initialized to an empty array
			and can be destroyed by the caller, e.g. by proper function exit after a "catch",
			or can be left unhandled, e.g. if the exception is not catched in the calling function.

		If multiple SharedArrays are instantiated in a function, then it is unclear after a bad_alloc exception
		which constructor failed and thus which SharedArrays were left uninitialized.
		=>	create multiple SharedArrays with the nothrow option and check errno or array sizes thereafter.
		I'm not shure how it is handled if one of the local instantiation fails and the exception is not catched.
		I'm not shure how it is handled if one of the local instantiation fails and the exception _is_ catched...
*/
	init();						// if alloc() throws bad_alloc
	data  = array = new T[n];
	count = n;
	clear( array, n );
}	




template<class T>
SharedArray<T>&
SharedArray<T>::Append ( T const* q, int n ) throw(bad_alloc)
{
/*	Append caller-supplied items to my array.
	does nothing if item count == 0.
	can be called to append subarray of own array!
	if bad_alloc is thrown, then this Array remained unmodified.
*/
	if(n>0)
	{
		T* new_data = new T[count+n];
		copy( new_data, array, count );		// copy own data
		copy( new_data+count, q, n );		// copy other data. may be subarray of my own array!

		kill();
		init(new_data,count+n);
	}
	return *this;
}


template<class T>
SharedArray<T>&
SharedArray<T>::MidSet ( int i, int n ) throw()
{
	if(i<0) { n+=i; i=0; } if(i+n>count) n=count-i; 
	if(n<=0) { Purge(); } else { array+=i; count=n; } 
	return *this;
}


template<class T>
SharedArray<T>&
SharedArray<T>::Reallocate ( ) throw(bad_alloc)		
{
/*	Unshare a shared array
	or reallocate a subsliced array
	or reallocate a caller-supplied array
	
	does not test whether shuffling around memory makes sense
	e.g.	reallocating an unshared array or
			reallocating an empty (count==0) array

	if bad_alloc is thrown, then Array remained unmodified
*/
	if( count<=0 ) Purge();
	else
	{
		T* new_data = new T[count];
		copy( new_data, array, count );
		kill();
		init( new_data, count );
	}
	return *this;
}	


template<class T>		
void 
SharedArray<T>::Shrink ( int new_count ) throw()
{
/*	Shrink Array:
	does nothing if requested new_count >= current count
	purges data if new_count == 0.
	silently handles new_count < 0 as new_count == 0.
	silently keeps the old data if data reallocation fails.
	note: this call may or may not unshare the array.
*/
	if( new_count<count )
	{
		count = new_count;		// limited to >= 0 by Reallocate()
		try { Reallocate(); }	// try to reallocate data
		catch(bad_alloc){}		// but just don't do it if reallocation fails
	}
}


template<class T>
void 
SharedArray<T>::Grow ( int new_count ) throw(bad_alloc)	
{
/*	Grow Array:
	does nothing if requested new_count <= current count.	
	if bad_alloc is thrown, then array remained unmodified.
*/
	if( new_count>count )
	{
		T* new_data = new T[new_count];
		copy( new_data, array, count );
		clear( new_data+count, new_count-count );
		kill();
		init( new_data, new_count );
	}
}	


#endif











