/*	Copyright  (c)	Günter Woigk 2004 - 2008
  					mailto:kio@little-bat.de

 	This program is distributed in the hope that it will be useful,
 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

 	Permission to use, copy, modify, distribute, and sell this software and
 	its documentation for any purpose is hereby granted without fee, provided
 	that the above copyright notice appear in all copies and that both that
 	copyright notice and this permission notice appear in supporting
 	documentation, and that the name of the copyright holder not be used
 	in advertising or publicity pertaining to distribution of the software
 	without specific, written prior permission.  The copyright holder makes no
 	representations about the suitability of this software for any purpose.
 	It is provided "as is" without express or implied warranty.

 	THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 	INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 	EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 	CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 	DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 	TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 	PERFORMANCE OF THIS SOFTWARE.

	pthread wrappers
	
	2008-05-02 kio	made PLock a subclass of pthread_mutex_t (instead having a data member of this class)
*/


#ifndef pthreads_h
#define	pthreads_h

#include <pthread.h>
#include "kio/kio.h"



// =====================================================================
// simple lock
// lock can only be released by lock owner

class PLock : private pthread_mutex_t
{

public:
	enum Attr 
	{ 
		normal		= PTHREAD_MUTEX_NORMAL,			// no usage error checking
		recursive	= PTHREAD_MUTEX_RECURSIVE,		// recursive locking allowed
		errorcheck	= PTHREAD_MUTEX_ERRORCHECK,		// w/ usage error checking
	};

			PLock	( )						{ int e = pthread_mutex_init(this,NULL); XTRAP(e); }
			PLock	( Attr );	
			~PLock	( )						{ int e = pthread_mutex_destroy(this); XXXTRAP(e); }

	void	Lock	( ) volatile throw()	{ int e = pthread_mutex_lock(const_cast<PLock*>(this));    XXXTRAP(e); }
	void	Unlock	( ) volatile throw()	{ int e = pthread_mutex_unlock(const_cast<PLock*>(this));  XXXTRAP(e); }
	bool	TryLock	( )	volatile throw()	{ return pthread_mutex_trylock(const_cast<PLock*>(this)) == 0; }	// yes = success
};


class RPLock : public PLock		// convenience class: recursive mutex lock
{
public:		RPLock	( )				: PLock( recursive ) { }
};


// =====================================================================
// class which locks a PLock in it's creator and 
// unlocks it in it's destructor

class PLocker
{
	PLock volatile*			lock;
public:
					PLocker		( PLock volatile* p )	{ (lock=p)->Lock(); }
					PLocker		( PLock volatile& p )	{ (lock=&p)->Lock(); }
					~PLocker	()						{  lock->Unlock();  }
};
		


// =====================================================================
// basic semaphore
// lock can be released by owner or any other thread
// recursive locking is possible						kio 2006-10-14
// recursive locking is _not_ possible					kio 2007-05-09
// unlocking of unlocked semaphore is silently ignored

class PSema
{
	pthread_mutex_t	lock;
	pthread_cond_t	cond;
	pthread_t		who;
	int				sema;
	cstr			name;

public:	
					PSema		();
					PSema		( cstr static_name_str );
					~PSema		();

	void			Lock		();
	void			Unlock		();
	bool			TryLock		();					// return 1  =>  got it
	bool			TryLock		( double timeout );	// return 1  =>  got it
//	bool			WaitUnlock	( double timeout );	// wait for s.o. else to unlock & return 1  =>  got it
	pthread_t		Owner		()					{ return sema>0?who:0; }
	bool			IsLockedByMe()					{ return sema>0&&who==pthread_self(); }
}; 



// Timer
extern double CurrentTime		( );

extern void WaitDelay			( double delay );
extern void	WaitUntil			( double time );

extern void	ExecuteEvery		( double delay, bool(*fu)(void*), void* arg );
extern void	ExecuteWithDelay	( double delay, bool(*fu)(void*), void* arg );

extern void	ExecuteAt			( double time,  double(*fu)(void*), void* arg );
extern void	ExecuteAfter		( double delay, double(*fu)(void*), void* arg );




#endif






